{*******************************************************}
{                                                       }
{          Turbo Pascal 7.0                             }
{          FreePacal                                    }
{          Delphi 4                                     }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit Cursors;
interface
uses Views, Drivers, Objects, Dialogs, App, Register;
type
    PCursor = ^TCursor;
    TCursor = object(TView)
      EditView:PView;
      ComponentClipBoard:PCollection;
      constructor Init;
      destructor Done;virtual;

      procedure HandleEvent(var Event:TEvent);virtual;
      procedure Draw;virtual;
      procedure SetLink(ALink:PView);
      procedure SetState(AState: Word; Enable: Boolean);virtual;
      procedure RecalcBounds;
      procedure DragLinkView(Event: TEvent; Mode: Byte;
              var Limits: TRect; MinSize, MaxSize: TPoint);
      procedure SelectView( AWin:PGroup;var Event:TEvent);

      function  FindView( Dialog:PGroup; Event:TEvent):PView;
      function GetPalette: PPalette; virtual;
      function GetHelpCtx: Word; virtual;
      procedure SetGrowMode(AGrowMode:word);
    end;

const
   _Cursor:PCursor = nil;

implementation
uses Ext_Str, Ext_Math, Ext_Dos, Global, ResTools;

constructor TCursor.Init;
var
  R:TRect;
begin
  R.Assign(1,1,10,2);
  inherited Init(R);
  Options := Options or (ofSelectable + ofFirstClick +
             ofPostProcess);
  EventMask := EventMask or evBroadcast or evCommand;
  ComponentType:=ctCursor;
end;

destructor TCursor.Done;
begin
  _Cursor:=nil;
  if ComponentClipBoard<>nil then Dispose(ComponentClipBoard, Done);
  inherited Done;
end;

procedure TCursor.HandleEvent(var Event:TEvent);
var
  Mouse: TPoint;

procedure DragWindow(Mode: Byte);
var
  Limits: TRect;
  Min, Max: TPoint;
begin
  if EditView=nil then exit;
{  EditView^.Owner^.GetExtent(Limits);}
  Limits.A.X:=-255;
  Limits.B.X:=255;
  Limits.A.Y:=-255;
  Limits.B.Y:=255;
  EditView^.SizeLimits(Min, Max);
  if Event.What<>evKeyDown then SetState(sfDragging, True);
  DragLinkView(Event, EditView^.DragMode or Mode, Limits, Min, Max);
  if Event.What<>evKeyDown then SetState(sfDragging, False);
  ClearEvent(Event);
  RecalcBounds;
end;

procedure CopyToClipBoard;
var B:PStream;
    P:PCollection;
begin
 if EditView<>nil then
 begin
   B:=GetTempStream;
   P:=New(PCollection, Init(10,5));
   P^.Insert(EditView);
   B^.Put(P);
   P^.DeleteAll;
   Dispose(P, Done);
   if ComponentClipBoard<>nil then
      Dispose(ComponentClipBoard, Done);
   B^.Seek(0);
   ComponentClipBoard:=PCollection(B^.Get);
   B^.Free;
   SetCmdState([cmPaste],true);
 end;
end;

procedure PasteFromClipBoard;
var B:PStream;
    P:PCollection;
    Mask:string;
procedure DoInsert(P:PView);far;
var i:Longint;
    EndDigit:boolean;
begin
  P^.MoveTo(1,1);
  EndDigit:=false;
  i:=Length(P^.ComponentName^);
  while (I>1) and (P^.ComponentName^[i] in Digit) do
  begin
    EndDigit:=true;
    dec(i);
  end;
  if EndDigit then Mask:=Copy(P^.ComponentName^,1,i) else Mask:=P^.ComponentName^;
  Mask:=Mask+'%d';
  i:=1;
  while TestComponentName(Format(Mask,I), Owner) do inc(i);
  AssignStr(P^.ComponentName,Format(Mask,I));
  Owner^.Insert(P);
end;
begin
 if Owner<>nil then
 begin
   B:=GetTempStream;
   B^.Put(ComponentClipBoard);
   B^.Seek(0);
   P:=PCollection(B^.Get);
   P^.ForEach(@DoInsert);
   P^.DeleteAll;
   Dispose(P, Done);
   B^.Free;
   Message(Application, evBroadcast, cmCurentChang, Owner);
 end;
end;
begin
  TView.HandleEvent(Event);
  case Event.What of
    evMouseDown:begin
                  MakeLocal(Event.Where, Mouse);
                  if (State and sfActive <> 0) then
                    if (Mouse.X >= Size.X - 2) then
                     DragWindow(dmDragGrow)
                    else
                     DragWindow(dmDragMove);
                end;
    evKeyDown:
              if (Event.KeyCode=kbLeft) or (Event.KeyCode=kbRight)
                 or (Event.KeyCode=kbUp) or (Event.KeyCode=kbDown) then
              begin
                 if GetShiftState and $03 = 0 then DragWindow(dmDragMove)
                    else DragWindow(dmDragGrow);
                 ClearEvent(Event);
              end;
    evCommand:
      begin
        case Event.Command of
          cmCopy:CopyToClipBoard;
          cmPaste:PasteFromClipBoard;
        else
          exit;
        end;
        ClearEvent(Event);
      end;
  end;
end;

procedure TCursor.Draw;
var
  C:char;
  CFrame,Width:word;
  B: TDrawBuffer;
begin
  if State and sfDragging <> 0 then
     CFrame := $0505
  else
    if State and sfActive = 0 then
       CFrame := $0101
    else
       CFrame := $0503;
  CFrame := GetColor(CFrame);
  Width := Size.X;
  if (State and sfActive) = 0 then
    C:=''
  else
    if (State and sfDragging) = 0 then
       C:=#205
    else
       C:=#208;
  MoveCStr( B, DupChar( C, Width), CFrame);
  MoveCStr(B[Width - 2], '~~', CFrame);
  if (State and sfDragging) = 0 then
    MoveCStr(B[0], #192, CFrame);
  WriteLine(0, 0, Size.X, 1, B);
end;

procedure TCursor.SetGrowMode(AGrowMode:word);
begin
  GrowMode:=(GrowMode and (not gfGrowAll)) or
            (EditView^.GrowMode and (gfGrowLoX or gfGrowHiX)) or
            (((EditView^.GrowMode and gfGrowHiY) shr 2) or
            (EditView^.GrowMode and gfGrowHiY));
end;

procedure TCursor.SetLink(ALink:PView);
begin
   if ALink=@Self then exit;
   if Owner<>nil then Owner^.Delete(@Self);
   EditView:=ALink;
   if EditView<>nil then
   begin
     RecalcBounds;
     EditView^.Owner^.Insert(@Self);
     SetGrowMode(EditView^.GrowMode);
   end
   else
     if ComponentClipBoard<>nil then
     begin
       Dispose(ComponentClipBoard, Done);
       ComponentClipBoard:=nil;
     end;

   if Owner<>nil then
   begin
     Owner^.Select;
     Select;
     Owner^.ReDraw;
   end;
   Message(Application, evBroadcast, cmCurSelect, EditView);
end;

function TCursor.GetPalette: PPalette;
const
  P: String[Length(CFrame)] = CFrame;
begin
  GetPalette := @P;
end;

procedure TCursor.SetState(AState: Word; Enable: Boolean);
begin
  TView.SetState(AState, Enable);
  if AState and (sfActive + sfDragging) <> 0 then DrawView;
  if AState and sfSelected <> 0 then
  begin
     SetCmdState([cmCopy],Enable);
     SetCmdState([cmPaste],Enable and (ComponentClipBoard<>nil));
  end;
end;


procedure TCursor.RecalcBounds;
var
  AX,AY,BX:Byte;
  R:TRect;
begin
   AX:=EditView^.Origin.X;
   AY:=EditView^.Origin.Y+EditView^.Size.Y;
   BX:=EditView^.Size.X+EditView^.Origin.X;
   if (BX-AX)<3 then
      BX:=AX+3;
   R.Assign(AX,AY,BX,AY+1);
   SetBounds(R);
end;


procedure TCursor.DragLinkView(Event: TEvent; Mode: Byte;
        var Limits: TRect; MinSize, MaxSize: TPoint);
var
  P, S: TPoint;
  SaveBounds: TRect;
  Modif:Boolean;

procedure MoveGrow(P, S: TPoint);
var
  R: TRect;
begin
  with EditView^ do
  begin
     S.X := Min(Max(S.X, MinSize.X), MaxSize.X);
     S.Y := Min(Max(S.Y, MinSize.Y), MaxSize.Y);
     P.X := Min(Max(P.X, Limits.A.X - S.X + 1), Limits.B.X - 1);
     P.Y := Min(Max(P.Y, Limits.A.Y - S.Y + 1), Limits.B.Y - 1);
     if Mode and dmLimitLoX <> 0 then P.X := Max(P.X, Limits.A.X);
     if Mode and dmLimitLoY <> 0 then P.Y := Max(P.Y, Limits.A.Y);
     if Mode and dmLimitHiX <> 0 then P.X := Min(P.X, Limits.B.X - S.X);
     if Mode and dmLimitHiY <> 0 then P.Y := Min(P.Y, Limits.B.Y - S.Y);
{     if P.X<0 then P.X:=0;}
{     if (P.X+S.X<Owner^.Size.X+1) and (P.Y+S.Y<Owner^.Size.Y+1) then
     begin                  }
       R.Assign(P.X, P.Y, P.X + S.X, P.Y + S.Y);
       Locate(R);
{     end;}
  end;
  RecalcBounds;
  Owner^.Redraw;
end;

procedure Change(DX, DY: Integer);
begin
  if (Mode and dmDragMove <> 0) and (GetShiftState and $03 = 0) then
  begin
    Inc(P.X, DX);
    Inc(P.Y, DY);
  end else
  if (Mode and dmDragGrow <> 0) and (GetShiftState and $03 <> 0) then
  begin
    Inc(S.X, DX);
    Inc(S.Y, DY);
  end;
end;

procedure Update(X, Y: Integer);
begin
  if Mode and dmDragMove <> 0 then
  begin
    P.X := X;
    P.Y := Y;
  end;
end;

begin
   Modif:=false;
   with EditView^ do
   begin
     if Event.What = evMouseDown then
     begin
       SetState(sfDragging, True);
       if Mode and dmDragMove <> 0 then
       begin
         P.X := Origin.X - Event.Where.X;
         P.Y := Origin.Y - Event.Where.Y;
         Modif:=(P.X+P.Y)=0;
         repeat
           Inc(Event.Where.X, P.X);
           Inc(Event.Where.Y, P.Y);
           MoveGrow(Event.Where, Size);
         until not MouseEvent(Event, evMouseMove);
       end else
       begin
         P.X := Size.X - Event.Where.X;
         P.Y := EditView^.Size.Y - Event.Where.Y;
         Modif:=(P.X+P.Y)=0;
         repeat
           Inc(Event.Where.X, P.X);
           Inc(Event.Where.Y, P.Y);
           MoveGrow(Origin, Event.Where);
         until not MouseEvent(Event, evMouseMove);
       end;
       SetState(sfDragging, False);
     end
     else
     begin
          P := Origin;
          S := Size;
          case Event.KeyCode and $FF00 of
            kbLeft: Change(-1, 0);
            kbRight: Change(1, 0);
            kbUp: Change(0, -1);
            kbDown: Change(0, 1);
          end;
          MoveGrow(P, S);
      end;
  end;
  Message(Application, evBroadcast, cmCurentChang, Owner);
end;

function TCursor.FindView( Dialog:PGroup; Event:TEvent):PView;

function ContainsMouse(P: PView): Boolean; far;
begin
  ContainsMouse := (P^.State and sfVisible <> 0) and
    P^.MouseInView(Event.Where);
end;
var P,PP:PView;
begin
  P:=nil;
  if (Event.What and PositionalEvents <> 0) then
  begin
      P:=Dialog^.FirstThat(@ContainsMouse);
      if (P<>nil) and (P^.ComponentType in GroupControl) then
      begin
        PP:=PGroup(P)^.FirstThat(@ContainsMouse);
        if (PP<>nil) and not (PP^.ComponentType in BackGComp) then P:=PP;
      end;
  end;
  FindView:=P;
end;

function TCursor.GetHelpCtx:Word;
begin
  if EditView<>nil then GetHelpCtx:=EditView^.HelpCtx
  else GetHelpCtx:=0
end;

procedure TCursor.SelectView( AWin:PGroup; var Event:TEvent);
var S:String;
    PV:PDialog;
    View,AView:PView;
begin
  if TypeOf(DeskTop^.Current^)=TypeOf(TDialog) then
  begin
    View:=FindView( AWin, Event);
    if (View<>nil) and (View<>PView(_Cursor)) and
       (TypeOf(View^)<>TypeOf(TFrame)) then
    begin
       if View^.ComponentType in GroupControl then
       begin
          AView:=FindView( PGroup(View), Event);
          if (AView<>nil) and (not (AView^.ComponentType in BackGComp)) then
             View:=AView;
       end;
       SetLink(View);
       ClearEvent(Event);
    end;
  end;
end;

end.
