{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       Repots                                          }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit Reports;
interface
uses Objects, Views, Dialogs, Drivers, DAE, DataCtrl,
     Ext_Prn, Ext_Type, Clases, Register, Parser, Ext_Sys;

const
  evPrint                = $1000;
  cmBeforePrinBand       = 1;
  cmAfterPrintBand       = 2;

type
   TReportBandType = (rbtDocHeader, rbtPageHeader, rbtPageFooter,
                      rbtDocFooter, rbtMain, rbtDetail, rbtBandHeader,
                      rbtBandFooter, rbtGroupHeader, rbtGroupFooter);

   TReportBandOptions = set of ( rbFramed, rbColored,
                                 rbHScrollBar, rbVScrollBar,
                                 rbNewPagePrint, rbStreched,
                                 rbBreaked);
type
  PVirtualDataSet = ^TVirtualDataSet;
  TVirtualDataSet = object(TDataSourse)
    fRecordCount:Longint;
    constructor Init(ARecordCount:Longint);
    function EOF:boolean;virtual;
    function RecNo:LongInt;virtual;
    procedure Go( APos:Longint);virtual;
    function NextRecEx:LongInt;virtual;
    function FirstRecEx:LongInt;virtual;
    function RecordCount:Longint;virtual;
    constructor Load(var S:TStream);
    procedure Store(var S:TStream);
    function FieldDefs:PCollection;virtual;{$IFDEF EAM}abstract;{$ENDIF}
  private
    fRecNo:Longint;
  end;

  PUserDataSet = ^TUserDataSet;
  TUserDataSet = object(TDataSourse)
    constructor Init;
    function EOF:boolean;virtual;
    function RecNo:LongInt;virtual;
    procedure Go( APos:Longint);virtual;
    function NextRecEx:LongInt;virtual;
    function FirstRecEx:LongInt;virtual;
    function RecordCount:Longint;virtual;
    constructor Load(var S:TStream);
    procedure Store(var S:TStream);
    function FieldDefs:PCollection;virtual;{$IFDEF EAM}abstract;{$ENDIF}
  end;

type
  PPageItem = ^TPageItem;
  TPageItem = object(TObject)
    Text:PString;
    X,Y:integer;
    constructor Init(AX,AY:integer;const S:string);
    destructor Done;virtual;
  end;

  PPageStrCollection = ^TPageStrCollection;
  TPageStrCollection = object(TSortedCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
  end;


type
  PReportBandCollection = ^TReportBandCollection;
  TReportBandCollection = object(TSortedCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
  end;

type
   PReport = ^TReport;
   TPrinterSetupEvent = procedure(Report:PReport);
   PReportBand = ^TReportBand;
   TReportBand = object(TGroup)
     Report:PReport;
     Detail:PReportBand;
     RBType:TReportBandType;
     DataSourse:PDataSourse;
     rbOption:TReportBandOptions;
     Header:PReportBand;
     Footer:PReportBand;
     constructor Init(var Bounds: TRect; AReport:PReport;
                      ADataSourse:PDataSourse;ADetail:PReportBand);
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     procedure Print;
     procedure SetDetail(ARB:PReportBand; aReportBandType:TReportBandType);virtual;
     procedure HandleEvent(var Event:TEvent);virtual;
     procedure Awaken;virtual;
     procedure NewDataSourse(ADataSourse:PDataSourse);virtual;
     function GetBufLine(i:integer):string;
     procedure SetState(AState: Word; Enable: Boolean); virtual;
     function GetSizeY:integer;
   private
     HScrollBar:PScrollBar;
     VScrollBar:PScrollBar;
     BackGround:PView;
     Delta:TPoint;
     LineNo:integer;
     LockedScrollBar:boolean;
     procedure ScrollDraw(DeltaX,DeltaY:integer);
     procedure BeforePrint;virtual;
     procedure AfterPrint;virtual;
   end;

   PGroupBand = ^TGroupBand;
   TGroupBand = object(TReportBand)
     GroupExpresion:PString;
     constructor Init(var Bounds: TRect; AReport:PReport;
                      const AGroupExpresion:string);
     destructor Done;virtual;
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     procedure SetGroupExpresion(const S:string);
     procedure NewDataSourse(ADataSourse:PDataSourse);virtual;
     procedure SetDetail(ARB:PReportBand; aReportBandType:TReportBandType);virtual;
     procedure CheckPrint;
   private
     GroupOPZStr:PString;
     OldExpresion:PString;
     fParser:PParser;
     procedure BeforePrint;virtual;
   end;

  TReport = object(TView)
    DocHeaderBand:PReportBand;
    PageHeaderBand:PReportBand;
    PageFooterBand:PReportBand;
    DocFooterBand:PReportBand;
    MainBand:PReportBand;
    PageHight:integer;
    PageWidth:integer;
    ShowProgres:boolean;
    DoublePass:boolean;
    LeftMargin:integer;
    RightMargin:integer;
    TopMargin:integer;
    BottomMargin:integer;
    ColumnCount:integer;
    ColumnMargin:integer;
    CurCollumn:integer;
    PrintLines:Longint;
    Printing:Boolean;
    constructor Init;
    constructor Load(var S:TStream);
    destructor Done;virtual;
    procedure Awaken;virtual;
    procedure Store(var S:TStream);
    procedure Draw;virtual;
    procedure Print;
    procedure OnPrintLine(Lines:integer);
    procedure OnPrintBand;
    function CurentPage:integer;
    procedure HandleEvent(var Event:TEvent);virtual;
    procedure PreviewReport;
    procedure SaveReport;
    procedure SetReportBand(ARB:PReportBand; ARBType:TReportBandType);
    procedure NewPage;
    procedure PrinterSetup;
    procedure Stop(AMsg:string);

    procedure WriteStr(S:string; Band:PReportBand);
    procedure WriteStrXY(X,Y:integer; S:string);
    procedure Clear;
  private
    CurPage:integer;
    CurentLine:integer;
    MaxLineInPage:integer;
    LastPrintLine:string;
    GroupBandCollection:PCollection;
    isError:boolean;
    Msg:PString;
    _Rec:PPrinterRec;
    Strings:PPageStrCollection;
    StatusForm:PView;
    CurentPass:integer;
    function StartPageY:integer;
    function EndPageY:integer;
    function Str(I:integer):string;
    procedure OutToPrn;
    procedure UpdateStatus;
  end;

type
   TRLStyle = (rlsDate, rlsTime, rlsDateTime, rlsPageNo, rlsLineNo);
type
   PReportLabel = ^TReportLabel;
   TReportLabel = object(TView)
     RLStyle:TRLStyle;
     Report:PReport;
     ParamStr:PString;
     constructor Init(var Bounds: TRect; AReport:PReport; ARLStyle:TRLStyle; AParam:String);
     destructor Done;virtual;
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     procedure Draw;virtual;
     procedure SetParam(AParam:string);
     procedure HandleEvent(var Event:TEvent);virtual;
   private
     LastState:byte;
   end;

type
   PReportMemo = ^TReportMemo;
   TReportMemo = object(TView)
     Field:PField;
     DS:PDataSourse;
     FieldName:PString;
     AutoSize:Boolean;
     constructor Init(Bounds:TRect; ADS:PDataSourse; Name:String);
     destructor Done;virtual;
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     procedure Draw;virtual;
     function GetPalette: PPalette; virtual;
     procedure HandleEvent(var Event:TEvent);virtual;
     procedure Update;
     function GetVertSize:integer;
   private
     TextBuf:PMemoRec;
     SaveSize:TPoint;
   end;

type
   TRShapeType = (rsRectangle,rsTopAndBottom,rsRightAndLeft);
   TPenStyle   = (psSingleLine, psDoubleLine);

type
   PReportShape = ^TReportShape;
   TReportShape = object(TView)
     Shape:TRShapeType;
     PenStyle:TPenStyle;
     constructor Init(Bounds:TRect; AShape:TRShapeType; APenStyle:TPenStyle);
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     procedure Draw;virtual;
   private
     function GetChar(X,Y:integer):char;
     function ScanFrame(AX,AY:integer):byte;
   end;

{Preview dialog}
type
   PTextView = ^TTextView;
   TTextView = object(TScroller)
     Strings:PStrings;
     Stream:PStream;
     constructor Init(var Bounds: TRect; AHScrollBar, AVScrollBar:
                                         PScrollBar; S:PStream);
     destructor Done; virtual;
     procedure Draw; virtual;
     procedure ReadFile(StartPos, EndPos:Longint);
     procedure SetState(AState: Word; Enable: Boolean); virtual;
     procedure FormatLine(var DrawBuf; Line: string;
               Width: Integer; C: byte);
   end;

type
   PPreviewDialog = ^TPreviewDialog;
   TPreviewDialog = object(TDialog)
     InfoPane:PStaticText;
     TextView:PTextView;
     PrinterRec:PPrinterRec;
     CurPage:PPageBreak;
     constructor Init(S:PPrinterRec);
     destructor Done;virtual;
     procedure HandleEvent(var Event:TEvent);virtual;
     procedure UpdateParam;
     procedure FistPagePreView;
     procedure PrevPagePreView;
     procedure NextPagePreView;
     procedure LastPagePreView;
   end;

const
  RReportBand: TStreamRec = (
     ObjType: rcReportBand;
     VmtLink: Ofs(TypeOf(TReportBand)^);
     Load:    @TReportBand.Load;
     Store:   @TReportBand.Store
  );
const
  RReport: TStreamRec = (
     ObjType: rcReport;
     VmtLink: Ofs(TypeOf(TReport)^);
     Load:    @TReport.Load;
     Store:   @TReport.Store
  );
const
  RReportLabel: TStreamRec = (
     ObjType: rcReportLabel;
     VmtLink: Ofs(TypeOf(TReportLabel)^);
     Load:    @TReportLabel.Load;
     Store:   @TReportLabel.Store
  );
const
  RReportMemo: TStreamRec = (
     ObjType: rcReportMemo;
     VmtLink: Ofs(TypeOf(TReportMemo)^);
     Load:    @TReportMemo.Load;
     Store:   @TReportMemo.Store
  );
const
  RReportShape: TStreamRec = (
     ObjType: rcReportShape;
     VmtLink: Ofs(TypeOf(TReportShape)^);
     Load:    @TReportShape.Load;
     Store:   @TReportShape.Store
  );
const
  RVirtualDataSet: TStreamRec = (
     ObjType: rcVirtualDataSet;
     VmtLink: Ofs(TypeOf(TVirtualDataSet)^);
     Load:    @TVirtualDataSet.Load;
     Store:   @TVirtualDataSet.Store
  );
const
  RUserDataSet : TStreamRec = (
     ObjType: rcUserDataSet;
     VmtLink: Ofs(TypeOf(TUserDataSet)^);
     Load:    @TUserDataSet .Load;
     Store:   @TUserDataSet .Store
  );
const
  RGroupBand : TStreamRec = (
     ObjType: rcGroupBand;
     VmtLink: Ofs(TypeOf(TGroupBand)^);
     Load:    @TGroupBand .Load;
     Store:   @TGroupBand .Store
  );

type
  TGetFormType = (gftSinglePassDlg, gftDoublePassDlg);
  TGetPrintForm = function(GetFormType:TGetFormType):PView;
procedure RegisterReports;
function StdGetPrintForm(GetFormType:TGetFormType):PView;
const
  GetPrintForm:TGetPrintForm = StdGetPrintForm;
  OnPrinterSetupEvent:TPrinterSetupEvent = nil;

procedure UnionLines(var S1,S2:string);
implementation
uses Ext_Dos, Ext_Math, Ext_Str, AddCntrl, GauGes, App, Memory,
     ResTools, Dos, MsgBox;

{var PrintProgresDlg:PView;
procedure DoPrint(Cur,Max:Longint);
var A:TPersentRec;
begin
  if PrintProgresDlg=nil then exit;
  A.MaxValue:=Max;
  A.CurValue:=Cur;
  PrintProgresDlg^.SetData(A)
end;}

type
   PReportBandBackground = ^TReportBandBackground;
   TReportBandBackground = object(TView)
     constructor Init(var Bounds: TRect);
     procedure Draw;virtual;
     procedure FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);
   end;

const
  RReportBandBackground: TStreamRec = (
     ObjType: 5007;
     VmtLink: Ofs(TypeOf(TReportBandBackground)^);
     Load:    @TReportBandBackground.Load;
     Store:   @TReportBandBackground.Store
  );

procedure UnionLines(var S1,S2:string);
var i:integer;
begin
  if Length(S1)<Length(S2) then
     S1:=S1+copy(S2,Length(S1)+1,255);
  for i:=1 to Length(S1) do
  begin
    case S1[i] of
      '':case S2[i] of
            '':begin
                  S1[i]:='';
                  S2[i]:='';
                end;
            '':begin
                  S1[i]:='';
                  S2[i]:='';
                end;
          end;
      '':case S2[i] of
            '':begin
                  S1[i]:='';
                  S2[i]:='';
                end;
            '':begin
                  S1[i]:='';
                  S2[i]:='';
                end;
            '':S1[i]:='';
          end;
      '':case S2[i] of
            '':S2[i]:=' ';
            '','',''
               :begin
                  S1[i]:='';
                  S2[i]:='';
                end;
          end;
      '':case S2[i] of
            '':S2[i]:=' ';
            '','',''
               :begin
                  S1[i]:='';
                  S2[i]:='';
                end;
          end;
    end;
  end;
end;


  {TPageItem}
constructor TPageItem.Init(AX,AY:integer;const S:string);
begin
  inherited Init;
  X:=AX; Y:=AY;
  Text:=NewStr(TrimRight(S));
end;

destructor TPageItem.Done;
begin
  AssignStr(Text, '');
  inherited Done;
end;

  {TPageStrCollection}
function TPageStrCollection.Compare(Key1, Key2: Pointer): Integer;
var R1:PPageItem absolute Key1;
    R2:PPageItem absolute Key2;
begin
  if R1^.Y=R2^.Y then
  begin
    if R1^.X<R2^.X then Compare:=-1 else
    if R1^.X>R2^.X then Compare:=1 else
    Compare:=0
  end
  else
  if R1^.Y<R2^.Y then Compare:=-1
  else Compare:=1;
end;


  {TReportBandCollection}
function TReportBandCollection.Compare(Key1, Key2: Pointer): Integer;
var R1:PView absolute Key1;
    R2:PView absolute Key2;
begin
  if R1^.Origin.Y=R2^.Origin.Y then
  begin
    if R1^.Origin.X<R2^.Origin.X then Compare:=-1 else
    if R1^.Origin.X>R2^.Origin.X then Compare:=1 else
    Compare:=0
  end
  else
  if R1^.Origin.Y<R2^.Origin.Y then Compare:=-1
  else Compare:=1;
end;

  { TReportBandBackground }
constructor TReportBandBackground.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  GrowMode := gfGrowHiX + gfGrowHiY;
  ComponentType:=ctReportBandBackground;
end;

procedure TReportBandBackground.FrameLine(var FrameBuf; Y, N: Integer;
  Color: Byte); assembler;

const
  InitFrame: array[0..17] of Byte =
    ($06, $0A, $0C, $05, $00, $05, $03, $0A, $09,
     $16, $1A, $1C, $00, $00, $00, $13, $1A, $19);
  FrameChars: array[0..31] of Char =
    '          ϻ ';

var
  FrameMask: array[0..MaxViewWidth-1] of Byte;
asm
{        LES     BX,Self
        MOV     DX,ES:[BX].TReportBandBackground.Size.X
        MOV     CX,DX
        DEC     CX
        DEC     CX
        MOV     SI,OFFSET InitFrame
        ADD     SI,N
        LEA     DI,FrameMask
        PUSH    SS
        POP     ES
        CLD
        MOVSB
        LODSB
        REP     STOSB
        MOVSB
        LES     BX,Self
        LES     BX,ES:[BX].TFrame.Owner
        LES     BX,ES:[BX].TGroup.Last
        DEC     DX
@1:     LES     BX,ES:[BX].TView.Next
        CMP     BX,WORD PTR Self[0]
        JNE     @2
        MOV     AX,ES
        CMP     AX,WORD PTR Self[2]
        JE      @10
@2:     TEST    ES:[BX].TView.Options,ofFramed
        JE      @1
        TEST    ES:[BX].TView.State,sfVisible
        JE      @1
        MOV     AX,Y
        SUB     AX,ES:[BX].TView.Origin.Y
        JL      @3
        CMP     AX,ES:[BX].TView.Size.Y
        JG      @1
        MOV     AX,0005H
        JL      @4
        MOV     AX,0A03H
        JMP     @4
@3:     INC     AX
        JNE     @1
        MOV     AX,0A06H
@4:     MOV     SI,ES:[BX].TView.Origin.X
        MOV     DI,ES:[BX].TView.Size.X
        ADD     DI,SI
        CMP     SI,1
        JG      @5
        MOV     SI,1
@5:     CMP     DI,DX
        JL      @6
        MOV     DI,DX
@6:     CMP     SI,DI
        JGE     @1
        OR      BYTE PTR FrameMask[SI-1],AL
        XOR     AL,AH
        OR      BYTE PTR FrameMask[DI],AL
        OR      AH,AH
        JE      @1
        MOV     CX,DI
        SUB     CX,SI
@8:     OR      BYTE PTR FrameMask[SI],AH
        INC     SI
        LOOP    @8
        JMP     @1
@10:    INC     DX
        MOV     AH,Color
        MOV     BX,OFFSET FrameChars
        MOV     CX,DX
        LEA     SI,FrameMask
        LES     DI,FrameBuf
@11:    SEGSS   LODSB
        XLAT
        STOSW
        LOOP    @11}
end;

procedure TReportBandBackground.Draw;
var
  F, I,DX: Integer;
  B: TDrawBuffer;
  Color:byte;
function ColorOfs:byte;
begin
  case PReportBand(Owner)^.RBType of
        rbtDocHeader,rbtDocFooter:ColorOfs:=4;
        rbtPageHeader,rbtPageFooter,
        rbtBandHeader,rbtBandFooter:ColorOfs:=10;
        rbtMain:ColorOfs:=16;
        rbtDetail:ColorOfs:=32;
  end;
end;
begin
  if (rbColored in PReportBand(Owner)^.rbOption) and
     not PReportBand(Owner)^.Report^.Printing  then
     Color:=GetColor(ColorOfs)
  else Color:=GetColor(4);
  if rbFramed in PReportBand(Owner)^.rbOption then
  begin
     F:=9;
     FrameLine(B, 0, F+3, Color);
     WriteLine(0, 0, Size.X, 1, B);
     for I := 1 to Size.Y - 2 do
     begin
       FrameLine(B, I, F + 3, Color);
       WriteLine(0, I, Size.X, 1, B);
     end;
     FrameLine(B, Size.Y - 1, F + 3, Color);
     WriteLine(0, Size.Y - 1, Size.X, 1, B);
  end
  else
  begin
     with PReportBand(Owner)^ do
       if (Report<>nil) and (Report^.Printing) then
           MoveChar(B,' ', Color,Size.X)
       else
       begin
         MoveChar(B,'.', Color,Size.X);
         if (RBType in [rbtMain, rbtDetail, rbtBandHeader, rbtBandFooter,
                        rbtGroupHeader, rbtGroupFooter])
            and (Report<>nil) and (Report^.ColumnCount>1) then
         begin
           DX:=(Report^.PageWidth div Report^.ColumnCount)+Report^.ColumnMargin;
           F:=0;
           i:=1;
           while (F<Size.X) and (I<=Report^.ColumnCount) do
           begin
             inc(F,DX);
             MoveChar(B[F],'|', Color,1);
             inc(i);
           end;
         end;
       end;
     WriteLine(0, 0, Size.X, Size.Y, B);
  end;
end;


  { TReportBand }

constructor TReportBand.Init(var Bounds: TRect; AReport:PReport;
                      ADataSourse:PDataSourse;ADetail:PReportBand);
var R:TRect;
begin
  inherited Init(Bounds);
  GrowMode:=GrowMode or gfGrowHiX;
  GetExtent(R);
  BackGround:=New(PReportBandBackground, Init(R));
  Insert(BackGround);
  GetExtent(R);
  R.A.Y:=R.B.Y-1;
  Dec(R.B.X);
  HScrollBar:=New(PScrollBar, Init(R));
    HScrollBar^.ComponentType:=ctReportBandBackground;
    HScrollBar^.SetParams(0,0,250,1,1);
    Insert(HScrollBar);
  GetExtent(R);
  R.A.X:=R.B.X-1;
  Dec(R.B.Y);
  VScrollBar:=New(PScrollBar, Init(R));
    VScrollBar^.ComponentType:=ctReportBandBackground;
    VScrollBar^.SetParams(0,0,250,1,1);
    Insert(VScrollBar);
  Report:=AReport;
  Detail:=ADetail;
  ComponentType:=ctReportBand;
  NewDataSourse(ADataSourse);
  EventMask:=$FFFF;
end;

const
  B:boolean = false;
constructor TReportBand.Load(var S:TStream);
begin
  B:=true;
  inherited Load(S);
  B:=false;
  GetSubViewPtr(S,VScrollBar);
  GetSubViewPtr(S,HScrollBar);
  GetSubViewPtr(S,BackGround);
  GetPeerViewPtr(S,Detail);
  GetPeerViewPtr(S,DataSourse);
  GetPeerViewPtr(S,Report);
  GetPeerViewPtr(S,Header);
  GetPeerViewPtr(S,Footer);

  S.Read(RBType, SizeOf(TReportBandType));
  S.Read(rbOption, SizeOf(byte));
  S.Read(Delta, SizeOf(TPoint));
end;

procedure TReportBand.Store(var S:TStream);
begin
  inherited Store(S);
  PutSubViewPtr(S,VScrollBar);
  PutSubViewPtr(S,HScrollBar);
  PutSubViewPtr(S,BackGround);
  PutPeerViewPtr(S,Detail);
  PutPeerViewPtr(S,DataSourse);
  PutPeerViewPtr(S,Report);
  PutPeerViewPtr(S,Header);
  PutPeerViewPtr(S,Footer);

  S.Write(RBType, SizeOf(TReportBandType));
  S.Write(rbOption, SizeOf(byte));
  S.Write(Delta, SizeOf(TPoint));
end;

procedure TReportBand.Awaken;
begin
  if B then exit;
  inherited Awaken;
end;

procedure TReportBand.NewDataSourse(ADataSourse:PDataSourse);
begin
  DataSourse:=ADataSourse;
end;

function TReportBand.GetBufLine(i:integer):string;assembler;
asm
{        PUSH    DS
        MOV     BX,I
        CLD
        LDS     SI,Self
        LES     DI,@Result
        MOV     AX,DS:[SI].TView.Size.X
        MOV     CX,AX
        STOSB
        MUL     BX
        SHL     AX,1
        LDS     SI,DS:[SI].TGroup.Buffer
        ADD     SI,AX
@@1:    LODSW
        STOSB
        LOOP    @@1
        POP     DS}
end;

function TReportBand.GetSizeY:integer;
var MaxY:integer;
{procedure DoFind(P:PView);far;
begin
  if P^.ComponentType=ctReportBandBackground then exit;
  if (P^.ComponentType=ctReportMemo) and (rbStreched in rbOption) then
     Max.Y:=Ext_Math.Max(Max.Y,P^.Origin.Y+PReportMemo(P)^.GetVertSize)
  else
    Max.Y:=Ext_Math.Max(P^.Origin.Y+P^.Size.Y,Max.Y);
  Max.X:=Ext_Math.Max(P^.Origin.X+P^.Size.X,Max.X);
end;}
begin
  GetSizeY:=Size.Y;
end;

procedure TReportBand.Print;
var i:integer;
    L:longint;
    SaveStateVis,SaveStateExp:word;
    Max,OldDelta:TPoint;
    OldR,R:TRect;
    SaveOwner:PGroup;
    NewLine:string;


procedure DoFind(P:PView);far;
begin
  if P^.ComponentType=ctReportBandBackground then exit;
  if (P^.ComponentType=ctReportMemo) and (rbStreched in rbOption) then
     Max.Y:=Ext_Math.Max(Max.Y,P^.Origin.Y+PReportMemo(P)^.GetVertSize)
  else
    Max.Y:=Ext_Math.Max(P^.Origin.Y+P^.Size.Y,Max.Y);
  Max.X:=Ext_Math.Max(P^.Origin.X+P^.Size.X,Max.X);
end;

procedure PrintDetailBand;
begin
   if Detail^.Header<>nil then Detail^.Header^.Print;
   if Detail^.DataSourse=nil then Detail^.Print
   else
   begin
        Detail^.LineNo:=1;
        L:=Detail^.DataSourse^.RecNo;
        Detail^.DataSourse^.FirstRec;
        while not Detail^.DataSourse^.Eof do
        begin
          Detail^.Print;
          Detail^.DataSourse^.NextRec;
          inc(Detail^.LineNo);
        end;
        Detail^.DataSourse^.Go(L);
   end;
  if Detail^.Footer<>nil then Detail^.Footer^.Print;
end;

begin
  if (Detail<>nil) and (Origin.Y>Detail^.Origin.Y) then PrintDetailBand;
  if rbNewPagePrint in rbOption then Report^.NewPage;
  Report^.OnPrintBand;
  FillChar(Max, SizeOf(TPoint), 0);
  OldDelta:=Delta;
  ScrollDraw(-Delta.X, -Delta.Y);
  ForEach(@DoFind);
  R.Assign( 0, 0, Ext_Math.Max(Size.X,Max.X)+1, Ext_Math.Max(Size.Y,Max.Y));
  GetBounds(OldR);
  ChangeBounds(R);
  if (Options and ofBuffered <> 0) and (Buffer <> nil) then
     DisposeCache(Pointer(Buffer));
  SaveOwner:=Owner;
  Owner:=nil;
  GetMem(Buffer, Size.X*Size.Y*SizeOf(word));
  SaveStateVis:=State and sfVisible;
  SaveStateExp:=State and sfExposed;
  SetState(sfVisible,true);
  SetState(sfExposed,true);
  if rbHScrollBar in rbOption then HScrollBar^.Hide;
  if rbVScrollBar in rbOption then VScrollBar^.Hide;
  Message(@Self, evPrint, cmBeforePrinBand, @Self);
  Redraw;
{  if rbFramed in rbOption then
  begin
    NewLine:=TrimRight(GetBufLine(0));
    if Report^.LastPrintLine<>'' then
    begin
       UnionLines(Report^.LastPrintLine,NewLine);
       Report^.WriteStr(Report^.LastPrintLine, @Self);
    end
    else
    Report^.WriteStr(NewLine, @Self);
    for i:=1 to  Size.Y-2 do Report^.WriteStr(GetBufLine(i), @Self);
    Report^.LastPrintLine:=TrimRight(GetBufLine(Size.Y-1));
  end
  else}
  begin
    if Report^.LastPrintLine<>'' then
    begin
       Report^.WriteStr(Report^.LastPrintLine, @Self);
       Report^.LastPrintLine:='';
    end;
    for i:=0 to  Size.Y-1 do
    begin
       Report^.WriteStr(GetBufLine(i), @Self);
       if (rbBreaked in rbOption) and (Report<>nil) then
          Report^.OnPrintLine(1);
    end;
  end;
  if (RBType in [rbtMain, rbtDetail]) and
     (Report<>nil) and not (rbBreaked in rbOption) then
     Report^.OnPrintLine(Size.Y);
  Message(@Self, evPrint, cmAfterPrintBand, @Self);
  FreeMem(Buffer, Size.X*Size.Y*SizeOf(word));
  Buffer:=nil;
  Owner:=SaveOwner;
  SetState(sfExposed, SaveStateExp and sfExposed<>0);
  SetState(sfVisible, SaveStateVis and sfVisible<>0);
  ChangeBounds(OldR);
  ScrollDraw(OldDelta.X,OldDelta.Y);
  if Detail<>nil then
     if Origin.Y<=Detail^.Origin.Y then
     PrintDetailBand;
end;

procedure TReportBand.SetDetail(ARB:PReportBand; aReportBandType:TReportBandType);

procedure _SetDetail;
begin
  Detail:=ARB;
  if Detail<>nil then
     Detail^.SetDetail(Detail^.Detail, rbtDetail);
end;

begin
  case aReportBandType of
    rbtDetail:_SetDetail;
    rbtBandHeader:Header:=ARB;
    rbtBandFooter:Footer:=ARB;
  end;
  if ARB<>nil then
  begin
    ARB^.RBType:=aReportBandType;
    ARB^.Report:=Report;
  end;
end;

procedure TReportBand.ScrollDraw(DeltaX,DeltaY:integer);
procedure DoScroll(P:PView);far;
begin
  if P^.ComponentType<>ctReportBandBackground then
     P^.MoveTo(P^.Origin.X+DeltaX,P^.Origin.Y+DeltaY);
end;
begin
  if LockedScrollBar then exit;
  if (Delta.X+DeltaX<-250) or (Delta.X+DeltaX>0) or
     (Delta.Y+DeltaY<-250) or (Delta.Y+DeltaY>0) then exit;
  Lock;
  ForEach(@DoScroll);
  UnLock;
  BackGround^.DrawView;
  Delta.X:=Delta.X+DeltaX;
  Delta.Y:=Delta.Y+DeltaY;
  LockedScrollBar:=true;
  VScrollBar^.SetValue(-Delta.Y);
  HScrollBar^.SetValue(-Delta.X);
  LockedScrollBar:=false;
end;

procedure TReportBand.BeforePrint;
begin
end;

procedure TReportBand.AfterPrint;
begin
  ReDraw;
end;

procedure TReportBand.HandleEvent(var Event:TEvent);
procedure DestroyLinks;
begin
  if Event.InfoPtr=Report then Report:=nil;
  if Event.InfoPtr=Detail then Detail:=nil;
  if Event.InfoPtr=DataSourse then DataSourse:=nil;
  if Event.InfoPtr=Header then Header:=nil;
  if Event.InfoPtr=Footer then Footer:=nil;
end;
begin
  inherited HandleEvent(Event);
  case Event.What of
    evKeyDown:
    case Event.KeyCode of
      kbCtrlPgDn:ScrollDraw(0,-1);
      kbCtrlPgUp:ScrollDraw(0,1);
      kbCtrlLeft:ScrollDraw(1,0);
      kbCtrlRight:ScrollDraw(-1,0);
    else
      exit;
    end;
    evBroadcast:
    case Event.Command of
      cmDestroed:DestroyLinks;
      cmScrollBarChanged:if Event.InfoPtr=HScrollBar then
                            ScrollDraw(Delta.X+HScrollBar^.Value,0)
                         else
                         if Event.InfoPtr=VScrollBar then
                            ScrollDraw(0,Delta.Y+VScrollBar^.Value)
                         else
                            exit;
    else
      exit;
    end;
    evPrint:
    if Event.InfoPtr=@Self then exit else
    case Event.Command of
      cmAfterPrint:AfterPrint;
      cmBeforePrint:BeforePrint;
    else
      exit;
    end;
  else
    exit;
  end;
end;

procedure TReportBand.SetState(AState: Word; Enable: Boolean);
procedure ShowSBar(SBar: PScrollBar; Enable:boolean);
begin
  if (SBar <> nil) then
    if GetState(sfActive) and GetState(sfVisible) and Enable then SBar^.Show
    else SBar^.Hide;
end;
begin
  inherited SetState(AState, Enable);
  if AState and (sfSelected + sfActive + sfVisible) <> 0 then
  begin
    ShowSBar(HScrollBar,rbHScrollBar in rbOption);
    ShowSBar(VScrollBar,rbVScrollBar in rbOption);
    ReDraw;
  end;
end;

   { TGroupBand }
procedure GetPValueEvent(Owner:PObject; Sender:PParser; const s:String; var v:TParserValue);far;
var Band:PGroupBand absolute Owner;
    S1,S2:string[60];
    i:integer;
    DS:PDataSourse;
    F:PField;
begin
  if S='#LINE' then V.AssignNum(Band^.Report^.PrintLines) else
  if S='#PAGE' then V.AssignNum(Band^.Report^.CurPage) else
  begin
    I:=Pos('.',S);
    if I<>0 then
    begin
      S1:=Copy(S,1,I-1);
      S2:=Copy(S,I+1,255);
      DS:=Band^.Report^.Owner^.FindComponent(S1);
      if DS<>nil then
      begin
        F:=DS^.FieldByName(S2);
        if F<>nil then
        case F^.FieldType of
          ftString:V.AssignStr(F^.AsString);
          ftInteger,
          ftFloat:V.AssignNum(F^.AsFloat);
          ftBoolean:V.AssignBool(F^.AsBoolean);
          ftDate,
          ftDateTime
             :begin
                Sender^.Format:=fDate;
                V.AssignNum(F^.AsDateTime);
              end;
        end
        else Band^.Report^.Stop('');
      end else Band^.Report^.Stop('Not found '''+S1+'''');
    end;
  end;
end;

procedure OnFunctionEvent(Owner:PObject; Sender:PParser; const name:String; var p1,p2,p3, Val:string);far;
begin
end;
constructor TGroupBand.Init(var Bounds: TRect; AReport:PReport;
                            const AGroupExpresion:string);
begin
  inherited Init(Bounds, AReport, nil, nil);
  SetGroupExpresion(AGroupExpresion);
  fParser:=New(PParser, Init(@Self));
  fParser^.OnGetValue:=GetPValueEvent;
  fParser^.OnFunction:=OnFunctionEvent;
  RBType:=rbtGroupHeader;
  ComponentType:=ctGroupBand;
{  rbtGroupFooter}
end;

destructor TGroupBand.Done;
begin
  SetGroupExpresion('');
  fParser^.Free;
  inherited Done;
end;

constructor TGroupBand.Load(var S:TStream);
begin
  inherited Load(S);
  GroupExpresion:=S.ReadStr;
  fParser:=New(PParser, Init(@Self));
  fParser^.OnGetValue:=GetPValueEvent;
  fParser^.OnFunction:=OnFunctionEvent;
  if GroupExpresion<>nil then
     AssignStr(GroupOPZStr, fParser^.Str2OPZ(GroupExpresion^))
  else AssignStr(GroupOPZStr, '');
end;

procedure TGroupBand.Store(var S:TStream);
begin
  inherited Store(S);
  S.WriteStr(GroupExpresion);
end;

procedure TGroupBand.SetGroupExpresion(const S:string);
begin
  if (GroupExpresion=nil) or (stUpCase(GroupExpresion^)<>stUpCase(S)) then
     AssignStr(GroupExpresion, S);
  if GroupExpresion<>nil then
     AssignStr(GroupOPZStr, fParser^.Str2OPZ(GroupExpresion^))
  else AssignStr(GroupOPZStr, '');
end;

procedure TGroupBand.NewDataSourse(ADataSourse:PDataSourse);
begin
end;

procedure TGroupBand.SetDetail(ARB:PReportBand; aReportBandType:TReportBandType);
begin
  if aReportBandType<>rbtDetail then
     inherited SetDetail(ARB, aReportBandType);
end;

procedure TGroupBand.CheckPrint;
var R:TParserValue;
begin
  if GroupOPZStr=nil then exit;
  R.Init;
  fParser^.CalcOPZ(GroupOPZStr^,R);
  if R.AsString<>DinamicStr(OldExpresion) then
  begin
    AssignStr(OldExpresion, R.AsString);
    Print;
  end;
  R.Done;
end;

procedure TGroupBand.BeforePrint;
begin
  inherited BeforePrint;
  if Report<>nil then
     Report^.GroupBandCollection^.Insert(@Self);
end;

   { TReport }
constructor TReport.Init;
var R:TRect;
begin
  R.Assign(1,1,10,2);
  inherited Init(R);
  ComponentName:=NewStr('DataReport');
  Options:= Options or ofPreProcess;
  EventMask := evMouseDown + evKeyDown + evCommand + evBroadcast;
  PageHight:=65;
  PageWidth:=80;
  ComponentType:=ctReport;
   GroupBandCollection:=New(PCollection, Init(10,5));
end;

destructor TReport.Done;
begin
  GroupBandCollection^.DeleteAll;
  inherited Done;
end;

constructor TReport.Load(var S:TStream);
begin
  inherited Load(S);
  GetPeerViewPtr(S,DocHeaderBand);
  GetPeerViewPtr(S,PageHeaderBand);
  GetPeerViewPtr(S,PageFooterBand);
  GetPeerViewPtr(S,DocFooterBand);
  GetPeerViewPtr(S,MainBand);
  S.Read(PageHight, SizeOf(integer));
  S.Read(PageWidth, SizeOf(integer));
  S.Read(ShowProgres, SizeOf(boolean));
  S.Read(DoublePass, SizeOf(boolean));
  S.Read(LeftMargin, SizeOf(integer));
  S.Read(RightMargin, SizeOf(integer));
  S.Read(TopMargin, SizeOf(integer));
  S.Read(BottomMargin, SizeOf(integer));
  S.Read(ColumnCount, SizeOf(integer));
  S.Read(ColumnMargin, SizeOf(integer));
  GroupBandCollection:=New(PCollection, Init(10,5));
end;

procedure TReport.Store(var S:TStream);
begin
  inherited Store(S);
  PutPeerViewPtr(S,DocHeaderBand);
  PutPeerViewPtr(S,PageHeaderBand);
  PutPeerViewPtr(S,PageFooterBand);
  PutPeerViewPtr(S,DocFooterBand);
  PutPeerViewPtr(S,MainBand);
  S.Write(PageHight, SizeOf(integer));
  S.Write(PageWidth, SizeOf(integer));
  S.Write(ShowProgres, SizeOf(boolean));
  S.Write(DoublePass, SizeOf(boolean));
  S.Write(LeftMargin, SizeOf(integer));
  S.Write(RightMargin, SizeOf(integer));
  S.Write(TopMargin, SizeOf(integer));
  S.Write(BottomMargin, SizeOf(integer));
  S.Write(ColumnCount, SizeOf(integer));
  S.Write(ColumnMargin, SizeOf(integer));
end;

procedure TReport.Draw;
var
  B: TDrawBuffer;
  C:Char;
begin
  MoveChar(B[0], ' ', GetColor(1), Size.X);
  MoveStr(B[0], ComponentName^, GetColor(1));
  WriteLine(0, 0, Size.X, Size.Y, B);
end;

function TReport.StartPageY:integer;
var R:integer;
begin
  R:=1;
  if (CurPage=1) and (DocHeaderBand<>nil) then
     Inc(R,DocHeaderBand^.GetSizeY);
  if PageHeaderBand<>nil then Inc(R,PageHeaderBand^.GetSizeY);
  StartPageY:=R;
end;

function TReport.EndPageY:integer;
var R:integer;
begin
  R:=PageHight;
  if PageFooterBand<>nil then Dec(R,PageFooterBand^.GetSizeY);
{  if (Report^.DocFooterBand)}
  EndPageY:=R;
end;

procedure TReport.Stop(AMsg:string);
var P:PString;
begin
  isError:=true;
  P:=@AMsg;
  DBError(edbfParserError, P);
end;

procedure TReport.Print;
var i:integer;
    L:longint;
procedure SetPrinterOptions;
begin
{  with PrinterOptionsRecord,PrinterManager^ do
  begin
     if UsePrinterCodes<>0 then
     begin
       SetCondensed(_Rec,CheckBoxes1 and $01 <>0);
       SetNLQMode(_Rec,CheckBoxes1 and $02 <>0);
       if Set216Inch<>0 then Set216InchLF(_Rec, Set216Inch);
       if LeftMargin<>0 then SetLeftMargin(_Rec, LeftMargin);
       if RightMargin<>0 then SetRightMargin(_Rec, RightMargin);
     end;
  end;}
end;

procedure _Print;
begin
  if IsError then exit;
  if DocHeaderBand<>nil then DocHeaderBand^.Print;
  if IsError then exit;
  if PageHeaderBand<>nil then PageHeaderBand^.Print;
  if IsError then exit;
  OnPrintBand;
  if IsError then exit;
  if MainBand<>nil then
  begin
     if MainBand^.Header<>nil then MainBand^.Header^.Print;
     if IsError then exit;
     MainBand^.LineNo:=1;
     if (MainBand^.DataSourse = nil)
     or (not MainBand^.DataSourse^.Active) then
        MainBand^.Print
     else
     begin
       L:=MainBand^.DataSourse^.RecNo;
       MainBand^.DataSourse^.FirstRec;
       while not MainBand^.DataSourse^.Eof do
       begin
         MainBand^.Print;
         if IsError then exit;
         MainBand^.DataSourse^.NextRec;
         inc(MainBand^.LineNo);
       end;
       MainBand^.DataSourse^.Go(L);
     end;
    if MainBand^.Footer<>nil then MainBand^.Footer^.Print;
    if IsError then exit;
  end;
  if CurCollumn>0 then CurentLine:=MaxLineInPage+1;
  if PageFooterBand<>nil then PageFooterBand^.Print;
  if IsError then exit;
  if DocFooterBand<>nil then DocFooterBand^.Print;
  if IsError then exit;
  if LastPrintLine<>'' then
  begin
     WriteStr(LastPrintLine, nil);
     LastPrintLine:='';
  end;
end;

begin
  if PrinterManager=nil then
  begin
    SingleError(edbfPrinterInit);
    exit;
  end;
  PrintLines:=0;
  CurCollumn:=0;
  CurentPass:=1;
  if ColumnCount<1 then ColumnCount:=1;
  Strings:=New(PPageStrCollection, Init(100,50));
  Strings^.Duplicates:=true;
  Clear;
  Message(Owner, evPrint, cmBeforePrint, @Self);
  Printing:=true;
  CurentLine:=0;
  CurPage:=1;
  if ShowProgres then
  begin
    StatusForm:=Desktop^.Insert(GetPrintForm(TGetFormType(DoublePass)));
    UpdateStatus;
  end;
  _Rec:=PrinterManager^.OpenPrinter(ComponentName^);
  SetPrinterOptions;
  Owner^.Lock;
  _Print;
  if LastPrintLine<>'' then
  begin
     WriteStr(LastPrintLine, nil);
     LastPrintLine:='';
  end;
  OutToPrn;
  Owner^.UnLock;
  IsError:=false;
  PrinterManager^.EndPage(_Rec);
  PrinterManager^.ClosePrinter(_Rec);
  Printing:=false;
  CurentLine:=0;
  CurPage:=1;
  Dispose(Strings, Done);
  Strings:=nil;
  Message(Owner, evPrint, cmAfterPrint, @Self);
  GroupBandCollection^.DeleteAll;
  if StatusForm<>nil then
  begin
    StatusForm^.Free;
    StatusForm:=nil;
  end;
  Owner^.ReDraw;
end;

function TReport.Str(I:integer):string;
var j:integer;
    T:TPageItem;
    S,S1:string;
    P:PPageItem;
begin
  T.X:=0;
  T.Y:=I;
  S[0]:=#0;
  FillChar(S[1],255,' ');
  if Strings^.Search(@T,J) then
  begin
    P:=Strings^.At(j);
    while (P^.Y=I) and (J<Strings^.Count) do
    begin
      S1:=DinamicStr(P^.Text);
      byte(S[0]):=Max(Length(S), P^.X+Length(S1));
      Move(S1[1], S[P^.X+1], Min(255-P^.X,Length(S1)));
      inc(j);
      if (J<Strings^.Count) then P:=Strings^.At(j);
    end;
  end;
  Str:=TrimRight(S);
end;

procedure TReport.WriteStr(S:string; Band:PReportBand);
begin
  if Band=nil then WriteStrXY(LeftMargin,CurentLine, S)
  else
  if Band^.RBType in [rbtDocHeader, rbtPageHeader, rbtDocFooter,rbtPageFooter] then
     WriteStrXY(LeftMargin,CurentLine, S)
  else
    WriteStrXY(LeftMargin+CurCollumn*((PageWidth div ColumnCount)+ColumnMargin),
             CurentLine, S);
  inc(CurentLine);
  if (CurentLine>=EndPageY) and not (Band^.RBType in
            [rbtDocHeader, rbtPageHeader, rbtDocFooter,rbtPageFooter]) then
  begin
    inc(CurCollumn);
    if CurCollumn<ColumnCount then CurentLine:=StartPageY;
  end;
  PrintLines:=Max(PrintLines, CurentLine);
  MaxLineInPage:=Max(MaxLineInPage, CurentLine);
end;

procedure TReport.WriteStrXY(X,Y:integer; S:string);
begin
  Strings^.Insert(New(PPageItem, Init(X,Y,S)));
end;

procedure TReport.OutToPrn;
var I:integer;
begin
  for i:=1 to PageHight do
    PrinterManager^.Writeln(_Rec,Str(I));
end;

procedure TReport.Clear;
begin
  Strings^.FreeAll;
end;

procedure TReport.NewPage;
begin
  if (CurPage=1) and (CurentLine=0) then exit;
  if CurCollumn>1 then CurentLine:=MaxLineInPage;
  if PageFooterBand<>nil then PageFooterBand^.Print;
  if LastPrintLine<>'' then
  begin
     WriteStr(LastPrintLine, nil);
     LastPrintLine:='';
  end;
  OutToPrn;
  CurentLine:=0;
  PrinterManager^.EndPage(_Rec);
  PrinterManager^.NewPage(_Rec);
  Clear;
  Inc(CurPage);
  UpdateStatus;
  CurCollumn:=0;
  CurentLine:=0;
  MaxLineInPage:=0;
  if PageHeaderBand<>nil then PageHeaderBand^.Print;
end;

procedure TReport.UpdateStatus;
var Rec:TPersentRec;
begin
  if StatusForm=nil then exit;
  Rec.MaxValue:=CurPage;
  Rec.CurValue:=CurentPass;
  StatusForm^.SetData(Rec)
end;

procedure TReport.OnPrintBand;
var i:integer;
begin
  for i:=0 to GroupBandCollection^.Count-1 do
      PGroupBand(GroupBandCollection^.At(i))^.CheckPrint;
end;

procedure TReport.OnPrintLine(Lines:integer);
begin
  if CurentLine>EndPageY then NewPage;
end;

function TReport.CurentPage:integer;
begin
  CurentPage:=CurPage;
end;

procedure TReport.PrinterSetup;
begin
  if Assigned(OnPrinterSetupEvent) then OnPrinterSetupEvent(@Self);
end;

procedure TReport.HandleEvent(var Event:TEvent);
procedure DestroyLinks;
begin
  if Event.InfoPtr=DocHeaderBand then DocHeaderBand:=nil;
  if Event.InfoPtr=PageHeaderBand then PageHeaderBand:=nil;
  if Event.InfoPtr=PageFooterBand then PageFooterBand:=nil;
  if Event.InfoPtr=DocFooterBand then DocFooterBand:=nil;
  if Event.InfoPtr=MainBand then MainBand:=nil;
end;
begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
    case Event.Command of
      cmPrintReport:Print;
      cmPreviewReport:PreViewReport;
      cmSaveReport:SaveReport;
      cmFindReport:ClearEvent(Event);
      cmPrinterSetup:PrinterSetup;
    else
      exit;
    end;
    evBroadcast:
    case Event.Command of
      cmDestroed:DestroyLinks;
    else
      exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TReport.PreViewReport;
var P:PPrinterRec;
    D:PDialog;
    TempD:word;
begin
  if Application<>nil then
  begin
{     TempD:=PrinterOptionsRecord.UsePrinterCodes;
     PrinterOptionsRecord.UsePrinterCodes:=0;}
     Print;
{     PrinterOptionsRecord.UsePrinterCodes:=TempD;}
     P:=PrinterManager^.At(PrinterManager^.Count-1);
     PrinterManager^.Delete(P);
     D:=New(PPreviewDialog, Init(P));
     D^.HelpCtx:=HelpCtx;
     case Application^.ExecuteDialog(D,nil) of
       cmOk:PrinterManager^.Insert(P);
       cmYes:begin
               SaveReport;
               PrinterManager^.FreeItem(P);
               _Rec:=nil;
             end;
       cmCancel:begin
                  PrinterManager^.FreeItem(P);
                  _Rec:=nil;
                end;
     end;
  end;
end;

procedure TReport.SaveReport;
var S:TDosStream;
    S1:PDosStream;
    FName:FNameStr;
    L:longint;
begin
  if _Rec=nil then exit;
  FName:='*.txt';
{  if ErrorMessage(mbpSaveAsReport,@FName)<>cmCancel then
  begin
    if Rec^.S<>nil then
    begin
      S1:=Rec^.S;
      L:=S1^.GetPos;
      S1^.Seek(0);
    end
    else S1:=New(PDosStream, Init(Rec^.FileName, stOpenRead));
    S.Init(FName, stCreate);
    S.CopyFrom(S1^,S1^.GetSize);
    S.Done;
    if Rec^.S<>nil then Rec^.S^.Seek(L)
    else S1^.Free;
  end;}
end;

procedure TReport.SetReportBand(ARB:PReportBand; ARBType:TReportBandType);
begin
  if ARBType = rbtDetail then ARBType:=rbtMain;
  case ARBType of
    rbtDocHeader:DocHeaderBand:=ARB;
    rbtPageHeader:PageHeaderBand:=ARB;
    rbtPageFooter:PageFooterBand:=ARB;
    rbtDocFooter:DocFooterBand:=ARB;
    rbtMain:MainBand:=ARB;
  end;
  if ARB<>nil then
  begin
    ARB^.Report:=@Self;
    ARB^.RBType:=ARBType;
    if ARBType=rbtMain then ARB^.SetDetail(ARB^.Detail, rbtDetail);
    ARB^.SetDetail(ARB^.Header, rbtBandHeader);
    ARB^.SetDetail(ARB^.Footer, rbtBandFooter);
  end;
end;

procedure TReport.Awaken;
begin
  SetReportBand(DocHeaderBand,rbtDocHeader);
  SetReportBand(PageHeaderBand,rbtPageHeader);
  SetReportBand(PageFooterBand,rbtPageFooter);
  SetReportBand(DocFooterBand,rbtDocFooter);
  SetReportBand(MainBand,rbtMain);
  Message(Owner, evBroadcast, cmAfterLoadReport, @Self);
end;


   { TReportLabel }
constructor TReportLabel.Init(var Bounds: TRect; AReport:PReport;
                                  ARLStyle:TRLStyle; AParam:String);
begin
  inherited Init( Bounds);
  RLStyle:=ARLStyle;
  Report:=AReport;
  ParamStr:=NewStr(AParam);
  ComponentType:=ctReportLabel;
  EventMask:=EventMask or evBroadcast;
end;

destructor TReportLabel.Done;
begin
  DisposeStr(ParamStr);
  inherited Done;
end;

constructor TReportLabel.Load(var S:TStream);
begin
  inherited Load(S);
  S.Read(RLStyle, SizeOf(TRLStyle));
  ParamStr:=S.ReadStr;
  PString(Report):=S.ReadStr;
end;

procedure TReportLabel.Store(var S:TStream);
var P:PString;
begin
  inherited Store(S);
  S.Write(RLStyle, SizeOf(TRLStyle));
  S.WriteStr(ParamStr);
  P:=nil;
  if Report=nil then S.WriteStr(P)
  else S.WriteStr(Report^.ComponentName);
end;

procedure TReportLabel.Draw;
var
  B: TDrawBuffer;
  S,S1:string;
  P:PString;
  CurLine:Longint;
begin
  MoveChar(B[0], ' ', GetColor(1), Size.X);
  P:=@S1;
  case RLStyle of
    rlsDate:
           begin
              S1:=DateToStr(Date);
              FormatStr(S,ParamStr^,P)
           end;
    rlsTime:
           begin
              S1:=TimeToStr(Time);
              FormatStr(S,ParamStr^,P)
           end;
    rlsDateTime:
           begin
              S1:=DateToStr(Date)+' - '+TimeToStr(Time);
              FormatStr(S,ParamStr^,P)
           end;
    rlsPageNo:
           begin
              if Report<>nil then Longint(P):=Report^.CurentPage
              else Longint(P):=0;
              FormatStr(S,ParamStr^,P)
           end;
    rlsLineNo:begin
                if (Owner<>nil) and (TypeOf(Owner^)=TypeOf(TReportBand)) then
                   if Report^.Printing then CurLine:=PReportBand(Owner)^.LineNo
                   else CurLine:=1;
                FormatStr(S,ParamStr^,CurLine)
              end;
  else
    S:=ComponentName^;
  end;
  MoveStr(B,S,GetColor(1));
  WriteLine(0, 0, Size.X, Size.Y, B);
end;

procedure TReportLabel.SetParam(AParam:string);
begin
  AssignStr(ParamStr, AParam);
end;

procedure TReportLabel.HandleEvent(var Event:TEvent);
procedure AfterLoad;
var P:PString;
begin
  if Report=nil then exit;
  P:=PString(Report);
  if stUpCase(P^)<>stUpCase(PReport(Event.InfoPtr)^.ComponentName^) then exit;
  DisposeStr(P);
  Report:=Event.InfoPtr;
end;

procedure Update;
var Hour, Minute, Second, Sec100: Word;
begin
   GetTime(Hour, Minute, Second, Sec100);
   if LastState <> Second and 1 then
   begin
     DrawView;
     LastState:=Second and 1;
   end;
end;
begin
  inherited HandleEvent(Event);
  case Event.What of
    evBroadcast:
    case Event.Command of
      cmAfterLoadReport:AfterLoad;
      cmIdle:if RLStyle in [rlsDate,rlsTime,rlsDateTime] then Update;
      cmDestroed:if Event.InfoPtr=Report then Report:=nil;
    else
      exit;
    end;
  else
    exit;
  end;
end;

function StdGetPrintForm(GetFormType:TGetFormType):PView;
begin
  StdGetPrintForm:=nil;
end;

   { TTextView }
constructor TTextView.Init(var Bounds: TRect; AHScrollBar, AVScrollBar:PScrollBar;
                               S:PStream);
begin
  inherited Init(Bounds, AHScrollBar, AVScrollBar);
  if S=nil then fail;
  Strings:=New(PStrings, Init);
  Stream:=S;
end;

destructor TTextView.Done;
begin
  Dispose(Strings, Done);
  inherited Done;
end;

procedure TTextView.Draw;
var
  B: TDrawBuffer;
  C: Byte;
  I: Integer;
  S: String;
  P: PString;
begin
  C := GetColor(1);
  for I := 0 to Size.Y - 1 do
  begin
    MoveChar(B, ' ', C, Size.X);
    if Delta.Y + I < Strings^.Count then
    begin
      P :=Strings^.At(Delta.Y + I);
      if P = nil then
      begin
        S := '';
        MoveStr(B, S, C);
      end
      else FormatLine(B, P^, Size.X,C);
    end;
    WriteLine(0, I, Size.X, 1, B);
  end;
end;

procedure TTextView.ReadFile(StartPos, EndPos:Longint);
var S:string;
    MaxLen:integer;
begin
  Strings^.FreeAll;
  Stream^.Seek(StartPos);
  if Stream^.Status<>stOk then Stream^.Reset;
  MaxLen:=82;
  while (Stream^.GetPos<EndPos) and (Stream^.Status=stOk) do
  begin
     S:=ReadText(Stream^);
     if Length(S)>MaxLen then MaxLen:=Length(S)+2;
     Strings^.Insert(NewStr(S));
  end;
  SetLimit(MaxLen, Strings^.Count);
  if Stream^.Status<>stOk then Stream^.Reset;
  DrawView;
end;

procedure TTextView.SetState(AState: Word; Enable: Boolean);
begin
  TScroller.SetState(AState, Enable);
  if Enable and (AState and sfExposed <> 0) then
     SetLimit(Limit.X, Limit.Y);
end;

procedure TTextView.FormatLine(var DrawBuf; Line: string; Width: Integer; C: byte);
var
  S:string;
  Len,i:integer;
begin
  S:='';
  if Length(Line)>Delta.X+1+Width then Len:=Delta.X+1+Width
  else len:=Length(Line);
  for i:=Delta.X+1 to Len do
    if Line[i]=#9 then
       S:=S+DupChar(' ', i or 8 - i)
    else S:=S+Line[i];
  MoveStr(DrawBuf,S,C);
end;

   { TPreviewDialog }
constructor TPreviewDialog.Init(S:PPrinterRec);
var R:TRect;
    ScrollBar1,ScrollBar2:PScrollBar;
begin
   Desktop^.GetExtent(R);
   inherited Init(R,'Preview report');
   Options:=Options or ofCentered;
   PrinterRec:=S;
   PrinterRec^.S:=New(PBufStream,Init(PrinterRec^.FileName, stOpenRead, 1024));
   GetExtent(R);
   R.A.Y:=R.B.Y-1;
   Dec(R.B.X);
   inc(R.A.X,20);
     ScrollBar1:=New(PScrollBar, Init( R));
     ScrollBar1^.Options := ScrollBar1^.Options or ofPostProcess;
     Insert(ScrollBar1);
   GetExtent(R);
   R.A.X:=R.B.X-1;
   Dec(R.B.Y);
   Inc(R.A.Y);
     ScrollBar2:=New(PScrollBar, Init( R));
     ScrollBar2^.Options := ScrollBar2^.Options or ofPostProcess;
     Insert(ScrollBar2);
   GetExtent(R);
   R.Grow(-1,-1);
   inc(R.A.Y,2);
     TextView:=New(PTextView, Init( R, ScrollBar1, ScrollBar2, PrinterRec^.S));
     Insert(TextView);
   R.Assign(  1,  1, 10,  3);
     Insert(New(PButton, Init( R, '<<', cmFistPagePreView, bfNormal)));
   R.Assign( 10,  1, 19,  3);
     Insert(New(PButton, Init( R, '<', cmPrevPagePreView, bfNormal)));
   R.Assign( 19,  1, 28,  3);
     Insert(New(PButton, Init( R, '>', cmNextPagePreView, bfNormal)));
   R.Assign( 28,  1, 37,  3);
     Insert(New(PButton, Init( R, '>>', cmLastPagePreView, bfNormal)));
   R.Assign( 41,  1, 51,  3);
     Insert(New(PButton, Init( R, {$IFNDEF RUS}'~P~rint'{$ELSE}'~~'{$ENDIF}, cmOk, bfNormal)));
   R.Assign( 51,  1, 61,  3);
     Insert(New(PButton, Init( R, {$IFNDEF RUS}'~S~ave'{$ELSE}'~~࠭.'{$ENDIF}, cmYes, bfNormal)));
   R.Assign( 61,  1, 71,  3);
     Insert(New(PButton, Init( R, {$IFNDEF RUS}'~C~lose'{$ELSE}'~~'{$ENDIF}, cmCancel, bfNormal)));
   GetExtent(R);
   R.A.Y:=R.B.Y-1;
   inc(R.A.X,2);
   R.B.X:=R.A.X+18;
     InfoPane:=New(PStaticText, Init( R,'Page 0 of 0'));
     Insert(InfoPane);
  SelectNext(false);
  FistPagePreView;
  UpdateParam;
end;

destructor TPreviewDialog.Done;
begin
  PrinterRec^.S^.Free;
  PrinterRec^.S:=nil;
  inherited Done;
end;

procedure TPreviewDialog.FistPagePreView;
var L:longint;
begin
  CurPage:=PrinterRec^.PageBreak;
  if CurPage^.Next<>nil then L:=CurPage^.Next^.Pos else
     L:=PrinterRec^.S^.GetSize;
  TextView^.ReadFile( CurPage^.Pos, L);
  UpdateParam;
end;

procedure TPreviewDialog.PrevPagePreView;
var L:longint;
    P:PPageBreak;
begin
  if CurPage=PrinterRec^.PageBreak then exit;
  P:=PrinterRec^.PageBreak;
  while P^.Next<>CurPage do P:=P^.Next;
  CurPage:=P;
  if CurPage^.Next<>nil then L:=CurPage^.Next^.Pos else
     L:=PrinterRec^.S^.GetSize;
  TextView^.ReadFile( CurPage^.Pos, L);
  UpdateParam;
end;

procedure TPreviewDialog.NextPagePreView;
var L:longint;
begin
  if CurPage^.Next=nil then exit;
  CurPage:=CurPage^.Next;
  if CurPage^.Next<>nil then L:=CurPage^.Next^.Pos else
     L:=PrinterRec^.S^.GetSize;
  TextView^.ReadFile( CurPage^.Pos, L);
  UpdateParam;
end;

procedure TPreviewDialog.LastPagePreView;
var L:longint;
begin
  while CurPage^.Next<>nil do CurPage:=CurPage^.Next;
  L:=PrinterRec^.S^.GetSize;
  TextView^.ReadFile( CurPage^.Pos, L);
  UpdateParam;
end;

procedure TPreviewDialog.HandleEvent(var Event:TEvent);
begin
  inherited HandleEvent(Event);
  case Event.What of
  evCommand:
    case Event.Command of
      cmFistPagePreView:FistPagePreView;
      cmPrevPagePreView:PrevPagePreView;
      cmNextPagePreView:NextPagePreView;
      cmLastPagePreView:LastPagePreView;
    else
      exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TPreviewDialog.UpdateParam;
var A:array[1..2] of longint;
    P:PPageBreak;
    S:string;
begin
  a[2]:=0;
  P:=PrinterRec^.PageBreak;
  while P<>nil do
  begin
    inc(a[2]);
    if P=CurPage then a[1]:=a[2];
    P:=P^.Next;
  end;
  FormatStr(S,'Page %d of %d',A);
  AssignStr(InfoPane^.Text, S);
  InfoPane^.DrawView;
end;

   { TReportMemo }
constructor TReportMemo.Init(Bounds:TRect; ADS:PDataSourse; Name:String);
begin
  inherited Init(Bounds);
  ComponentType:=ctReportMemo;
  EventMask:=EventMask or evMessage;
  DS:=ADS;
  if DS<>nil then
  begin
    Field:=DS^.FieldByName(Name);
    FieldName:=NewStr(Name);
  end;
end;

destructor TReportMemo.Done;
begin
  inherited Done;
  DisposeStr(FieldName);
  if TextBuf<>nil then FreeMem(TextBuf, TextBuf^.Length);
end;

constructor TReportMemo.Load(var S:TStream);
begin
  Inherited Load(S);
  FieldName:=S.ReadStr;
  PString(DS):=S.ReadStr;
  S.Read(AutoSize, SizeOf(Boolean));
end;

procedure TReportMemo.Store(var S:TStream);
var P:PString;
begin
  inherited Store(S);
  S.WriteStr(FieldName);
  if DS=nil then P:=nil
  else P:=DS^.TableName;
  S.WriteStr(P);
  S.Write(AutoSize, SizeOf(Boolean));
end;
const
  Separators = [' ','.',',','-'];

function TReportMemo.GetVertSize:integer;
function CalcVSize:integer;
var
  I, J, L, P, Y: Integer;
begin
  L := TextBuf^.Length-1;
  P := 0;
  Y := 0;
  while P < L do
  begin
    if P <= L then
    begin
      I := P;
      repeat
        J := P;
        while (P <= L) and (TextBuf^.Text[P] = ' ') do Inc(P);
        while (P <= L) and (TextBuf^.Text[P] <> ' ') and (TextBuf^.Text[P] <> #13) do Inc(P);
      until (P > L) or (P >= I + Size.X) or (TextBuf^.Text[P] = #13);
      if P > I + Size.X then
        if J > I then P := J else P := I + Size.X;
      J := 0;
      while (P <= L) and (TextBuf^.Text[P] = ' ') do Inc(P);
      if (P <= L) and (TextBuf^.Text[P] = #13) then
      begin
        Inc(P);
        if (P <= L) and (TextBuf^.Text[P] = #10) then Inc(P);
      end;
    end;
    Inc(Y);
  end;
  CalcVSize:=Y;
end;
begin
  GetVertSize:=max(CalcVSize,{CountChar(TextBuf^.Text, TextBuf^.Length, #13),} Size.Y);
end;

procedure TReportMemo.Draw;
var
  B: TDrawBuffer;
  C: Byte;
  I,J,L,P,Y: Integer;

procedure DoDraw;
var
  I, J, L, P, Y: Integer;
begin
  L := TextBuf^.Length-1;
  P := 0;
  Y := 0;
  while Y < Size.Y do
  begin
    MoveChar(B, ' ', C, Size.X);
    if P <= L then
    begin
      I := P;
      repeat
        J := P;
        while (P <= L) and (TextBuf^.Text[P] = ' ') do Inc(P);
        while (P <= L) and (TextBuf^.Text[P] <> ' ') and (TextBuf^.Text[P] <> #13) do Inc(P);
      until (P > L) or (P >= I + Size.X) or (TextBuf^.Text[P] = #13);
      if P > I + Size.X then
        if J > I then P := J else P := I + Size.X;
      J := 0;
      MoveBuf(B[J], TextBuf^.Text[I], C, P - I);
      while (P <= L) and (TextBuf^.Text[P] = ' ') do Inc(P);
      if (P <= L) and (TextBuf^.Text[P] = #13) then
      begin
        Inc(P);
        if (P <= L) and (TextBuf^.Text[P] = #10) then Inc(P);
      end;
    end;
    WriteLine(0, Y, Size.X, 1, B);
    Inc(Y);
  end;
end;

begin
  C:=GetColor(1);
  if (DS=nil) or (Field=nil) then
  begin
    MoveChar(B, ' ', C, Size.X);
    if ComponentName<>nil then
    begin
      MoveChar(B, ' ', C, Size.X);
      P:=(Size.Y div 2);
      if P>0 then WriteLine(0, 0, Size.X, P, B);
      i:=(Size.X-Length(ComponentName^)) div 2;
      if i<0 then i:=0;
      MoveStr(B[i], ComponentName^, C);
      WriteLine(0, P, Size.X, P+1, B);
      inc(P);
      if P<Size.Y then
      begin
        MoveChar(B, ' ', C, Size.X);
        WriteLine(0, P, Size.X, Size.Y, B);
      end;
    end
    else
    begin
      MoveChar(B, ' ', C, Size.X);
      WriteLine(0, 0, Size.X, Size.Y, B);
    end;
  end
  else
  begin
    if TextBuf=nil then
    begin
      MoveChar(B, ' ', C, Size.X);
      WriteLine(0, 0, Size.X, Size.Y, B);
    end
    else  DoDraw;
  end;
end;

function TReportMemo.GetPalette: PPalette;
const
  P: String[Length(CStaticText)] = CStaticText;
begin
  GetPalette := @P;
end;

procedure TReportMemo.Update;
var
  Len:Word;
begin
  if Field=nil then exit;
  Len:=Field^.GetSize;
  if TextBuf<>nil then FreeMem(TextBuf, TextBuf^.Length);
  if Len=0 then TextBuf:=nil
  else
  begin
    GetMem(TextBuf, Len+SizeOf(Word));
    TextBuf^.Length:=Len;
    Field^.GetData(TextBuf^);
  end;
  DrawView;
end;

procedure TReportMemo.HandleEvent(var Event:TEvent);
procedure AfterLoad;
var P:PString;
begin
  P:=PString(DS);
  if stUpCase(P^)<>stUpCase(PDataSourse(Event.InfoPtr)^.TableName^) then exit;
  DisposeStr(P);
  DS:=Event.InfoPtr;
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evPrint:
    case Event.Command of
      cmBeforePrinBand:begin
                         SaveSize.X:=Size.X;
                         SaveSize.Y:=Size.Y;
                         GrowTo(Size.X,GetVertSize);
                       end;
      cmAfterPrintBand:GrowTo(SaveSize.X,SaveSize.Y);
    else
      exit;
    end;
    evBroadcast:
     if (DS<>nil) and (Event.Command=cmAfterLoadDS) then
        AfterLoad
     else
     if (DS<>nil) and (Event.InfoPtr=DS) then
     begin
       case Event.Command of
         cmUpDataDS,
         cmAfterPost: Update;
         cmNewTable: if FieldName<>nil then
                     begin
                        Field:=DS^.FieldByName(FieldName^);
                        Update;
                     end;
         cmDestroed:begin
                      DS:=nil;
                      Field:=nil;
                      DrawView;
                    end;
       else
         Exit;
       end;
       DrawView;
     end;
  else
    exit;
  end;
end;

    { TReportShape }

constructor TReportShape.Init(Bounds:TRect; AShape:TRShapeType; APenStyle:TPenStyle);
begin
  inherited Init(Bounds);
  Shape:=AShape;
  PenStyle:=APenStyle;
  ComponentType:=ctReportShape;
end;

constructor TReportShape.Load(var S:TStream);
begin
  inherited Load(S);
  S.Read(Shape, SizeOf(TRShapeType));
  S.Read(PenStyle, SizeOf(TPenStyle));
end;

procedure TReportShape.Store(var S:TStream);
begin
  inherited Store(S);
  S.Write(Shape, SizeOf(TRShapeType));
  S.Write(PenStyle, SizeOf(TPenStyle));
end;

procedure TReportShape.Draw;
var
  C: Byte;
  I,J,L,P,Y: Integer;
  B:TDrawBuffer;
  S:string;
procedure DrawRectangle;
var i:integer;
begin
  for i:=0 to Size.X-1 do
    MoveChar(B[i], GetChar(i,0), GetColor(1), 1);
  WriteLine(0, 0, Size.X, 1, B);
  MoveChar(B[1], ' ', GetColor(1), Size.X-2);
  for i:=1 to Size.Y-1 do
  begin
    MoveChar(B[0], GetChar(0,i), GetColor(1), 1);
    MoveChar(B[Size.X-1], GetChar(Size.X-1,i), GetColor(1), 1);
    WriteLine(0, i, Size.X, i+1, B);
  end;
  for i:=0 to Size.X-1 do
    MoveChar(B[i], GetChar(i,Size.Y-1), GetColor(1), 1);
  WriteLine(0, Size.Y-1, Size.X, Size.Y, B);
end;

procedure DrawVertLine;
var i:integer;
begin
  for i:=0 to Size.Y-1 do
    MoveChar(B[i], GetChar(0,i), GetColor(1), 1);
  WriteLine(0, 0, 1, Size.Y, B);
end;
procedure DrawHorLine;
var i:integer;
begin
  for i:=0 to Size.X-1 do
    MoveChar(B[i], GetChar(i,0), GetColor(1), 1);
  WriteLine(0, 0, Size.X, 1, B);
end;
procedure DrawTopAndBottom;
var i:integer;
begin
  for i:=0 to Size.X-1 do
     WriteChar(i,0, GetChar(i,0), 1, 1);
  MoveChar(B[0], ' ', GetColor(1), Size.X);
  WriteLine(0, 1, Size.X, Size.Y-1, B);
  for i:=0 to Size.X-1 do
     WriteChar(i,Size.Y-1, GetChar(i,Size.Y-1), 1, 1);
end;
procedure DrawRightAndLeft;
var i:integer;
begin
  MoveChar(B[1], ' ', GetColor(1), Size.X-2);
  for i:=0 to Size.Y-1 do
  begin
    MoveChar(B[0], GetChar(0,i), GetColor(1), 1);
    MoveChar(B[Size.X-1], GetChar(Size.X-1,i), GetColor(1), 1);
    WriteLine(0, i, Size.X, i+1, B);
  end;
end;


begin
  C:=GetColor(1);
  if Size.X=1 then
      DrawVertLine
  else
  if Size.Y=1 then DrawHorLine
  else
  case Shape of
    rsRectangle:DrawRectangle;
    rsTopAndBottom:DrawTopAndBottom;
    rsRightAndLeft:DrawRightAndLeft;
  end;
end;

function TReportShape.GetChar(X,Y:integer):char;
const SingleChar = 'ĳĿ³';
      DoubleChar = 'ɻȼ̹';
var SS:string;
    C:integer;
begin
{  case PenStyle of
    psSingleLine:}SS:=SingleChar;
{    psDoubleLine:SS:=DoubleChar;
  end};
  c:=ScanFrame(X,Y);
  GetChar:=SS[C];
end;

function TReportShape.ScanFrame(AX,AY:integer):byte;
var
{$IFDEF VER70}
    Result:byte;
{$ENDIF}
    P:PView;
    CC:byte;
    BX,BY:integer;
begin
  Result:=0;
  CC:=byte(PenStyle) shl 4 + 1;
  if ((AY = 0) or (AY = (Size.Y-1))) and (Size.X>1)
    and (Shape<>rsRightAndLeft) then
    if AX=0 then Result:=Result or 1 else
    if AX=Size.X-1 then Result:=Result or 4 else Result:=Result or 5;
  if ((AX = 0) or (AX=Size.X-1)) and (Size.Y>1)
     and (Shape<>rsTopAndBottom) then
    if AY=0 then Result:=Result or 2 else
    if AY=Size.Y-1 then Result:=Result or 8 else Result:=Result or 10;
  BX:=AX; BY:=AY;
  AX:=Origin.X+AX; AY:=Origin.Y+AY;
  if @Self<>Owner^.Last then
  begin
    P:=Next;
    while (P<>Owner^.Last) do
    begin
      if (P^.ComponentType=ctReportShape) then
        if (AY>=P^.Origin.Y) and (AY<P^.Origin.Y+P^.Size.Y) and
           (AX>=P^.Origin.X) and (AX<P^.Origin.X+P^.Size.X) then
        begin
          if ((P^.Origin.X=AX) or (P^.Origin.X+P^.Size.X-1=AX) or
              (P^.Origin.Y=AY) or (P^.Origin.Y+P^.Size.Y-1=AY)) then
          begin
            Result:=Result or PReportShape(P)^.ScanFrame(AX-P^.Origin.X,AY-P^.Origin.Y);
            if (BY>0) and (BY<Size.Y-1) and (Size.X>1) then
            begin
               if (BX=0) then Result:=Result and $FE
               else
               if (BX=Size.X-1) then Result:=Result and $FB
            end
            else
            begin
            if (BX>0) and (BX<Size.X-1) and (Size.Y>1) then
               if (BY=0) then Result:=Result and $FD
               else
               if (BY=Size.Y-1) then Result:=Result and $F7;
            end;
          end;
          break;
        end;

      P:=P^.Next;
    end;
  end;
{$IFDEF VER70}
  ScanFrame:=Result;
{$ENDIF}
end;

  { TVirtualDataSet }
constructor TVirtualDataSet.Init(ARecordCount:Longint);
begin
  inherited Init;
  ComponentType:=ctVirtualDataSet;
  fRecordCount:=ARecordCount;
end;

function TVirtualDataSet.EOF:boolean;
begin
  Eof:=fRecNo>fRecordCount;
end;
function TVirtualDataSet.RecNo:LongInt;
begin
end;
procedure TVirtualDataSet.Go( APos:Longint);
begin
  fRecNo:=APos;
end;

function TVirtualDataSet.NextRecEx:LongInt;
begin
  if fRecNo<=fRecordCount then Inc(fRecNo);
end;

function TVirtualDataSet.FirstRecEx:LongInt;
begin
  Go(1);
end;

constructor TVirtualDataSet.Load(var S:TStream);
begin
  inherited Load(S);
  S.Read(fRecNo, SizeOf(Longint));
  S.Read(fRecordCount, SizeOf(Longint));
end;

procedure TVirtualDataSet.Store(var S:TStream);
begin
  inherited Store(S);
  S.Write(fRecNo, SizeOf(Longint));
  S.Write(fRecordCount, SizeOf(Longint));
end;

function TVirtualDataSet.RecordCount:Longint;
begin
  RecordCount:=fRecordCount;
end;

function TVirtualDataSet.FieldDefs:PCollection;
begin
  FieldDefs:=nil;
end;

  { TUserDataSet }
constructor TUserDataSet.Init;
begin
  inherited Init;
  ComponentType:=ctUserDataSet;
end;

function TUserDataSet.EOF:boolean;
begin
end;
function TUserDataSet.RecNo:LongInt;
begin
end;
procedure TUserDataSet.Go( APos:Longint);
begin
end;
function TUserDataSet.NextRecEx:LongInt;
begin
end;
function TUserDataSet.FirstRecEx:LongInt;
begin
end;
function TUserDataSet.RecordCount:Longint;
begin
end;
constructor TUserDataSet.Load(var S:TStream);
begin
  inherited Load(S);
end;
procedure TUserDataSet.Store(var S:TStream);
begin
  inherited Store(S);
end;
function TUserDataSet.FieldDefs:PCollection;
begin
  FieldDefs:=nil;
end;


procedure RegisterReports;
begin
  RegisterType(RReportBandBackground);
  RegisterType(RReportBand);
  RegisterType(RReport);
  RegisterType(RReportLabel);
  RegisterType(RReportMemo);
  RegisterType(RReportShape);
  RegisterType(RVirtualDataSet);
  RegisterType(RUserDataSet);
  RegisterType(RGroupBand);
end;

end.
