{*******************************************************}
{                                                       }
{       Turbo Pascal 7                                  }
{       Free Pascal                                     }
{       Delphi 2 - 5                                    }
{       Virtual Pascal 2.1                              }
{                                                       }
{       Data access unit                                }
{                                                       }
{       Copyright (c) 1995-2001 by Alexs                }
{                                                       }
{*******************************************************}

{$IFDEF VER70}
{$O-}
{$ENDIF}
{$I Define.inc}

unit DAE;

interface

uses
{$IFDEF WIN32} Windows,{$ENDIF}
  Objects,
{$IFDEF UseDrivers}
  Drivers,
{$ENDIF}
  Ext_Dos, Ext_Type, Dos, Clases, Ext_Sys, VMM, DAE_Type, IniFile;

{$I DBConsts.inc}

type
   TTableType = (ttDefault, ttParadox, ttDBase, ttASCII);

   TIndexOptions = set of (ixPrimary, ixUnique, ixDescending,
                           ixCaseInsensitive, ixExpression);

   TSearchOptions = set of (soCaseInSensitive, soRightTrim,
                            soInclude, soGoFirst);
type
    TFieldType = (ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
                  ftBoolean, ftFloat,  ftCurrency, ftBCD,     ftDate,
                  ftTime, ftDateTime, ftBytes, ftVarBytes, ftAutoInc,
                  ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftParadoxOle,
                  ftDBaseOle, ftTypedBinary);

    TIndexStatus = set of (iERead, iEWRite, iOpenErr, iDupKey, iEofFind,
                    iBofFind, iKeyNotFound);
type
    TFieldInfoRecord = record
      Name:string;
      Size:integer;
      Dec:integer;
      Offset:word;
      fType:TFieldType;
      Align:TAlign;
    end;

type
   TErrorCode = (
       edbNoError, edbFieldNotString, edbFieldNotInteger,
       edbFieldNotBoolean, edbFieldNotDateTime,
       edbFieldNotCurrency, edbFieldNotFloat,
       edbFieldNotSetData,
       edbDSNotEditInsert, edbFieldNotFound, edbFileNotFound, edbFileNotOpen,
       edbFileNotCreate, edbSesionNotInit, edbFieldNonMemo, edbFileNotIndex,
       edbDupKey, edbStartRangeNotInstal, edbEndRangeNotInstal,
       edbDataSetNotClosed, edbDataSetNotLock, edbDataSetNotUnLock,
       edbSCalcFieldNA, edbGCalcFieldNA, edbfDriverNotAssigned,
       edbLockUpFieldType, edbIndexNotFound,
       edbFileNotIndexed, edbFileAlreadyOpen, edbIncompatableTable,
       edbErrorFreeBk, edbErrorPosBk, edbNonExclus, edbDSNotsetKey,
       edbRecAlreadyExist, edbDataSetReadOnly, edbBathCopyBase,
       edbTableNameRecvared, edbDSNotBrowse, edbIndexNotAssigned,
       edbIndexFieldTypeMismatch, edbFileCorrupt,
       edbNotValidOperation, edbInvalidIntegerValue, edbInvalidFloatValue,
       edbInvalidDateValue,
       edbIntValueOutOfRange, edbFValueOutOfRange, edbDAEInvalidValue,
       edbDataSetClosed, edbfInvalidFilter,

       {訡 SQL}
       edbSELECTExpected, edbFieldListExpected, edbFieldOrFROMExpected,
       edbFROMExpected, edbORDERBYExpected, edbIndexExpected,
       edbFilesListExpected, edbSQLParserInit,
       edbNotFindDS,
       {}
       edbExprError, edbParamNotFound, edbIndexFileMistMath,
       {訡 PXEngine}
       edbPXInitError,
       edbPXError,
       {㣨 訡}
       edbfPrinterInit, edbfParserError);
   TTextMessage = ( txtSQLWait, txtReportWait);

type
  TDataSetState = (dsInactive, dsBrowse, dsEdit, dsInsert, dsAppend,
                   dsEditRangeStart, dsEditRangeEnd, dsSetKey, dsCreate);
type
  TDataSetStatus =(dsOk, dsErrorReadFields, dsErrorReadFile, dsErrorOpenFile,
     dsErrorReadRecord, dsErrorSeek, dsErrorCreate, dsErrorWriteHeader,
     dsErrorMakeIndex, dsErrorGoTop, dsErrorRecordNotFind, dsNotFind,
     dsFindSet);
  TErrorCodeSet = set of TErrorCode;

type
  TDBExitProc=procedure;

type
   PDBStream = ^TDBStream;
   TDBStream = object(TDosStream)
     CountUse:integer;
     Info:Pointer;
     InfoSize:word;
     FileName:PString;
     constructor Init(AFileName:FNameStr; AMode:word);
     destructor Done;virtual;
     procedure CopyStructure(S:PStream);virtual;{$IFDEF EAM} abstract;{$ENDIF}
     function CreateInfoBlock(ASize:longint):pointer;
   end;


type
  PCodePageDrv = ^TCodePageDrv;
  PSession = ^TSession;

  PAliasRecord = ^TAliasRecord;
  TAliasRecord = object(TObject)
    WorkDir:PString;
    AliasName:PString;
    WaitTime:Word;
    isNet:boolean;
    DefaultDriver:PCodePageDrv;
    constructor Init(const AWorkDir, AAliasName:string);
    destructor Done;virtual;
    constructor Load(S:PIniFile; Item:integer);
    procedure Store(S:PIniFile; Item:integer);
    procedure SetFilesChange(aFilesChange:boolean);virtual;
    function FilesChange:boolean;virtual;
  private
    fFilesChange:boolean; {}
  end;

   PDatabase = ^TDatabase;
   TDatabase = object(TObject)
   end;

   PAliasCollection = ^TAliasCollection;
   TAliasCollection = object(TSortedCollection)
     function Compare(Key1, Key2: Pointer): Integer; virtual;
     function KeyOf(Item: Pointer): Pointer; virtual;
   end;

  TSession = object(TObject)
    AliasCollection:PAliasCollection;
    CodePageCollections:PCollection;
    DefaultDriver:PCodePageDrv;
    DataCollection:PCollection;
    DBStreams : PCollection;
    constructor Init;
    destructor Done;virtual;
    function DatabaseCount:integer;
    procedure AddPassword(const APassword:string);
    function ByAlias(const Name:String):DirStr;
    function AliasByName(const Name:String):PAliasRecord;
    function AddAlias(const AWorkDir, AAliasName:string):PAliasRecord;
    function GetDBStream(Name:string):PDBStream;
    procedure FreeDBStream(P:PDBStream);
    procedure StoreCfgFile(AFileName:FNameStr);
    procedure LoadCfgFile(AFileName:FNameStr);
    procedure LoadAliases(AFileName:FNameStr);
    function SystemPath:string;
    procedure SetSystemPath(APath:string);
    function GetCPDriver(Version:Cardinal):PCodePageDrv;
  private
    DefPassword:PString;
    fSystemPath:PString;
    procedure InitCodePages;
  end;

  TCodePageArray = array [0..255] of char;
  TCodePageDrv = object(TObject)
  private
    CP:TCodePageArray;
    RevertCP:TCodePageArray;
    fName:PString;
  public
    Version:Cardinal;
    SubVers:Cardinal;
    constructor Init;
    constructor Load(var S:TStream);
    destructor Done;virtual;
    function StrConvertTO(const S:string):string;
    function StrConvertFrom(const S:string):string;
    procedure BufConvertTO(var Buf; Len:word);
    procedure BufConvertFrom(var Buf; Len:word);
    function Name:string;
  end;



type
  PIndexStream = ^TIndexStream;
  PIndexObject = ^TIndexObject;
  PNode = ^TNode;
  TNode = object(TVMMPage)
    Stream:PIndexStream;
    constructor Init(AStream:PIndexStream; ARef:longint; ABufferSize:word);
    destructor Done;virtual;
    procedure Store;virtual;
    procedure Load(ARef:Longint);virtual;
  end;

  TIndexStream = object(TDBStream)
    Modified:boolean;
    destructor Done;virtual;
    function NewPage:longint;virtual;
    procedure FreePage(L:Longint);virtual;{$IFDEF EAM} abstract;{$ENDIF}
    procedure Flush;virtual;
    procedure WriteHeader;virtual;{$IFDEF EAM} abstract;{$ENDIF}
    {頥 ᫮ ᮢ   䠩}
    function IndexCount:integer;virtual;
    { ꥪ    ஬}
    function GetIndex(ANum:integer):PIndexObject;virtual;{$IFDEF EAM} abstract;{$ENDIF}
  end;

   PField = ^TField;
   PDataSet = ^TDataSet;
   PMemoFile = ^TMemoFile;
   PBookmark = ^TBookmark;

   TDataSetNotifyEvent = procedure(DataSet:PDataSet);
   TDataSetLangProc = procedure(var Data; Length:word; isWrite:boolean);
   TFilterEvent = function(Sender:PDataSet):boolean;
   TCreateMemoFileFunc = function(ADataSet:PDataSet):PMemoFile;

   TDataSet = object(TObject)
      Mode        : word;
      State     : TDataSetState;
      DSStatus    : TDataSetStatus;
      FieldDefs   : PCollection;
      IndexDefs   : PCollection;
      Index       : PIndexObject;
      Master      : PDataSet;
      MasterField : PField;
      CountRelation:integer;
      EndRangeSet : TByteSet;
      StartRangeSet :TByteSet;
      StartRange  : PByteArray;
      EndRange    : PByteArray;
      RecBuf      : PByteArray;
      CancelBuf   : PByteArray;
      RecNo       : longint;
      TableName   : PString;
      MasterFieldName:PString;
      CodePageDrv : PCodePageDrv;
      Filter      : PString;
      Filtred     : boolean;
      fEof        : Boolean;
      fBof        : Boolean;
      constructor Init;
      destructor Done;virtual;
      function Open:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Close;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      procedure CreateIndex(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure OpenIndex(AIndexName:FNameStr);virtual;{$IFDEF EAM}abstract;{$ENDIF}

      function RecordSize:word;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function FieldCount:integer;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function RecordCount:Longint;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      function FieldByName( Name:String):PField;
      function FindField( Name:String):PField;
      function Fields(N:integer):PField;
      procedure FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function GetNumberFieldByName(AName:string):integer;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      procedure PutFields(AName:byte;Value:string);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function GetFields(AName:byte):string;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      function GetMemoFile:PMemoFile;virtual;

      procedure SetKey;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure EditKey;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function GotoKey:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function FindKey:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function Location:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Post;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Edit;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Cancel;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Append;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Insert;virtual;
      procedure Delete;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure EmptyTable;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      procedure Go(ARecNo:longint);virtual;
      procedure GotoCurent(P:PDataSet);
      function EOF:boolean;
      function BOF:boolean;
      procedure Last;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure First;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Next;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure Prior;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure MoveBy(Distance:longint);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure FreeBookmark(Bookmark: PBookmark);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function GetBookmark:PBookmark;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure GotoBookmark(Bookmark: PBookmark);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      function isPosBookmark(Bookmark: PBookmark):boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure RenameTable(NewTableName: string);virtual;{$IFDEF EAM}abstract;{$ENDIF}

      function DeletedRec:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure UpData;

      procedure Lock;
      procedure UnLock;
      procedure FieldChanged(Field:PField);virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure SetMaster(AMaster:PDataSet; AField:string);
      procedure SetFiltred(F:boolean);virtual;
      procedure SetFilter(AFilter:string);virtual;
      function DoFilter:boolean;virtual;
      function ReadOnly:boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure ApplyUpdates;virtual;{$IFDEF EAM}abstract;{$ENDIF}
      procedure CancelUpdates;virtual;{$IFDEF EAM}abstract;{$ENDIF}

      procedure DoAfterInsert;
      procedure DoBeforeInsert;
      procedure DoBeforeEdit;
      procedure DoAfterEdit;
      procedure DoBeforePost;
      procedure DoAfterPost;
      procedure DoBeforeCancel;
      procedure DoAfterCancel;
      procedure DoBeforeDelete;
      procedure DoAfterDelete;
   private
      LookFlag:integer;
   public
      OnFilter:TFilterEvent;
      BeforeInsert: TDataSetNotifyEvent;
      AfterInsert : TDataSetNotifyEvent;
      BeforeEdit  : TDataSetNotifyEvent;
      AfterEdit   : TDataSetNotifyEvent;
      BeforePost  : TDataSetNotifyEvent;
      AfterPost   : TDataSetNotifyEvent;
      BeforeCancel: TDataSetNotifyEvent;
      AfterCancel : TDataSetNotifyEvent;
      BeforeDelete: TDataSetNotifyEvent;
      AfterDelete : TDataSetNotifyEvent;
   end;

   TIndexObject = object( TObject)
     Owner:PDataSet;
     IndexField:PField;
     IndexStream:PIndexStream;
     FileName:PString;
     iStatus:TIndexStatus;
     BufSize:word;
     PP:integer;
     PagePtrArray:TPagePtrArray;
     OldPagePtrArray:TPagePtrArray;
     OldPP:integer;
     Options:TIndexOptions;
     constructor Init(ALink:PDataSet);
     constructor Create(ALink:PDataSet; AOptions:TIndexOptions);
     destructor Done;virtual;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function IndexExpresion:string;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function KeySize:word;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function NextKey:Longint;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function PrevKey:Longint;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function Unique:Boolean;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure AddKey( AKey:String; ALink:LongInt);virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure DeleteKey( AKey:String; ALink:longint);virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure ClearKey;
     function FindCurent:Longint;virtual;
     function FirstKey:Longint;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function LastKey:Longint;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure DeleteCyrrent;virtual;
     procedure EmptyIndex; virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function GetNode(ARef:Longint):PNode;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure ReIndexs;
     function StartRange(const AVal:String):boolean;
     function EndRange(const AVal:String):boolean;
     function StartRangeValue:string;
     function EndRangeValue:string;
     procedure Close;
     function Open:Boolean;virtual;
     procedure RenameIndex;virtual;{$IFDEF EAM}abstract;{$ENDIF}
   end;

   TBookmark = record
     Owner:PDataSet;
     RecNo:Longint;
     fEof : Boolean;
     fBof : Boolean;
   end;


  PMemoFilesRecord = ^TMemoFilesRecord;
  TMemoFilesRecord = record
    ByteID:byte;
    CreateFunc:TCreateMemoFileFunc;
    SystemName:PString;
    Next:PMemoFilesRecord;
  end;

  TMemoFile = object(TObject)
     Owner:PDataSet;
     FileName:PString;
     constructor Init(AOwner:PDataSet);
     destructor Done;virtual;
     function GetSize(Offset:LongInt):Word;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure Get(Offset:LongInt;var D:TMemoRec);virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure AtFree(Offset:LongInt);virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function Put(var D:TMemoRec):LongInt;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure RenameMemo;virtual;
     procedure Close;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     procedure Open;virtual;{$IFDEF EAM}abstract;{$ENDIF}
     function CalcOffset(P:Pointer):Longint;virtual;
     procedure StoreOffset(P:Pointer; Value:Longint);virtual;
  end;


  TCalcProc=procedure(Sender:PField; DataSet:PDataSet; var Data);

   TField = object(TObject)
     Owner:PDataSet;
     FieldNo:int16;
     Name:PString;
     Size:word;
     Precision:byte;
     FieldType:TFieldType;
     PtrInBuf:word;
     Align:TAlign;
     Info:Pointer;
     isCurrency:boolean;
     ReadOnly:boolean;
     constructor Init(AOwner:PDataSet; AFieldNo:integer);
     destructor Done;virtual;
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     function AsDateTime:TDateTime;virtual;
     function AsCurrency:Currency;virtual;
     procedure SetAsString(D:string);virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsFloat(D:Double);virtual;
     procedure SetAsBoolean(D:boolean);virtual;
     procedure SetAsDateTime(D:TDateTime);virtual;
     procedure SetAsCurrency(D:Currency);virtual;
     function DisplayText:string;virtual;
     procedure Assign(AField:PField);virtual;
     procedure SetData(var Rec:TMemoRec);virtual;
     procedure GetData(var Rec:TMemoRec);virtual;
     function GetSize:word;virtual;
     function GetWidth:word;virtual;
     function IndexString:string;virtual;
     function Calculated:boolean;virtual;
     function IsIndexField: Boolean;virtual;
     function IsValidValue(S:String):boolean;virtual;
     procedure Clear;virtual;
     procedure UpdateField;virtual;
     function TestIndexField(P:PField):Boolean;virtual;
     procedure Load(S:PReader);virtual;
     procedure Store(S:PWriter);virtual;
     function IsNull:boolean;virtual;
     procedure OnNewRecord;virtual;
     procedure SetDataPtr;
     function DataPtr:Pointer;
{   private
     fDataPtr:Pointer;}
   end;

   PCustomMemoField = ^TCustomMemoField;
   TCustomMemoField = object(TField)
     function GetWidth:word;virtual;
     procedure AssignStrings(P:PStrings);
   end;

type
   PCalcField = ^TCalcField;
   TCalcField = object(TField)
     OnGet:TCalcProc;
     OnSet:TCalcProc;
     constructor Init(AOwner:PDataSet; AName:string; AType:TFieldType;
                      ASize:word; ADec:byte);
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     function AsDateTime:TDateTime;virtual;
     procedure SetAsString(D:string);virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsFloat(D:double);virtual;
     procedure SetAsBoolean(D:boolean);virtual;
     procedure SetAsDateTime(D:TDateTime);virtual;
     function IndexString:string;virtual;
     function Calculated:boolean;virtual;
     function IsNull:boolean;virtual;
     function IsValidValue(S:String):boolean;virtual;
     function GetWidth:word;virtual;
   end;

function FullName(Name:FNameStr):FNameStr;
function ExtractAliasName(Name:FNameStr):string;

type
   TDBErrorFunc = function( Code:TErrorCode;var Info):word;
   TDBTextMessageFunc = function( Code:TTextMessage):string;
   TDBOpenProc = function(AFileName:FNameStr):PDataSet;
function fTypeToChar(T:TFieldType):char;
function stdDBErrorproc( Code:TErrorCode;var Info):word;{$IFNDEF FPC}far;{$ENDIF}
function defDBErrorproc( Code:TErrorCode;var Info):word;{$IFNDEF FPC}far;{$ENDIF}
function StdErrorNames(Code:TErrorCode):string;
function defDBTextMessageFunc( Code:TTextMessage):string;{$IFNDEF FPC}far;{$ENDIF}
function stdDBTextMessageFunc( Code:TTextMessage):string;{$IFNDEF FPC}far;{$ENDIF}
function OpenDataSet(FileName:FNameStr):PDataSet;
function IndexOpen(FileName:FNameStr; DS:PDataSet; OnOpenTable:boolean):PIndexObject;
function IndexCreate( FileName:FNameStr; DS:PDataSet;
                      AField:string; AOptions:TIndexOptions):PIndexObject;
function GetPage(IO:PIndexObject; Ref:Longint):Pointer;
procedure FreePage(Node:PNode);
procedure CloseStream(Stream:PIndexStream);

type
  TIndexOpen=function(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet; OnOpenTable:boolean):PIndexObject;
  TIndexCreate=function(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                        AField:string; AOptions:TIndexOptions):PIndexObject;

const
  dbDesignTime   : boolean = false;
  LocateOptions  : TSearchOptions=[soCaseInSensitive, soGoFirst];
  TmpStream      : boolean = inMemory;
  DefIndexExt    : ExtStr = '.IND';
  ShowMemoValue  : boolean = false;
  ShowSysField   : boolean = false;
  DefMemoString  : string[10] = '<MEMO>';
  Session        : PSession = nil;
  CalcMessage    : string[20] = '塞.';

const
  MemoFilesCreatedList : PMemoFilesRecord = nil;

const
{$IFDEF COMPACT_EXE}
   DBError:TDBErrorFunc       = defDBErrorproc;
   DBText:TDBTextMessageFunc  = defDBTextMessageFunc;
{$ELSE}
   DBError:TDBErrorFunc       = StdDBErrorproc;
   DBText:TDBTextMessageFunc  = stdDBTextMessageFunc;
{$ENDIF}
   DBOpen:TDBOpenProc         = nil;

const
   OnOpenIndex:TIndexOpen=nil;
   OnCreateIndex:TIndexCreate=nil;

procedure SingleError(Code:TErrorCode);
procedure InitDataSets;
procedure DoneDataSets;
procedure DBInit;{$IFNDEF FPC}far;{$ENDIF}
procedure DBDone;{$IFNDEF FPC}far;{$ENDIF}
const
  DBExitProc  :TDBExitProc = DBDone;
  DBBeginProc :TDBExitProc = DBInit;
  DoReindex   : TProgresEvent = nil;
  HasError    : boolean = false;  {⠭  ⨭  砥 訡}

const
  NonCriticalError = [edbDataSetNotClosed,edbDataSetNotLock,
      edbDataSetNotUnLock, edbNonExclus, edbRecAlreadyExist,
      edbFileCorrupt,
      edbDAEInvalidValue, edbInvalidFloatValue, edbInvalidDateValue,
      edbInvalidIntegerValue, edbIntValueOutOfRange, edbFValueOutOfRange];

(*procedure CP1251(var Data; Length:word; isWrite:boolean);{$IFNDEF FPC}far;{$ENDIF}*)
procedure RegisterMemoFileType(ByteID:byte; AFunc:TCreateMemoFileFunc; ASystemName:string);
function DaeLibCfgFileName:FNameStr;
implementation
uses
  Ext_Math, Ext_Str;


function DaeLibCfgFileName:FNameStr;
var S:FNameStr;
begin
  if FileExists('DAE_LIB.CFG') then S:='DAE_LIB.CFG'
  else
  begin
    S:=GetEnv('DAE_CFG');
    if (S<>'') and (S[Length(s)]<>'\') then S:=S+'\';
    S:=S+'DAE_LIB.CFG';
  end;
  DaeLibCfgFileName:=S;
end;

  { TSession }

constructor TSession.Init;
begin
  inherited Init;
  Session:=@Self;
  DataCollection:=New(PCollection, Init(10,10));
  DBStreams:=New(PCollection, Init( 10 ,5));
  CodePageCollections:=New(PCollection, Init(10,5));
  LoadCfgFile(DaeLibCfgFileName);
  InitCodePages;
  LoadAliases(DaeLibCfgFileName);
  if AliasCollection=nil then
     AliasCollection:=New(PAliasCollection, Init(10,5));
end;

procedure TSession.InitCodePages;
var R:SearchRec;
    S:PStream;
    P:PCodePageDrv;
    Path:string;
begin
  if CodePageCollections=nil then exit;
  Path:=SystemPath;
  if (Path<>'') and (Path[Length(Path)]<>'\') then Path:=Path+'\';
  FindFirst( Path+'*.cpt', AnyFile, R);
  while DosError = 0 do
  begin
    S:=New(PDosStream, Init(Path+R.Name, stOpenRead));
    P:=New(PCodePageDrv, Load(S^));
    if P<>nil then CodePageCollections^.Insert(P);
    S^.Free;
    FindNext(R);
  end;
  FindClose(R);
end;

destructor TSession.Done;
var i:integer;
begin
  if DataCollection^.Count<>0 then
  begin
     for i:=0 to DataCollection^.Count-1 do
       DBError(edbDataSetNotClosed, PDataSet(DataCollection^.At(i))^.TableName);
  end;
  Dispose(DataCollection, Done);
  for i:=0 to DBStreams^.Count-1 do DBStreams^.FreeItem(DBStreams^.At(0));
  Dispose( DBStreams, Done);
  Dispose(AliasCollection, Done);
  AssignStr(fSystemPath, '');
  Dispose(CodePageCollections, Done);
  inherited Done;
  Session:=nil;
end;

function TSession.GetCPDriver(Version:Cardinal):PCodePageDrv;
function DoTest(Item:PCodePageDrv):boolean;far;
begin
  DoTest:=Item^.Version=Version;
end;
begin
  GetCPDriver:=CodePageCollections^.FirstThat(@DoTest);
end;


function TSession.DatabaseCount:integer;
begin
  DatabaseCount:=DataCollection^.Count;
end;

function TSession.GetDBStream(Name:string):PDBStream;
var
  P:PDbStream;
  I:integer;
function Test(DS:PDbStream):boolean;
begin
  Test:=Name = stUpCase(DS^.FileName^);
end;
begin
  Name:=FullName(Name);
  P:=nil;
  for i:=0 to DBStreams^.Count-1 do
    if Test(DBStreams^.At(i)) then
    begin
      P:=DBStreams^.At(i);
      break;
    end;
  if P<>nil then inc(P^.CountUse);
  GetDBStream:=P;
end;

procedure TSession.FreeDBStream(P:PDBStream);
begin
  if P=nil then exit;
  Dec(P^.CountUse);
  if P^.CountUse = 0 then Dispose(P, Done);
end;

function TSession.SystemPath:string;
begin
  if fSystemPath=nil then SystemPath:=ExePath
  else SystemPath:=fSystemPath^;
end;

procedure TSession.SetSystemPath(APath:string);
begin
  AssignStr(fSystemPath, APath);
end;

procedure TSession.AddPassword(const APassword:string);
begin
  AssignStr(DefPassword, APassword);
end;

function TSession.ByAlias(const Name:String):DirStr;
var P:PAliasRecord;
    S:string;
begin
  P:=AliasByName(Name);
  if P<>nil then
  begin
    S:=stUpCase(P^.WorkDir^);
    if (S<>'') and (S[Length(S)]<>'\') then S:=S+'\';
    ByAlias:=S;
  end
  else ByAlias:='';
end;

function TSession.AddAlias(const AWorkDir, AAliasName:string):PAliasRecord;
var i:integer;
    P:PAliasRecord;
begin
  if AliasCollection^.Search(@AAliasName,i) then AliasCollection^.AtFree(i);
  P:=New(PAliasRecord, Init(AWorkDir, AAliasName));
  AliasCollection^.Insert(P);
  AddAlias:=P;
end;

function TSession.AliasByName(const Name:String):PAliasRecord;
var i:integer;
begin
  if AliasCollection^.Search(@Name,i) then AliasByName:=AliasCollection^.At(i)
  else AliasByName:=nil;
end;

procedure TSession.StoreCfgFile(AFileName:FNameStr);
var S:PIniFile;
    i:integer;
begin
  S:=New(PIniFile, Init(AFileName));
  S^.WriteString(sLocalInfo,sDateSeparator,DateSeparator);
  S^.WriteString(sLocalInfo,sTimeSeparator,TimeSeparator);
  S^.WriteBoolean(sLocalInfo,sExpandDate,ExpandDate);
  S^.WriteString(sLocalInfo,sDecimalSeparator,DecimalSeparator);
  S^.WriteString(sLocalInfo,sCurrencyString,CurrencyString);
  S^.WriteInteger(sLocalInfo,sThousandSep,byte(ThousandSeparator));
  S^.WriteInteger(sLocalInfo, sCurrencyDec, CurrencyDecimals);
  S^.WriteString(sLocalInfo,sTextTrue,BooleanText[true]);
  S^.WriteString(sLocalInfo,sTextFalse,BooleanText[false]);
  S^.WriteString(sLocalInfo,sShortTimeFormat,ShortTimeFormat);
  S^.WriteString(sLocalInfo,sLongTimeFormat,LongTimeFormat);
  S^.WriteString(sLocalInfo,sShortDateFormat,ShortDateFormat);
  S^.WriteString(sLocalInfo,sLongDateFormat,LongDateFormat);
  S^.WriteString(sGlobalOptions,sSystemPath,DinamicStr(fSystemPath));
  S^.WriteInteger(sGlobalOptions, sCountAlias, AliasCollection^.Count);
  if DefaultDriver<>nil then
    S^.WriteInteger(sGlobalOptions, sDefaultDriver, DefaultDriver^.Version);

  for i:=0 to AliasCollection^.Count-1 do
      PAliasRecord(AliasCollection^.At(i))^.Store(S,i);
  S^.Free;
end;

procedure TSession.LoadAliases(AFileName:FNameStr);
var S:PIniFile;
    Count,i:integer;
begin
  if FileExists(AFileName) then
  begin
    if AliasCollection<>nil then Dispose(AliasCollection, Done);
    AliasCollection:=New(PAliasCollection, Init(50,50));
    S:=New(PIniFile, Init(AFileName));
    DefaultDriver:=Session^.GetCPDriver(S^.ReadInteger(sGlobalOptions, sDefaultDriver, -1));
    Count:=S^.ReadInteger(sGlobalOptions, sCountAlias, 0);
    for i:=0 to Count-1 do
       AliasCollection^.Insert(New(PAliasRecord, Load(S,i)));
    S^.Free;
  end;
end;

procedure TSession.LoadCfgFile(AFileName:FNameStr);
var S:PIniFile;
    sc:string[10];
begin
  if FileExists(AFileName) then
  begin
    S:=New(PIniFile, Init(AFileName));

    SC:=S^.ReadString(sLocalInfo,sDateSeparator,DateSeparator);
    DateSeparator:=SC[1];
    SC:=S^.ReadString(sLocalInfo,sTimeSeparator,TimeSeparator);
    TimeSeparator:=SC[1];
    ExpandDate:=S^.ReadBoolean(sLocalInfo,sExpandDate,ExpandDate);
    SC:=S^.ReadString(sLocalInfo,sDecimalSeparator,DecimalSeparator);
    DecimalSeparator:=SC[1];
    CurrencyString:=S^.ReadString(sLocalInfo,sCurrencyString,CurrencyString);
    ThousandSeparator:=Char(S^.ReadInteger(sLocalInfo,sThousandSep,byte(ThousandSeparator)));
    CurrencyDecimals:=S^.ReadInteger(sLocalInfo, sCurrencyDec, CurrencyDecimals);
    BooleanText[true]:=S^.ReadString(sLocalInfo,sTextTrue,BooleanText[true]);
    BooleanText[false]:=S^.ReadString(sLocalInfo,sTextFalse,BooleanText[false]);
    ShortTimeFormat:=S^.ReadString(sLocalInfo,sShortTimeFormat,ShortTimeFormat);
    LongTimeFormat:=S^.ReadString(sLocalInfo,sLongTimeFormat,LongTimeFormat);
    ShortDateFormat:=S^.ReadString(sLocalInfo,sShortDateFormat,ShortDateFormat);
    LongDateFormat:=S^.ReadString(sLocalInfo,sLongDateFormat,LongDateFormat);
    SetSystemPath(S^.ReadString(sGlobalOptions,sSystemPath,''));
  end;
end;

procedure CloseStream(Stream:PIndexStream);
var FreeC:PCollection;
    i:integer;
procedure DoAdd(P:PNode);
begin
  if P^.Stream=Stream then FreeC^.Insert(P);
end;
begin
  FreeC:=New(PCollection, Init(50,25));
  for i:=0 to Nodes^.Count-1 do DoAdd(Nodes^.At(i));
  for i:=0 to FreeC^.Count-1 do Nodes^.Delete(FreeC^.At(i));
  Dispose(FreeC, Done);
end;


function stdDBErrorProc( Code:TErrorCode;var Info):word;
var S:string;
begin
  stdDBErrorProc:=0;
  FormatStr(S, StdErrorNames(Code), Info);
  PrintStrLn(S);
  HasError:=true;
  if Code in NonCriticalError then exit;
  Halt(1);
end;

function StdErrorNames(Code:TErrorCode):string;
begin
  case Code of
    edbFieldNotString      : StdErrorNames := SFieldNotStrign;
    edbFieldNotInteger     : StdErrorNames := SFieldNotInteger;
    edbFieldNotBoolean     : StdErrorNames := SFieldNotBoolean;
    edbFieldNotCurrency    : StdErrorNames := SFieldNotCurrency;
    edbFieldNotDateTime    : StdErrorNames := SFieldNotDateTime;
    edbFieldNotFloat       : StdErrorNames := SFieldNotFloat;
    edbFieldNotSetData     : StdErrorNames := SFieldNotSetData;
    edbDSNotEditInsert     : StdErrorNames := SNotEditing;
    edbDSNotBrowse         : StdErrorNames := SNotBrowse;
    edbFieldNotFound       : StdErrorNames := SFieldNotFound;
    edbFileNotFound        : StdErrorNames := SFileNotFound;
    edbFileNotOpen         : StdErrorNames := SFileNotOpen;
    edbFileNotCreate       : StdErrorNames := SFileNotCreate;
    edbSesionNotInit       : StdErrorNames := SDAENotInit;
    edbInvalidIntegerValue : StdErrorNames := SInvalidIntegerValue;
    edbInvalidFloatValue   : StdErrorNames := SInvalidFloatValue;
    edbInvalidDateValue    : StdErrorNames := SInvalidDateValue;
    edbIntValueOutOfRange  : StdErrorNames := SIntValueOutOfRange;
    edbFValueOutOfRange    : StdErrorNames := SFlValueOutOfRange;
    edbfDriverNotAssigned  : StdErrorNames := SDriverNotAssigned;
    edbFileNotIndex        : StdErrorNames := SFileNotIndex;
    edbDupKey              : StdErrorNames := SDupKey;
    edbStartRangeNotInstal : StdErrorNames := SStartRangeNotInstal;
    edbEndRangeNotInstal   : StdErrorNames := SEndRangeNotInstal;
    edbSCalcFieldNA        : StdErrorNames := SSCalcFieldNA;
    edbGCalcFieldNA        : StdErrorNames := SGCalcFieldNA;
    edbLockUpFieldType     : StdErrorNames := SLockUpFieldType;
    edbIndexNotFound       : StdErrorNames := SIndexNotFound;
    edbFileNotIndexed      : StdErrorNames := SFileNotIndexed;
    edbFileAlreadyOpen     : StdErrorNames := SFileAlreadyOpen;
    edbIncompatableTable   : StdErrorNames := SIncompatableTable;
    edbErrorFreeBk         : StdErrorNames := SErrorFreeBk;
    edbErrorPosBk          : StdErrorNames := SErrorPosBk;
    edbDataSetNotClosed    : StdErrorNames := SDataSetNotClosed;
    edbDAEInvalidValue     : StdErrorNames := SDAEInvalidValue;
    edbDataSetNotLock      : StdErrorNames:='  -> %s.  ஢ .';
    edbDataSetNotUnLock    : StdErrorNames:='  -> %s.  ࠧ஢ .';
    edbNonExclus           : StdErrorNames:='  -> %s    쭮 ०.';
    edbDSNotsetKey         : StdErrorNames:='  -> %s   ० ⠭ .';
    edbRecAlreadyExist     : StdErrorNames := SRecAlreadyExist;
    edbDataSetReadOnly     : StdErrorNames := SDataSetReadOnly;
    edbBathCopyBase        :StdErrorNames:='DAE -> ।  ';
    edbTableNameRecvared   : StdErrorNames:='DAE ->   ⠡';
    edbIndexNotAssigned    :StdErrorNames:='DAE ->  %s  ᯮ (  祭 ).';
    edbIndexFieldTypeMismatch:StdErrorNames:='DAE -> 訡   ᭮ ࠦ "%s"';
    edbFileCorrupt         : StdErrorNames := SFileCorrupt;
    edbDataSetClosed       : StdErrorNames := SDataSetClosed;
    edbfInvalidFilter      : StdErrorNames := sInvalidFilter;
    edbNotValidOperation   :StdErrorNames:='DAE ->  믮   ⮬  ';
    edbSELECTExpected      :StdErrorNames:='SQL error - "SELECT" expected "%s"';
    edbFieldListExpected   :StdErrorNames:='SQL error - Field list expected "%s"';
    edbFieldOrFROMExpected :StdErrorNames:='SQL error - Field or "FROM" expected "%s"';
    edbFROMExpected        :StdErrorNames:='SQL error - "FROM" expected "%s"';
    edbFilesListExpected   :StdErrorNames:='SQL error - Files list expected "%s"';
    edbORDERBYExpected     :StdErrorNames:='SQL error -  "ORDER BY"';
    edbIndexExpected       :StdErrorNames:='SQL error - Index expected';
    edbSQLParserInit       :StdErrorNames:='SQL error - Parser init error';
    edbNotFindDS           :StdErrorNames:='SQL error -    , ᮤঠ騩  "%s"';
    edbExprError           :StdErrorNames:='訡  ࠦ ''%s''';
    edbParamNotFound       : StdErrorNames := SParamNotFound;
    edbfPrinterInit        : StdErrorNames:='ਭ  ந樠஢.';
    edbIndexFileMistMath   : StdErrorNames:='DAE ->   ᮮ⢥ ⠡ ''%s''' ;
    edbPXInitError         : StdErrorNames := SPXInitError;
    edbPXError             : StdErrorNames := SPXError;
    edbfParserError        : StdErrorNames := SParserError;
{    edbFloatRangeError:StdErrorNames:=' ''%s''. 祭 %s 諮  ।  %s - %s';}
  else
    StdErrorNames := SOtherError;
  end;
end;

function defDBTextMessageFunc( Code:TTextMessage):string;
begin
  defDBTextMessageFunc:='';
end;

function stdDBTextMessageFunc( Code:TTextMessage):string;
begin
  case Code of
    txtSQLWait:stdDBTextMessageFunc:=SWaitSQL;
    txtReportWait:stdDBTextMessageFunc:=SWaitPrinter;
  else
    stdDBTextMessageFunc:='';
  end;
end;


function defDBErrorproc( Code:TErrorCode;var Info):word;
begin
  defDBErrorproc:=0;
  HasError:=true;
  halt(1);
end;

procedure SingleError(Code:TErrorCode);
var P:Pointer;
    S:string;
begin
  P:=@S;
  S:='';
  DBError(Code{ edbSesionNotInit}, P);
end;

function fTypeToChar(T:TFieldType):char;
begin
  case t of
    ftString:fTypeToChar:='C';
    ftSmallint,
    ftInteger,
    ftWord,
    ftFloat,
    ftCurrency:fTypeToChar:='N';
    ftBoolean:fTypeToChar:='L';
    ftDate,
    ftTime,
    ftDateTime:fTypeToChar:='D';
    ftMemo:fTypeToChar:='M';
  else
    fTypeToChar:='?';
  end;
end;



  { TDBStream }
constructor TDBStream.Init(AFileName:FNameStr; AMode:word);
var P:PAliasRecord;
begin
  if isMultiUser then AMode:=AMode or stDenyNone;
  if (Session^.AliasCollection<>nil) then
  begin
    P:=Session^.AliasByName(ExtractAliasName(AFileName));
    if (P<>nil) and P^.isNet then AMode:=AMode or stDenyNone;
  end;
  AFileName:=FullName(AFileName);
  inherited Init(AFileName, AMode);
  Session^.DBStreams^.Insert(@Self);
  CountUse:=1;
  FileName:=NewStr(AFileName);
end;

destructor TDBStream.Done;
begin
  Session^.DBStreams^.Delete(@Self);
  Flush;
  if Info<>nil then
  begin
     FreeMem(Info, InfoSize);
     Info:=nil;
  end;
  DisposeStr(FileName);
  inherited Done;
end;

function TDBStream.CreateInfoBlock(ASize:longint):pointer;
begin
  GetMem(Info, ASize);
  InfoSize:=ASize;
  CreateInfoBlock:=Info;
end;

{$IFNDEF EAM}
procedure TDBStream.CopyStructure(S:PStream);
begin abstract;
end;
{$ENDIF}

   {TDataSet}
constructor TDataSet.Init;
begin
  inherited Init;
  if Session=nil then SingleError(edbSesionNotInit);
  Session^.DataCollection^.Insert(@Self);
  FieldDefs:=New(PCollection,Init(10,5));
end;

destructor TDataSet.Done;
begin
  Index:=nil;
  if FieldDefs<>nil then Dispose(FieldDefs, Done);
  Session^.DataCollection^.Delete(@Self);
  DisposeStr(MasterFieldName);
  SetFilter('');
  DisposeStr(Filter);
  inherited Done;
end;

procedure TDataSet.SetFilter(AFilter:string);
begin
  AssignStr(Filter, AFilter);
  if AFilter='' then SetFiltred(false);
end;

procedure TDataSet.SetFiltred(F:boolean);
begin
  Filtred:=F;
end;

function TDataSet.DoFilter:boolean;
begin
  DoFilter:=true;
  if Filtred and Assigned(OnFilter) then
     DoFilter:=OnFilter(@Self);
end;

procedure TDataSet.SetMaster(AMaster:PDataSet; AField:string);
begin
  if Master<>nil then Dec(Master^.CountRelation);
  Master:=AMaster;
  if AMaster<>nil then
  begin
    Inc(Master^.CountRelation);
    if Index<>nil then
       if AField<>'' then MasterField:=Master^.FieldByName(AField)
       else
       MasterField:=Master^.FieldByName( PIndexObject(Index)^.IndexField^.Name^);
    if MasterField<>nil then AssignStr(MasterFieldName, MasterField^.Name^);

    if Index<>nil then First
    else UpData;
  end;
end;

function TDataSet.GetMemoFile:PMemoFile;
begin GetMemoFile:=nil;
end;

function TDataSet.Fields(N:integer):PField;
begin
  if FieldDefs<>nil then Fields:=FieldDefs^.At(N)
  else Fields:=nil;
end;

function TDataSet.FieldByName(Name:String):PField;
var P:PField;
begin
  P:=FindField(Name);
  FieldByName:=P;
  if P=nil then DBError(edbFieldNotFound, Name);
end;

function TDataSet.FindField(Name:String):PField;
function Test(Item:PField):boolean;
begin
  Test:=stUpCase(Item^.Name^)=Name;
end;
var i:integer;
begin
  FindField:=nil;
  Name:=stUpCase(Name);
  if FieldDefs<>nil then
  for i:=0 to FieldDefs^.Count-1 do
    if stUpCase(PField(FieldDefs^.At(i))^.Name^)=Name then
    begin
      FindField:=FieldDefs^.At(i);
      break;
    end;
end;

{$IFNDEF EAM}
function TDataSet.ReadOnly:boolean;
begin abstract
end;
procedure TDataSet.RenameTable(NewTableName: string);
begin abstract
end;
procedure TDataSet.CreateIndex(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);
begin abstract
end;
function TDataSet.Open:boolean;
begin abstract
end;
procedure TDataSet.Close;
begin abstract
end;
procedure TDataSet.OpenIndex(AIndexName:FNameStr);
begin abstract
end;
function TDataSet.RecordSize:word;
begin abstract
end;
procedure TDataSet.PutFields(AName:byte;Value:string);
begin abstract
end;
function TDataSet.GetFields(AName:byte):string;
begin abstract
end;
procedure TDataSet.FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);
begin abstract
end;
function TDataSet.GetNumberFieldByName(AName:string):integer;
begin abstract
end;
function TDataSet.FieldCount:integer;
begin abstract
end;
procedure TDataSet.SetKey;
begin abstract
end;
procedure TDataSet.EditKey;
begin abstract
end;
function TDataSet.GotoKey:boolean;
begin abstract
end;
function TDataSet.FindKey:boolean;
begin abstract
end;
procedure TDataSet.Post;
begin abstract
end;
procedure TDataSet.Edit;
begin abstract
end;
procedure TDataSet.Cancel;
begin abstract
end;
procedure TDataSet.Append;
begin abstract
end;
function TDataSet.Location:boolean;
begin abstract
end;
procedure TDataSet.Delete;
begin abstract
end;
procedure TDataSet.EmptyTable;
begin abstract
end;
function TDataSet.RecordCount:Longint;
begin abstract
end;
procedure TDataSet.Last;
begin abstract
end;
procedure TDataSet.First;
begin abstract
end;
procedure TDataSet.Next;
begin abstract
end;
procedure TDataSet.Prior;
begin abstract
end;
procedure TDataSet.MoveBy(Distance:longint);
begin abstract
end;
function TDataSet.DeletedRec:boolean;
begin abstract
end;
procedure TDataSet.FieldChanged(Field:PField);
begin abstract;
end;
procedure TDataSet.FreeBookmark(Bookmark: PBookmark);
begin abstract;
end;
function TDataSet.GetBookmark:PBookmark;
begin abstract;
end;
procedure TDataSet.GotoBookmark(Bookmark: PBookmark);
begin abstract;
end;
function TDataSet.isPosBookmark(Bookmark: PBookmark):boolean;
begin abstract;
end;
procedure TDataSet.ApplyUpdates;
begin abstract;
end;
procedure TDataSet.CancelUpdates;
begin abstract;
end;
{$ENDIF}

function TDataSet.EOF:boolean;
begin
  EOF:=fEOF;
end;

function TDataSet.BOF:boolean;
begin
  BOF:=fBOF
end;

procedure TDataSet.Insert;
begin
  Append;
end;

procedure TDataSet.Go(ARecNo:longint);
procedure DoUpdata(DataSet:PDataSet);
begin
  if (DataSet^.Master=@Self) and (DataSet^.MasterField<>nil) then
     DataSet^.UpData
end;
var i:integer;
begin
  if (CountRelation<>0) and (LookFlag=0) then
    for i:=0 to Session^.DataCollection^.Count-1 do
       DoUpdata(Session^.DataCollection^.At(i));
end;

procedure TDataSet.UpData;
begin
  if (Index<>nil) and (MasterField<>nil) then
     Go(PIndexObject(Index)^.FindKey(MasterField^.IndexString, -1, true))
  else Go(Master^.RecNo)
end;

procedure TDataSet.DoAfterInsert;
begin
  if Assigned(AfterInsert) then AfterInsert(@Self);
end;

procedure TDataSet.DoBeforeInsert;
begin
  if Assigned(BeforeInsert) then BeforeInsert(@Self);
end;

procedure TDataSet.DoBeforeEdit;
begin
  if Assigned(BeforeEdit) then BeforeEdit(@Self);
end;

procedure TDataSet.DoAfterEdit;
begin
  if Assigned(AfterEdit) then AfterEdit(@Self);
end;
procedure TDataSet.DoBeforePost;
begin
  if Assigned(BeforePost) then BeforePost(@Self);
end;
procedure TDataSet.DoAfterPost;
begin
  if Assigned(AfterPost) then AfterPost(@Self);
end;
procedure TDataSet.DoBeforeCancel;
begin
  if Assigned(BeforeCancel) then BeforeCancel(@Self);
end;
procedure TDataSet.DoAfterCancel;
begin
  if Assigned(AfterCancel) then AfterCancel(@Self);
end;
procedure TDataSet.DoBeforeDelete;
begin
  if Assigned(BeforeDelete) then BeforeDelete(@Self);
end;
procedure TDataSet.DoAfterDelete;
begin
  if Assigned(AfterDelete) then AfterDelete(@Self);
end;

procedure TDataSet.Lock;
begin
  inc(LookFlag);
end;

procedure TDataSet.UnLock;
begin
  dec(LookFlag);
end;

procedure TDataSet.GotoCurent(P:PDataSet);
begin
  P^.Go(RecNo);
end;

{ TField }

constructor TField.Init(AOwner:PDataSet; AFieldNo:integer);
var fInfo:TFieldInfoRecord;
begin
  inherited Init;
  Owner:=AOwner;
  if Owner<>nil then
  begin
     FieldNo:=AFieldNo;
     Owner^.FieldInfo(FieldNo, fInfo);
     Name:=NewStr(fInfo.Name);
     FieldType:=fInfo.fType;
     Size:=fInfo.Size;
     Precision:=fInfo.Dec;
     PtrInBuf:=fInfo.Offset;
     Align:=fInfo.Align;
     SetDataPtr;
  end;
end;

procedure TField.SetDataPtr;
begin
{  fDataPtr:=Owner^.RecBuf;
  Inc(Longint(fDataPtr), PtrInBuf);}
end;

function TField.DisplayText:string;
var S:string;
begin
  case FieldType of
    ftString:S:=AsString;
    ftFloat:S:=FloatToStr(AsFloat);
    ftInteger:S:=IntToStr(AsInteger);
    ftBoolean:S:=BooleanToStr(AsBoolean);
    ftDateTime:S:=DateToStr(AsDateTime);
    else S:= '';
  end;
  {if (Length(S)<Size) and (FieldType<>ftString) then S:=DupChar(' ',Size-Length(S))+S;}
  DisplayText:=S;
end;

destructor TField.Done;
begin
  if Name<>nil then DisposeStr(Name);
  inherited Done;
end;

procedure TField.OnNewRecord;
begin
end;

procedure TField.UpdateField;
begin
  if Owner<>nil then Owner^.FieldChanged(@Self);
end;

function TField.AsString:string;
begin
  DBError(edbFieldNotString, Name);
  AsString:='';
end;

function TField.AsInteger:Longint;
begin
  DBError(edbFieldNotInteger,Name);
  AsInteger:=0;
end;

function TField.AsFloat:double;
begin
  DBError( edbFieldNotFloat, Name);
  AsFloat:=0;
end;

function TField.AsBoolean:boolean;
begin
  DBError(edbFieldNotBoolean, Name);
  AsBoolean:=false;
end;

function TField.AsDateTime:TDateTime;
begin
  DBError(edbFieldNotDateTime, Name);
  AsDateTime:=0;
end;

function TField.AsCurrency:Currency;
begin
  AsCurrency:=AsFloat;
end;

procedure TField.SetAsCurrency(D:Currency);
begin
  SetAsFloat(D);
end;

procedure TField.SetAsString(D:string);
begin
  DBError(edbFieldNotString, Name)
end;

procedure TField.SetAsInteger(D:Longint);
begin
  DBError(edbFieldNotInteger, Name)
end;

procedure TField.SetAsFloat(D:double);
begin
  DBError(edbFieldNotFloat, Name)
end;

procedure TField.SetAsBoolean(D:boolean);
begin
  DBError(edbFieldNotBoolean, Name)
end;

procedure TField.SetAsDateTime(D:TDateTime);
begin
  DBError(edbFieldNotDateTime, Name)
end;

procedure TField.Assign(AField:PField);
begin
  if AField<>nil then
  case FieldType of
    ftString:SetAsString(AField^.AsString);
    ftFloat:SetAsFloat(AField^.AsFloat);
    ftInteger:SetAsInteger(AField^.AsInteger);
    ftBoolean:SetAsBoolean(AField^.AsBoolean);
    ftDateTime:SetAsDateTime(AField^.AsDateTime);
    ftCurrency:SetAsCurrency(AField^.AsCurrency);
  end;
end;

procedure TField.SetData(var Rec:TMemoRec);
begin
  DBError(edbFieldNotSetData, Name)
end;

procedure TField.GetData(var Rec:TMemoRec);
var S:string;
begin
  S:=DisplayText;
  FillChar(Rec.Text, GetSize, 0);
  Move(S[1], Rec.Text, Length(s));
end;

{$IFDEF VER70}
function TField.IndexString:string;assembler;
asm
        PUSH    DS
        LDS     SI,Self
        LES     DI,@Result
        MOV     DX,DS:[SI].TField.PtrInBuf
        MOV     CX,DS:[SI].TField.Size
        LDS     SI,DS:[SI].TField.Owner
        LDS     SI,DS:[SI].TDataSet.RecBuf
        ADD     SI,DX
@@2:    MOV     AX,CX
        CLD
        STOSB
        JCXZ  @End
    REP MOVSB
@End:
        POP     DS
end;
{$ELSE}
function TField.IndexString:string;
begin
  IndexString:=Owner^.GetFields(FieldNo);
end;
{$ENDIF}

{$IFDEF VER70}
function TField.GetSize:word;assembler;
asm
        LES     DI,Self
        MOV     AX,ES:[DI].TField.Size
end;
{$ELSE}
function TField.GetSize:word;
begin
  GetSize:=Size;
end;
{$ENDIF}

function TField.GetWidth:word;
begin
  GetWidth:=Size+1;
end;

function TField.Calculated:boolean;
begin
  Calculated:=Owner=nil;
end;

function TField.IsIndexField:Boolean;
begin
  if (Owner=nil) or (Owner^.Index=nil) then IsIndexField:=false
  else
    IsIndexField:=PIndexObject(Owner^.Index)^.IndexField^.TestIndexField(@Self)
end;

function TField.TestIndexField(P:PField):Boolean;
begin
  TestIndexField:=@Self=P;
end;

function TField.IsValidValue(S:String):boolean;
begin
  IsValidValue:=true;
end;

procedure TField.Clear;
var i:integer;
begin
  if Owner^.State=dsInactive then DBError(edbDataSetClosed, Owner^.TableName) else
  if Owner^.State=dsBrowse then DBError(edbDSNotEditInsert, Owner^.TableName) else
  begin
    if not Calculated then
      for i:=0 to Size-1 do PCharArray(DataPtr)^[i]:=' ';
  end;
end;

procedure TField.Load(S:PReader);
begin
  Align:=TAlign(S^.ReadInteger('Align', Ord(Align)));
  isCurrency:=S^.ReadBoolean('isCurrency',isCurrency);
  ReadOnly:=S^.ReadBoolean('ReadOnly',ReadOnly);
end;

procedure TField.Store(S:PWriter);
begin
  S^.WriteInteger('Align',Ord(Align));
  S^.WriteBoolean('isCurrency', isCurrency);
  S^.WriteBoolean('ReadOnly',ReadOnly);
end;

function TField.IsNull:boolean;
var
  I:integer;
{$IFDEF VER70}
  Result:Boolean;
{$ENDIF}
begin
  Result:=true;
  for i:=0 to Size-1 do
    if PCharArray(DataPtr)^[i]<>' ' then Result:=false;
  {$IFDEF VER70}
  IsNull:=Result;
  {$ENDIF}
end;

function TField.DataPtr:Pointer;
{$IFDEF VER70}
var fDataPtr:Pointer;
{$ENDIF}
begin
  case Owner^.dsStatus of
    dsOk:Result:=Owner^.RecBuf;
    dsFindSet:Result:=Owner^.CancelBuf;
  else
    Result:=Owner^.RecBuf;
  end;
  Inc(Longint(Result), PtrInBuf);
  {$IFDEF VER70}
  DataPtr:=Result;
  {$ENDIF}
end;


  { TCustomMemoField }
function TCustomMemoField.GetWidth:word;
begin
  GetWidth:=Length(DefMemoString);
end;

procedure TCustomMemoField.AssignStrings(P:PStrings);
var L,i,Ofs:word;
    D:PMemoRec;
    S:string;
begin
  if P=nil then exit;
  if P^.Count=0 then Clear
  else
  begin
    L:=0;
    for i:=0 to P^.Count-1 do L:=L+Length(P^.Str(i))+2;
    GetMem(D, L+SizeOf(Word));
    D^.Length:=L;
    Ofs:=0;
    for i:=0 to P^.Count-1 do
    begin
      S:=P^.Str(i);
      Move(S[1], D^.Text[Ofs], Length(S));
      inc(Ofs, Length(s));
      S:=cEoL;
      Move(S[1], D^.Text[Ofs], 2);
      Inc(Ofs,2);
    end;
    SetData(D^);
    FreeMem(D, L+SizeOf(Word));
  end;
end;

{ TCalcField }
constructor TCalcField.Init(AOwner:PDataSet; AName:string; AType:TFieldType;
                            ASize:word; ADec:byte);
begin
  TObject.Init;
  Owner:=AOwner;
  Name:=NewStr(AName);
  FieldNo:=-1;
  FieldType:=AType;
  Size:=ASize;
  Precision:=ADec;
  if FieldType in [ftInteger, ftFloat, ftCurrency] then Align:=alRight
  else Align:=alLeft;
end;

function TCalcField.IsNull:boolean;
begin
  IsNull:=false;
end;

function TCalcField.IsValidValue(S:String):boolean;
begin
  IsValidValue:=true;
end;

function TCalcField.GetWidth:word;
begin
  GetWidth:=Size;
end;

function TCalcField.AsString:string;
var
  S:String;
begin
  if not Assigned(OnGet) then
     if dbDesignTime then AsString:= CalcMessage
     else  DBError( edbGCalcFieldNA, Name)
  else
     if FieldType = ftString then
     begin
        S:='';
        OnGet(@Self, Owner,S);
        AsString:=S;
     end
     else DBError( edbFieldNotString, Name)
end;

function TCalcField.AsInteger:Longint;
var
  S:Longint;
begin
  AsInteger:= 0;
  if not Assigned(OnGet) then
  begin
     if not dbDesignTime then DBError( edbGCalcFieldNA, Name)
  end
  else
     if FieldType = ftInteger then
     begin
        S:=0;
        OnGet(@Self, Owner,S);
        AsInteger:=S;
     end
     else DBError( edbFieldNotInteger, Name)
end;

function TCalcField.AsFloat:double;
var
  S:Double;
begin
  AsFloat:= 0;
  if Assigned(OnGet) then
     if FieldType = ftFloat then
     begin
        S:=0;
        OnGet(@Self, Owner,S);
        AsFloat:=S;
     end
     else DBError( edbFieldNotFloat, Name)
  else if dbDesignTime then DBError(edbGCalcFieldNA, Name);
end;

function TCalcField.AsBoolean:boolean;
var
  S:Boolean;
begin
  AsBoolean:=false;
  if not Assigned(OnGet) then
     if FieldType = ftBoolean then
     begin
        S:=false;
        OnGet(@Self, Owner,S);
        AsBoolean:=S;
     end
     else DBError( edbFieldNotBoolean, Name)
  else if not dbDesignTime then DBError( edbGCalcFieldNA, Name)
end;

function TCalcField.AsDateTime:TDateTime;
var D:TDateTime;
begin
  AsDateTime:=0;
  if not Assigned(OnGet) then
  begin
     if not dbDesignTime then DBError( edbGCalcFieldNA, Name)
  end
  else
     if FieldType = ftDateTime then
     begin
        D:=0;
        OnGet(@Self, Owner,D);
        AsDateTime:=D;
     end
     else DBError( edbFieldNotDateTime, Name)
end;

procedure TCalcField.SetAsString(D:string);
begin
  if not Assigned(OnSet) then
     DBError( edbSCalcFieldNA, Name)
  else
     if FieldType = ftString then
        OnSet(@Self, Owner,D)
     else DBError( edbFieldNotString, Name)
end;

procedure TCalcField.SetAsInteger(D:Longint);
begin
  if not Assigned(OnSet) then
     DBError( edbSCalcFieldNA, Name)
  else
     if FieldType = ftInteger then OnSet(@Self, Owner,D)
     else DBError( edbFieldNotInteger, Name)
end;

procedure TCalcField.SetAsFloat(D:Double);
begin
  if not Assigned(OnSet) then
     DBError( edbSCalcFieldNA, Name)
  else
     if FieldType = ftFloat then OnSet(@Self, Owner,D)
     else DBError( edbFieldNotFloat, Name)
end;

procedure TCalcField.SetAsBoolean(D:boolean);
begin
  if not Assigned(OnSet) then DBError( edbSCalcFieldNA, Name)
  else
  if FieldType = ftBoolean then  OnSet(@Self, Owner,D)
  else DBError( edbFieldNotBoolean, Name)
end;

procedure TCalcField.SetAsDateTime(D:TDateTime);
begin
  if Assigned(OnSet) then
     if FieldType = ftDateTime then OnSet(@Self, Owner,D)
     else DBError( edbFieldNotDateTime, Name)
  else
     DBError( edbSCalcFieldNA, Name)
end;

function TCalcField.IndexString:string;
begin
  IndexString:=DisplayText;
end;

function TCalcField.Calculated:boolean;
begin
  Calculated:=true;
end;

{MemoFile's functions}
procedure RegisterMemoFileType(ByteID:byte; AFunc:TCreateMemoFileFunc; ASystemName:string);
var R:PMemoFilesRecord;
begin
  New(R);
  R^.Next:=MemoFilesCreatedList;
  R^.ByteID:=ByteID;
  R^.CreateFunc:=AFunc;
  R^.SystemName:=NewStr(ASystemName);
  MemoFilesCreatedList:=R;
end;


 {Init Database engine}
procedure DBInit;
begin
  if Session = nil then
     Session:=New(PSession, Init);
end;

procedure DBDone;
var R:PMemoFilesRecord;
begin
  if Session<>nil then Session^.Free;
  while MemoFilesCreatedList<>nil do
  begin
    R:=MemoFilesCreatedList^.Next;
    DisposeStr(MemoFilesCreatedList^.SystemName);
    Dispose(MemoFilesCreatedList);
    MemoFilesCreatedList:=R;
  end;
  Session:=nil;
end;

  {TMemoFile}
constructor TMemoFile.Init(AOwner:PDataSet);
begin
  inherited Init;
  Owner:=AOwner;
end;

destructor TMemoFile.Done;
begin
  inherited Done;
  DisposeStr(FileName);
end;

procedure TMemoFile.RenameMemo;
var NewName:FNameStr;
begin
  NewName:=ReplaceExt(Owner^.TableName^, ExtractExt(FileName^), true);
  RenameFile(FullName(FileName^), FullName(NewName));
  AssignStr(FileName, NewName);
end;

type
  POffsetArray = ^TOffsetArray;
  TOffsetArray = array [1..10] of Char;
function TMemoFile.CalcOffset(P:Pointer):Longint;
var Offset:POffsetArray absolute P;
begin
  CalcOffset:=StrToInt(Trim(Offset^));
end;

procedure TMemoFile.StoreOffset(P:Pointer; Value:Longint);
var S:string[10];
    Offset:POffsetArray absolute P;
begin
  FillChar(Offset^, 10, ' ');
  S:=IntToStr(Value);
  Move(S[1],Offset^[10-Length(S)],Length(S));
end;

{$IFNDEF EAM}
procedure TMemoFile.Close;
begin abstract;
end;
procedure TMemoFile.Open;
begin abstract;
end;
function TMemoFile.GetSize(Offset:LongInt):Word;
begin abstract;
end;
procedure TMemoFile.Get(Offset:Longint;var D:TMemoRec);
begin abstract;
end;
function TMemoFile.Put(var D:TMemoRec):LongInt;
begin abstract;
end;
procedure TMemoFile.AtFree(Offset:LongInt);
begin abstract;
end;
{$ENDIF}

function GetPage(IO:PIndexObject; Ref:Longint):Pointer;
var Node:PNode;
    i:integer;
function DoFind(P:PNode):boolean;
begin
  if TypeOf(P^)<>TypeOf(TVMMPage) then
     DoFind:=(P^.Stream=IO^.IndexStream) and (Ref=P^.Handle)
  else DoFind:=false;
end;
begin
  Node:=nil;
  for i:=0 to Nodes^.Count-1 do
    if DoFind(Nodes^.At(i)) then
    begin
      Node:=Nodes^.At(i);
      break;
    end;
  if Node=nil then
  begin
     if Nodes^.Count<Nodes^.Limit then
        Node:=IO^.GetNode(Ref)
     else
     begin
        ReorderPagesForPage;
        Node:=IO^.GetNode(Ref)
     end;
     Nodes^.AtInsert(0, Node);
  end;
  GetPage:=Node;
end;


procedure FlushNodes(IO:PIndexObject);
procedure DoFlush(P:PNode);
begin
  if (P^.Stream=IO^.IndexStream) and P^.Modified then P^.Store;
end;
var I:integer;
begin
  for i:=0 to Nodes^.Count-1 do DoFlush(Nodes^.At(i));
end;

procedure FreePage(Node:PNode);
begin
  PIndexStream(Node^.Stream)^.FreePage(Node^.Handle);
  Nodes^.Delete(Node);
  Node^.Free;
end;

  { TNode }
constructor TNode.Init(AStream:PIndexStream; ARef:longint; ABufferSize:word);
begin
  TObject.Init;
  Stream:=AStream;
  BufferSize:=ABufferSize;
  GetMem(Buffer, BufferSize);
  Load(ARef);
end;

destructor TNode.Done;
begin
  if Modified then Store;
  FreeMem(Buffer, BufferSize);
  TObject.Done;
end;

procedure TNode.Store;
begin
  Modified:=false;
  PIndexStream(Stream)^.Modified:=true;
end;

procedure TNode.Load(ARef:Longint);
begin
  if Modified then Store;
  Handle:=ARef;
end;

 { TIndexStream }
destructor TIndexStream.Done;
begin
  CloseStream(@Self);
  inherited Done;
end;

function TIndexStream.NewPage:longint;
begin
  NewPage:=GetSize;
end;

procedure TIndexStream.Flush;
begin
  WriteHeader;
  inherited Flush;
end;

{$IFNDEF EAM}
procedure TIndexStream.FreePage(L:Longint);
begin abstract
end;
procedure TIndexStream.WriteHeader;
begin abstract
end;
function TIndexStream.GetIndex(ANum:integer):PIndexObject;
begin abstract
end;
{$ENDIF}

function TIndexStream.IndexCount:integer;
begin
  IndexCount:=1;
end;

{ TIndexObject }
constructor TIndexObject.Init(ALink:PDataSet);
begin
  inherited Init;
  Owner:=ALink;
end;

constructor TIndexObject.Create(ALink:PDataSet; AOptions:TIndexOptions);
begin
  inherited Init;
  Options:=AOptions;
  Owner:=ALink;
end;

destructor TIndexObject.Done;
begin
  Close;
  DisposeStr(FileName);
  inherited Done;
end;

procedure TIndexObject.Close;
begin
  FlushNodes(@Self);
  FileName:=NewStr(IndexStream^.FileName^);
  Session^.FreeDBStream(IndexStream);
  if ixExpression in Options then IndexField^.Free;
end;

function TIndexObject.Open:boolean;
begin
  IndexStream:=PIndexStream(Session^.GetDBStream(FileName^));
end;

{$IFNDEF EAM}
procedure TIndexObject.RenameIndex;
begin abstract;
end;
function TIndexObject.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
begin abstract;
end;
function TIndexObject.IndexExpresion:string;
begin abstract;
end;
function TIndexObject.KeySize:word;
begin abstract;
end;
function TIndexObject.NextKey:Longint;
begin abstract;
end;
function TIndexObject.PrevKey:Longint;
begin abstract;
end;
function TIndexObject.Unique:Boolean;
begin abstract;
end;
procedure TIndexObject.AddKey( AKey:String; ALink:LongInt);
begin abstract;
end;
procedure TIndexObject.DeleteKey( AKey:String; ALink:longint);
begin abstract;
end;
function TIndexObject.FirstKey:Longint;
begin abstract;
end;
function TIndexObject.LastKey:Longint;
begin abstract;
end;
procedure TIndexObject.EmptyIndex;
begin abstract;
end;
function TIndexObject.GetNode(ARef:Longint):PNode;
begin abstract;
end;
{$ENDIF}

procedure TIndexObject.ClearKey;
begin
  PP:=0;
  iStatus:=iOk;
end;

function TIndexObject.FindCurent:Longint;
begin
  FindCurent:=FindKey(IndexField^.IndexString, Owner^.RecNo, true);
end;

procedure TIndexObject.DeleteCyrrent;
begin
  DeleteKey(IndexField^.IndexString, Owner^.RecNo);
end;

procedure TIndexObject.ReIndexs;
var i,l:longint;
    P:Pointer;
begin
  P:=Owner^.Index;
  L:=Owner^.RecordCount;
  for i:=1 to L do
  begin
    Owner^.Go(i);
    if not Owner^.DeletedRec then AddKey(IndexField^.IndexString,i);
    if Assigned(DoReindex) then DoReindex(L,I, FileName);
  end;
  FlushNodes(@Self);
  Owner^.Index:=P;
end;

function TIndexObject.StartRange(const AVal:String):boolean;
var P:Pointer;
begin
  if (Owner^.Master<>nil) and (Owner^.MasterField<>nil) then
     StartRange:=ASCIICompare(AVal, TrimRight(Owner^.MasterField^.IndexString))<0
  else
  if Owner^.StartRange<>nil then
  begin
    P:=Owner^.RecBuf;
    Owner^.RecBuf:=Owner^.StartRange;
    StartRange:=ASCIICompare(AVal,TrimRight(IndexField^.IndexString))<0;
    Owner^.RecBuf:=P;
  end
  else StartRange:=false;
end;

function TIndexObject.EndRange(const AVal:String):boolean;
var P:Pointer;
begin
  if (Owner^.Master<>nil) and (Owner^.MasterField<>nil) then
     EndRange:=ASCIICompare(AVal, TrimRight(Owner^.MasterField^.IndexString))>0
  else
  if Owner^.EndRange<>nil then
  begin
    P:=Owner^.RecBuf;
    Owner^.RecBuf:=Owner^.EndRange;
    EndRange:=ASCIICompare(AVal,TrimRight(IndexField^.IndexString))>0;
    Owner^.RecBuf:=P;
  end
  else EndRange:=false;
end;

function TIndexObject.StartRangeValue:string;
var P:Pointer;
begin
  if (Owner^.Master<>nil) and (Owner^.MasterField<>nil) then
     StartRangeValue:=TrimRight(Owner^.MasterField^.IndexString)
  else
  if Owner^.StartRange<>nil then
  begin
    P:=Owner^.RecBuf;
    Owner^.RecBuf:=Owner^.StartRange;
    StartRangeValue:=TrimRight(IndexField^.IndexString);
    Owner^.RecBuf:=P;
  end
  else StartRangeValue:='';
end;

function TIndexObject.EndRangeValue:string;
var P:Pointer;
begin
  if (Owner^.Master<>nil) and (Owner^.MasterField<>nil) then
     EndRangeValue:=TrimRight(Owner^.MasterField^.IndexString)
  else
  if Owner^.StartRange<>nil then
  begin
    P:=Owner^.RecBuf;
    Owner^.RecBuf:=Owner^.EndRange;
    EndRangeValue:=TrimRight(IndexField^.IndexString);
    Owner^.RecBuf:=P;
  end
  else EndRangeValue:='';
end;

function IndexOpen(FileName:FNameStr; DS:PDataSet; OnOpenTable:boolean):PIndexObject;
var Dir: DirStr;
    Name: NameStr;
    Ext: ExtStr;
    R:PIndexObject;
begin
  R:=nil;
  if not OnOpenTable then
    FileName:=ReplaceExt(stUpCase(FileName),DefIndexExt, false);
  FSplit(FileName,Dir,Name,Ext);
  if Assigned(OnOpenIndex) then R:=OnOpenIndex(FileName, Ext, DS, OnOpenTable);
  if (R=nil) and not OnOpenTable then DBError(edbIndexNotAssigned, FileName);
  IndexOpen:=R;
end;

function IndexCreate( FileName:FNameStr; DS:PDataSet;
                      AField:string; AOptions:TIndexOptions):PIndexObject;
var Dir: DirStr;
    Name: NameStr;
    Ext: ExtStr;
    R:PIndexObject;
    P:Pointer;
begin
  R:=nil;
  FileName:=ReplaceExt(stUpCase(FileName),DefIndexExt, false);
  FSplit(FileName,Dir,Name,Ext);
  if Assigned(OnCreateIndex) then R:=OnCreateIndex(FileName, Ext, DS, AField, AOptions);
  if R=nil then
  begin
    P:=@FileName;
    DBError(edbIndexNotAssigned, P);
  end;
  IndexCreate:=R;
end;

function OpenDataSet(FileName:FNameStr):PDataSet;
var SS:PString;
    P:PDataSet;
    FName:FNameStr;
begin
  P:=nil;
  SS:=@FileName;
  FName:=ReplaceExt(FullName(FileName), '.DBF', false);
  if FileExists(FName) then
  begin
    if Assigned(DBOpen) then P:=DBOpen(FileName);
    if P=nil then DBError(edbfDriverNotAssigned, SS);
  end
  else DBError(edbFileNotFound, SS);
  OpenDataSet:=P;
end;

   {TAliasRecord}
constructor TAliasRecord.Init(const AWorkDir, AAliasName:string);
begin
  inherited Init;
  WorkDir:=NewStr(AWorkDir);
  AliasName:=NewStr(AAliasName);
end;

procedure TAliasRecord.SetFilesChange(aFilesChange:boolean);
begin
end;

function TAliasRecord.FilesChange:boolean;
begin
end;

destructor TAliasRecord.Done;
begin
  DisposeStr(WorkDir);
  DisposeStr(AliasName);
  inherited Done;
end;

constructor TAliasRecord.Load(S:PIniFile; Item:integer);
var SName:string;
begin
  inherited Init;
  SName:=sSectionName+IntToStr(Item);
  AssignStr(WorkDir,S^.ReadString(SName, sAliasDir, sCurrentDir));
  AssignStr(AliasName,S^.ReadString(SName, sAliasName, sDefAliasName+IntToStr(Item)));
  WaitTime:=S^.ReadInteger(SName, sWaitTime, 0);
  isNet:=S^.ReadBoolean(SName, sIsNet, false);
  DefaultDriver:=Session^.GetCPDriver(S^.ReadInteger(SName, sDefaultDriver, -1));
end;

procedure TAliasRecord.Store(S:PIniFile; Item:integer);
var SName:string;
begin
  SName:=sSectionName+IntToStr(Item);
  S^.WriteString(SName, sAliasDir, DinamicStr(WorkDir));
  S^.WriteString(SName, sAliasName, DinamicStr(AliasName));
  S^.WriteInteger(SName, sWaitTime, WaitTime);
  S^.WriteBoolean(SName, sIsNet, isNet);
  if DefaultDriver<>nil then
    S^.WriteInteger(SName, sDefaultDriver, DefaultDriver^.Version);
end;

   { TAliasCollection }

function TAliasCollection.Compare(Key1, Key2: Pointer): Integer;
var Rec1:PString absolute Key1;
    Rec2:PString absolute Key2;
begin
  if stUpCase(Rec1^)<stUpCase(Rec2^) then Compare:=-1
  else
  if stUpCase(Rec1^)>stUpCase(Rec2^) then Compare:=1
  else Compare:=0;
end;

function TAliasCollection.KeyOf(Item: Pointer): Pointer;
var Rec:PAliasRecord absolute Item;
begin
  KeyOf:=Rec^.AliasName;
end;

{**************}
{ TCodePageRec }

constructor TCodePageDrv.Init;
begin
  inherited Init;
  FillChar(CP, SizeOf(TCodePageArray), #32);
end;

constructor TCodePageDrv.Load(var S:TStream);
var St:string;
    C:char;
    I:Integer;
    FileVersion:Cardinal;
begin
  inherited Init;
  St:='';
  repeat
    S.Read(C, SizeOf(Char));
    if C<>#26 then St:=St+C;
  until (C=#26) or (S.Status<>stOk);
  if S.Status<>stOk then fail
  else
  begin
    S.Read(FileVersion, SizeOf(Cardinal)); { always $00000001 for DAE 2.0, reserved}
    if FileVersion>1 then fail;
    S.Read(Version, SizeOf(Cardinal));
    S.Read(SubVers, SizeOf(Cardinal));
    S.Read(CP, SizeOf(CP));
    RevertCP[32]:=' ';
    for i:=0 to 255 do
    begin
      C:=CP[i];
      if RevertCP[byte(c)]=#0 then RevertCP[byte(c)]:=char(i);
    end;
    for i:=0 to 255 do if RevertCP[i]=#0 then RevertCP[i]:=' ';
    AssignStr(fName, St);
  end;
end;

destructor TCodePageDrv.Done;
begin
  DisposeStr(fName);
  inherited Done;
end;

function TCodePageDrv.StrConvertFrom(const S:string):string;
var i:integer;
{$IFDEF VER70}
    Result:string;
{$ENDIF}
begin
  SetLength(Result, Length(S));
  for i:=1 to Length(S) do Result[i]:=RevertCP[byte(S[i])];
{$IFDEF VER70}
  StrConvertFrom:=Result;
{$ENDIF}
end;

procedure TCodePageDrv.BufConvertTO(var Buf; Len:word);
var i:word;
    RecBuf:array[0..0] of char absolute Buf;
begin
  for i:=0 to Len-1 do
    RecBuf[i]:=CP[Byte(RecBuf[i])];
end;

procedure TCodePageDrv.BufConvertFrom(var Buf; Len:word);
var i:word;
    RecBuf:array[0..0] of char absolute Buf;
begin
  for i:=0 to Len-1 do
    RecBuf[i]:=RevertCP[Byte(RecBuf[i])];
end;

function TCodePageDrv.StrConvertTo(const S:string):string;
var i:integer;
{$IFDEF VER70}
    Result:string;
{$ENDIF}
begin
  SetLength(Result, Length(S));
  for i:=1 to Length(S) do Result[i]:=CP[byte(S[i])];
{$IFDEF VER70}
  StrConvertTO:=Result;
{$ENDIF}
end;

function TCodePageDrv.Name:string;
begin
  Name:=DinamicStr(fName);
end;

{}
function FullName(Name:FNameStr):FNameStr;
var S1,S2:FNameStr;
begin
  if (Name<>'') and (Name[1]='<') and Assigned(Session) then
  begin
    S1:=Copy(Name,2,Pos('>',Name)-2);
    S2:=Copy(Name,Pos('>',Name)+1,255);
    if (S2<>'') and (S2[1]='\') then Delete(S2,1,1);
    Name:=Session^.ByAlias(S1)+S2;
  end;
  FullName:=Name;
end;

function ExtractAliasName(Name:FNameStr):string;
begin
  if (Name<>'') and (Name[1]='<') then
    ExtractAliasName:=Copy(Name,2,Pos('>',Name)-2)
  else ExtractAliasName:='';
end;

procedure InitDataSets;
begin
  InitCountryEnvironment;
  InitVMM;
  RegisterMemoFileType($02, nil, 'DBase II');
  RegisterMemoFileType($03, nil, 'DBase III Plus');
  RegisterMemoFileType($04, nil, 'DBase VII');
  RegisterMemoFileType($8C, nil, 'DBase VII with memo');
  RegisterMemoFileType($05, nil, 'DBase V');
  DBBeginProc;
end;

procedure DoneDataSets;
begin
  DBExitProc;
  DoneVMM;
end;

{$IFDEF VER70}
procedure Copyright;assembler;
asm
  ret;
  db 13,10,'DAE Lib (Data access engine library) (c) by Alexs 1995-2000'
end;
{$ELSE}
procedure Copyright;
const
  db = #13#10'DAE Lib (Data access engine library) (c) by Alexs 1995-2000';
var cc:string;
begin
  cc:=db;
end;
{$ENDIF}

procedure StartUp;
begin
  {$IFDEF WIN32}
  {$IFDEF OEM_FILE_API}
  SetFileApisToOEM;
  {$ENDIF}
  {$ENDIF}
end;

begin
  Copyright;
  StartUp;
end.
