unit MDX_Unit;
interface
uses Objects, DAE, Ext_Type;
const
   defIndExt = '.MDX';
const
   MDXBlokSize = 512;
   MDXSignature = $02;
   MDXTagsAvail = 48;

type
   PMDXTag = ^TMDXTag;
   TMDXTag = packed record
      TagHeader   : longint;              {Tag Header block (512-byte blocks)}
      TagName     : array[0..10] of char;  {Tag Name}
      FieldTag    : byte;                 {is $10 if index uses a DBF field}
      PrevTag     : byte;                 {Points to tag < this tag}
      NextTag     : byte;                 {Points to tag > this tag}
      OwnrTag     : byte;                 {Tag that pointed here}
      TagSig      : byte;                 {Unknown, appears to always be $02}
      IndexType   : char;                 {Type of index (char, date, number)}
      Unknown3    : array[0..10] of byte;
   end;

   PTagArray     = ^TTagArray;
   TTagArray     = array[0..255] of TMDXTag;{Up to 255 tag descriptions}

   PMDXHeader = ^TMDXHeader;
   TMDXHeader = packed record
      Version      : byte;                 {Version number (2)}
      BuildDate    : array[0..2] of byte;  {YMD of last reindex}
      DBFFileName  : array[0..8] of char;  {DBF name, null terminated}
      Unknown0     : array[0..6] of char;  {unknown}
      ChunksInPage : TrueWord;                 {pages in this header}
      BytesInPage  : TrueWord;                 {Bytes in a page}
      Production   : byte;                 {1=production index, else 0}
      TagsAvail    : byte;                 {Tags that can be used (48)}
      TagDescSize  : TrueWord;                 {Tag descriptor size (32)}
      TagsUsed     : TrueWord;                 {Tags in use}
      Unknown1     : TrueWord;
      NextBlock    : longint;              {Next available block}
      FreeList     : longint;              {Starting block in the free list}
      FreeBlks     : longint;
      CreateDate   : array[0..2] of byte;  {YMD of creation date}
      Unknown3     : byte;
      Unknown4     : array[0..463+32] of byte;{filler to 512 bytes}
      TagArray     : PTagArray;
   end;
const
  HeaderSize = SizeOf(TMDXHeader)-SizeOf(PTagArray);
type
  PMDXStream = ^TMDXStream;
  TMDXStream = object(TIndexStream)
    Header:TMDXHeader;
    constructor Init(AFileName:FNameStr; Mode:TrueWord);
    destructor Done;virtual;
    function IndexCount:integer;virtual;
    function GetIndex(ANum:integer):PIndexObject;virtual;
    function NewPage:longint;virtual;
    procedure FreePage(L:Longint);virtual;
    procedure WriteHeader;virtual;
    procedure CopyStructure(S:PStream);virtual;
  end;

type
  PMDXIndex = ^TMDXIndex;

  PMDXFile  = ^TMDXFile;
  TMDXFile  = object(TCollection)
    Stream:PMDXIndex;
//    constructor Init;
  end;

  TMDXIndex = object(TIndexObject)
    Header:PMDXHeader;
//    Owner:PMDXFile;
  end;

implementation
uses Dos;

  { TMDXStream }
constructor TMDXStream.Init(AFileName:FNameStr; Mode:TrueWord);
begin
  inherited Init(AFileName, Mode);
  if Mode<>stCreate then
  begin
    Read(Header, HeaderSize);
    GetMem(Header.TagArray, SizeOf(TMDXTag)*Header.TagsUsed);
    Read(Header.TagArray^, SizeOf(TMDXTag)*Header.TagsUsed);
  end;
end;

destructor TMDXStream.Done;
begin
  FreeMem(Header.TagArray, SizeOf(TMDXTag)*Header.TagsUsed);
  inherited Done;
end;

function TMDXStream.IndexCount:integer;
begin
end;

function TMDXStream.GetIndex(ANum:integer):PIndexObject;
begin
end;

function TMDXStream.NewPage:longint;
begin
end;

procedure TMDXStream.FreePage(L:Longint);
begin
end;

procedure TMDXStream.WriteHeader;
begin
  Seek(0);
  Write(Header, SizeOf(TMDXHeader));
end;

procedure TMDXStream.CopyStructure(S:PStream);
begin
end;


const
   OldOnOpenIndex:TIndexOpen=nil;
   OldOnCreateIndex:TIndexCreate=nil;

function MDXOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet; OnOpenTable:boolean):PIndexObject;far;
var TempName:string;
begin
  if OnOpenTable then
  begin
  end;
  if Ext=defIndExt then
     MDXOpen:=New(PMDXIndex, Init(DS))
  else
  if Assigned(OldOnOpenIndex) then
     MDXOpen:=OldOnOpenIndex(FileName,Ext,DS,OnOpenTable)
  else MDXOpen:=nil;
end;

function MDXCreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;far;
begin
//  if Ext=defIndExt then MDXCreate:=New(PMDXIndex, Create(DS, FileName, AField, AOptions))
//  else
  if Assigned(OldOnCreateIndex) then
     MDXCreate:=OldOnCreateIndex(FileName,Ext, DS, AField, AOptions)
  else MDXCreate:=nil;
end;

begin
(*{  OldOnOpenIndex:=OnOpenIndex;
   OldOnCreateIndex:=OnCreateIndex;
   {$IFNDEF FPC}
   OnCreateIndex:=IndCreate;
   OnOpenIndex:=IndOpen;
   {$ELSE}
   OnCreateIndex:=@IndCreate;
   OnOpenIndex:=@IndOpen;
   {$ENDIF}} *)
end.
