{$O-} {$I Define.inc}
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Paradox data file object interfase              }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}
unit PXTable;
interface
uses DAE, PXEngine, Objects, Ext_Dos, Ext_Type;
type
   PRecordHandle = ^RecordHandle;

type
   PPXTable = ^TPXTable;
   TPXTable = object(TDataSet)
      Handle:TableHandle;           { returns handle for opened table }
      constructor Init(AName:FNameStr; AMode:word);
      destructor Done;virtual;
      procedure IndexCreate(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);virtual;
      procedure IndexOpen(AIndexName:FNameStr);virtual;

      function RecordSize:word;virtual;
      function RecordCount:Longint;virtual;
      function FieldCount:integer;virtual;

      procedure FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);virtual;
      function GetNumberFieldByName(AName:string):integer;virtual;

{      function EOF:boolean;virtual;
      function BOF:boolean;virtual;}
      procedure Go(ARecNo:longint);virtual;
      procedure Last;virtual;
      procedure First;virtual;
      procedure Next;virtual;
      procedure Prior;virtual;
      procedure MoveBy(Distance:longint);virtual;

      procedure Post;virtual;
      procedure Edit;virtual;
      procedure Cancel;virtual;
      procedure Append;virtual;
      procedure Insert;virtual;
      procedure Delete;virtual;
      procedure EmptyTable;virtual;
   private
     RecHandle:RecordHandle;
     function CheckPX(C:integer):integer;
   end;
type
   PPXIndex = ^TPXIndex;
   TPXIndex = object(TIndexObject)
   end;

type
   PPXField = ^TPXField;
   TPXField = object(TField)
     constructor Init(AOwner:PDataSet; AFieldNo:integer);
     destructor Done;virtual;
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     function AsDateTime:TDateTime;virtual;
     procedure SetAsString(D:string);virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsFloat(D:Double);virtual;
     procedure SetAsBoolean(D:boolean);virtual;
     procedure SetAsDateTime(D:TDateTime);virtual;
     procedure SetData(var Rec:TMemoRec);virtual;
     procedure GetData(var Rec:TMemoRec);virtual;
     function IndexString:string;virtual;
     function Calculated:boolean;virtual;
     function IsIndexField: Boolean;virtual;
     function IsValidValue(S:String):boolean;virtual;
     procedure Clear;virtual;
     procedure UpdateField;virtual;
     function IsBlank:Boolean;
   private
     TableHandle:TableHandle;
     RecordHandle:PRecordHandle;
   end;

type
   PPXStringField = ^TPXStringField;
   TPXStringField = object(TPXField)
     function AsString:string;virtual;
     procedure SetAsString(D:string);virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsFloat(D:Double);virtual;
     procedure SetAsDateTime(D:TDateTime);virtual;
   end;

type
   PPXIntegerField = ^TPXIntegerField;
   TPXIntegerField = object(TPXField)
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsBoolean(D:boolean);virtual;
   end;

type
   PPXFloatField = ^TPXFloatField;
   TPXFloatField = object(TPXField)
     function AsString:string;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     procedure SetAsInteger(D:Longint);virtual;
     procedure SetAsFloat(D:Double);virtual;
     procedure SetAsBoolean(D:boolean);virtual;
     function DisplayText:string;virtual;
   end;

type
   PPXCurrencyField = ^TPXCurrencyField;
   TPXCurrencyField = object(TPXFloatField)
     function DisplayText:string;virtual;
   end;
type
   PPXDateField = ^TPXDateField;
   TPXDateField = object(TPXField)
     function AsString:string;virtual;
     function AsDateTime:TDateTime;virtual;
     procedure SetAsDateTime(D:TDateTime);virtual;
     function DisplayText:string;virtual;
   end;
type
   PPXBlobField = ^TPXBlobField;
   TPXBlobField = object(TPXField)
   end;

function PXFldTypeToStd(S:string; var Len:word):TFieldType;
implementation
uses Ext_Math, Ext_Str, Ext_Sys;

function PXFldTypeToStd(S:string; var Len:word):TFieldType;
begin
  Len:=StrToInt(Copy(S,2,255));
  case S[1] of
    'D':begin PXFldTypeToStd:=ftDate; Len:=8+(byte(ExpandDate) shl 1); end;
    'S':begin PXFldTypeToStd:=ftSmallint; Len:=2; end;
    'I':begin PXFldTypeToStd:=ftInteger; Len:=4; end;
    'N':begin PXFldTypeToStd:=ftFloat; Len:=15; end;
    '$':begin PXFldTypeToStd:=ftCurrency; Len:=15; end;
    'A':PXFldTypeToStd:=ftString;
    'M':PXFldTypeToStd:=ftMemo;
    'B':PXFldTypeToStd:=ftBlob;
    'F':PXFldTypeToStd:=ftFmtMemo;
    'O':PXFldTypeToStd:=ftParadoxOle;
    'G':PXFldTypeToStd:=ftGraphic;
  end;
end;

  { TPXTable }
constructor TPXTable.Init(AName:FNameStr; AMode:word);
var i,Temp,Temp2:integer;
    S:NameString;
    fldArr: FieldHandleArray;
begin
  inherited Init;
  TableName:=NewStr(AName);
  FillChar(fldArr, SizeOf(FieldHandleArray), 0);
  CheckPX(PXTblOpen(AName, Handle, 0, AMode<>stOpenRead));
  if DSStatus<>dsOk then
  begin
    DisposeStr(TableName);
    fail;
  end;

  Mode:=AMode;
  for i:=1 to FieldCount do
  begin
     PXFldType(Handle, i, S);
     case S[1] of
       'D':FieldDefs^.Insert(New(PPXDateField,Init(@Self,i)));
       'S',
       'I':FieldDefs^.Insert(New(PPXIntegerField,Init(@Self,i)));
       'N':FieldDefs^.Insert(New(PPXFloatField,Init(@Self,i)));
       '$':FieldDefs^.Insert(New(PPXCurrencyField,Init(@Self,i)));
       'A':FieldDefs^.Insert(New(PPXStringField,Init(@Self,i)));
       'M',
       'B',
       'F',
       'O',
       'G':FieldDefs^.Insert(New(PPXBlobField,Init(@Self,i)));
     else
       FieldDefs^.Insert(New(PPXField,Init(@Self,i)));
     end;
  end;
  CheckPX(PXRecNum(Handle,RecNo));
  CheckPX(PXRecBufOpen(Handle, RecHandle));
  CheckPX(PXRecGet(Handle, RecHandle));
end;

destructor TPXTable.Done;
begin
  CheckPX(PXRecBufClose(RecHandle));
  CheckPX(PXTblClose(Handle));
  DisposeStr(TableName);
  inherited Done;
end;

procedure TPXTable.IndexCreate(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);
begin abstract
end;

procedure TPXTable.IndexOpen(AIndexName:FNameStr);
begin abstract
end;

function TPXTable.RecordSize:word;
begin
end;

function TPXTable.FieldCount:integer;
var NFlds:integer;
begin
   CheckPX(PXRecNFlds(Handle, NFlds));
   FieldCount:=NFlds;
end;

function TPXTable.GetNumberFieldByName(AName:string):integer;
var I:word;
begin
   CheckPX(PXFldHandle(Handle, AName, i));
   GetNumberFieldByName:=i;
end;

procedure TPXTable.FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);
begin
end;

function TPXTable.RecordCount:Longint;
var N:Longint;
begin
  CheckPX(PXTblNRecs(Handle, N));
  RecordCount:=N;
end;

procedure TPXTable.Go(ARecNo:longint);
begin
  CheckPX(PXRecGoto(Handle, ARecNo));
  RecNo:=ARecNo;
  PXRecGet(Handle, RecHandle);
end;

{procedure TPXTable.UpdatePos;
begin
  PXRecNum(Handle,PtrRecord);
end;}

function TPXTable.CheckPX(C:integer):integer;
var Rec:array[0..1]of Longint;
begin
  fEof:=C=PXERR_ENDOFTABLE;
  fBof:=C=PXERR_STARTOFTABLE;
  if not (c in [0,PXERR_ENDOFTABLE,PXERR_STARTOFTABLE]) then
  begin
    case C of
      PXERR_FILEBUSY,
      PXERR_FILELOCKED,
      PXERR_FILENOTFOUND:DBError(edbFileNotOpen, TableName);
    else
      Rec[0]:=Longint(TableName);
      Rec[1]:=C;
      DBError(edbPXError, Rec);
    end;
    DSStatus:=dsErrorOpenFile;
  end;
end;

procedure TPXTable.Last;
begin
  CheckPX(PXRecLast(Handle));
  PXRecNum(Handle, RecNo);
  PXRecGet(Handle, RecHandle);
end;

procedure TPXTable.First;
begin
  CheckPX(PXRecFirst(Handle));
  PXRecNum(Handle,RecNo);
  PXRecGet(Handle, RecHandle);
end;

procedure TPXTable.Next;
begin
  CheckPX(PXRecNext(Handle));
  PXRecNum(Handle,RecNo);
  PXRecGet(Handle, RecHandle);
end;

procedure TPXTable.Prior;
begin
  CheckPX(PXRecPrev(Handle));
  PXRecNum(Handle,RecNo);
  PXRecGet(Handle, RecHandle);
end;

procedure TPXTable.MoveBy(Distance:longint);
begin
  if Distance=0 then exit;
  if Distance>0 then
  begin
    while (Distance<>0) and not fEof do
    begin
      CheckPX(PXRecNext(Handle));
      dec(Distance);
    end
  end
  else
  begin
    while (Distance<>0) and not fBof do
    begin
      CheckPX(PXRecPrev(Handle));
      dec(Distance);
    end
  end;
  PXRecNum(Handle,RecNo);
  PXRecGet(Handle, RecHandle);
end;

procedure TPXTable.Post;
begin
  if State=dsBrowse then exit;
  DoBeforePost;
  case State of
    dsEdit:CheckPX(PXRecUpdate(Handle, RecHandle));
    dsInsert:CheckPX(PXRecInsert(Handle, RecHandle));
    dsAppend:CheckPX(PXRecAppend(Handle, RecHandle));
{    dsSetKey:_SetKey;
    dsEditRangeStart,
    dsEditRangeEnd:ApplyRange;}
  else
    DBError(edbDSNotEditInsert, TableName^);
  end;
  DoAfterPost;
  State:=dsBrowse;
end;

procedure TPXTable.Cancel;
begin
  if State=dsBrowse then
     DBError(edbDSNotEditInsert, TableName^)
  else
  begin
    DoBeforeCancel;
    State:=dsBrowse;
    PXRecGet(Handle, RecHandle);
    DoAfterCancel;
  end;
end;

procedure TPXTable.Edit;
begin
  if State<>dsBrowse then
       DBError(edbDSNotBrowse, TableName^)
  else
  begin
    DoBeforeEdit;
    State:=dsEdit;
    DoAfterEdit;
  end;
end;

procedure TPXTable.Append;
begin
  if State<>dsBrowse then
       DBError(edbDSNotBrowse, TableName^)
  else
  begin
    DoBeforeInsert;
    PXRecBufEmpty(Handle);
    State:=dsAppend;
    DoAfterInsert;
  end;
end;

procedure TPXTable.Insert;
begin
  if State<>dsBrowse then
       DBError( edbDSNotBrowse, TableName^)
  else
  begin
    DoBeforeInsert;
    PXRecBufEmpty(Handle);
    State:=dsInsert;
    DoAfterInsert;
  end;
end;

procedure TPXTable.Delete;
begin
  DoBeforeDelete;
  CheckPX(PXRecDelete(Handle));
  DoAfterDelete;
end;

procedure TPXTable.EmptyTable;
begin
  CheckPX(PXTblEmpty(TableName^));
end;
   { TPXField }

constructor TPXField.Init(AOwner:PDataSet; AFieldNo:integer);
var C:integer;
    S:NameString;
begin
  TObject.Init;
  C:=0;
  Owner:=AOwner;
  FieldNo:=AFieldNo;
  if Owner<>nil then
  begin
    TableHandle:=PPXTable(Owner)^.Handle;
    RecordHandle:=@(PPXTable(Owner)^.RecHandle);
    C:=C+PXFldName(TableHandle, AFieldNo, S);
    Name:=NewStr(S);
    PXFldType(TableHandle,AFieldNo, S);
    FieldType:=PXFldTypeToStd(S, Size);
    if FieldType in [ftCurrency,ftFloat] then Align:=alRight
    else Align:=alLeft;
  end;
end;

destructor TPXField.Done;
begin
  inherited Done;
end;

function TPXField.AsString:string;
var S:string;
    C:integer;
begin
  DBError(edbFieldNotString, Name^);
end;
function TPXField.AsInteger:Longint;
begin
  DBError(edbFieldNotInteger, Name^);
end;
function TPXField.AsFloat:double;
begin
  DBError(edbFieldNotFloat, Name^);
end;
function TPXField.AsBoolean:boolean;
begin
  DBError(edbFieldNotBoolean, Name^);
end;
function TPXField.AsDateTime:TDateTime;
begin
  DBError(edbFieldNotDateTime, Name^);
end;
procedure TPXField.SetAsString(D:string);
begin
  DBError(edbFieldNotString, Name^);
end;
procedure TPXField.SetAsInteger(D:Longint);
begin
  DBError(edbFieldNotInteger, Name^);
end;
procedure TPXField.SetAsFloat(D:Double);
begin
  DBError(edbFieldNotFloat, Name^);
end;
procedure TPXField.SetAsBoolean(D:boolean);
begin
  DBError(edbFieldNotBoolean, Name^);
end;
procedure TPXField.SetAsDateTime(D:TDateTime);
begin
  DBError(edbFieldNotDateTime, Name^);
end;
procedure TPXField.SetData(var Rec:TMemoRec);
begin abstract;
end;
procedure TPXField.GetData(var Rec:TMemoRec);
begin abstract;
end;
function TPXField.IndexString:string;
begin abstract;
end;
function TPXField.Calculated:boolean;
begin abstract;
end;
function TPXField.IsIndexField: Boolean;
begin abstract;
end;
function TPXField.IsValidValue(S:String):boolean;
begin abstract;
end;

procedure TPXField.Clear;
begin
  if Owner^.State=dsBrowse then
    DBError(edbDSNotEditInsert, Owner^.TableName^)
  else
    PXPutBlank(RecordHandle^, FieldNo);
end;

procedure TPXField.UpdateField;
begin abstract;
end;

function TPXField.IsBlank:Boolean;
var
  Blank:Boolean;
begin
  PXFldBlank(RecordHandle^, FieldNo, Blank);
  IsBlank:=Blank;
end;

 { TPXStringField }
function TPXStringField.AsString:string;
var S:string;
    C:integer;
begin
  C:=PXGetAlpha(RecordHandle^, FieldNo, S);
  AsString:=S;
end;

procedure TPXStringField.SetAsString(D:string);
var C:integer;
begin
  if Owner^.State=dsBrowse then DBError(edbDSNotEditInsert, Owner^.TableName^)
  else
  begin
     C:=PXPutAlpha(RecordHandle^, FieldNo, D);
  end;
end;

procedure TPXStringField.SetAsInteger(D:Longint);
begin
  SetAsString(IntToStr(D));
end;

procedure TPXStringField.SetAsFloat(D:Double);
begin
  SetAsString(FloatToStr(D));
end;

procedure TPXStringField.SetAsDateTime(D:TDateTime);
begin
  SetAsString(DateToStr(D));
end;

   { TPXIntegerField }
function TPXIntegerField.AsString:string;
begin
  AsString:=IntToStr(AsInteger);
end;

function TPXIntegerField.AsInteger:Longint;
var L:Longint;
    LL:integer absolute L;
    C:integer;
begin
  L:=0;
  if not isBlank then
  case FieldType of
    ftSmallint:C:=PXGetShort(RecordHandle^, FieldNo, LL);
    ftInteger:C:=PXGetLong(RecordHandle^, FieldNo, L);
  end;
  AsInteger:=L;
end;

function TPXIntegerField.AsFloat:double;
begin
  AsFloat:=AsInteger;
end;

function TPXIntegerField.AsBoolean:boolean;
begin
  AsBoolean:=AsInteger<>0;
end;

procedure TPXIntegerField.SetAsInteger(D:Longint);
var C:word;
begin
  if Owner^.State=dsBrowse then DBError(edbDSNotEditInsert, Owner^.TableName^)
  else
  begin
     case FieldType of
        ftSmallint:C:=PXPutShort(RecordHandle^, FieldNo, D);
        ftInteger:C:=PXPutLong(RecordHandle^, FieldNo, D);
     end;
  end;
end;

procedure TPXIntegerField.SetAsBoolean(D:boolean);
begin
  SetAsInteger(byte(D));
end;

   { TPXFloatField }
function TPXFloatField.AsString:string;
begin
  AsString:=FloatToStr(AsFloat);
end;

function TPXFloatField.AsFloat:double;
var C:integer;
    D:Double;
begin
  if isBlank then D:=0
  else PXGetDoub(RecordHandle^, FieldNo, D);
  AsFloat:=D;
end;

function TPXFloatField.AsBoolean:boolean;
begin
  AsBoolean:=AsFloat<>0;
end;

procedure TPXFloatField.SetAsInteger(D:Longint);
begin
  SetAsFloat(D);
end;

procedure TPXFloatField.SetAsFloat(D:Double);
var C:integer;
begin
  if Owner^.State=dsBrowse then DBError(edbDSNotEditInsert, Owner^.TableName^)
  else
  begin
     C:=PXPutDoub(RecordHandle^, FieldNo, D);
  end;
end;

procedure TPXFloatField.SetAsBoolean(D:boolean);
begin
  SetAsFloat(byte(D));
end;

function TPXFloatField.DisplayText:string;
var
  s,s1:string[22];
  I,j: Integer;
  D:Double;
begin
  if isBlank then S1:=''
  else
  begin
    PXGetDoub(RecordHandle^, FieldNo, D);
    Str( D:21:CurrencyDecimals,S);
    i:=20-CurrencyDecimals;
    j:=0;
    S1:=DecimalSeparator+copy(S,21-CurrencyDecimals+1,255);
    S:=TrimLeft(S);
    while (i>0) and (s[i]<>' ') do
    begin
      S1:=S[i]+S1;
      inc(j);
      if j=3 then
      begin
        j:=0;
        S1:=ThousandSeparator+S1;
      end;
      System.dec(i);
    end;
  end;
  DisplayText:=S1;
end;

 { TPXCurrencyField }
function TPXCurrencyField.DisplayText:string;
begin
  if IsBlank then DisplayText:=''
  else DisplayText:=CurrencyToStr(AsFloat);
end;

  { TPXDateField }
function TPXDateField.AsString:string;
begin
  AsString:=DisplayText;
end;

function TPXDateField.AsDateTime:TDateTime;
var C:integer;
    Value:TDate;
begin
  if isBlank then AsDateTime:=0
  else
  begin
    C:=PXGetDate(RecordHandle^, FieldNo, Value);
    AsDateTime:=Value-DateDelta;
  end;
end;

function TPXDateField.DisplayText:string;
var
  Value:TDate;
begin
  if isBlank then DisplayText:=''
  else
  begin
    PXGetDate(RecordHandle^, FieldNo, Value);
    DisplayText:=DateToStr(Value-DateDelta);
  end;
end;

procedure TPXDateField.SetAsDateTime(D:TDateTime);
var
  Value:TDate;
begin
  if Owner^.State=dsBrowse then DBError(edbDSNotEditInsert, Owner^.TableName^)
  else
  begin
    Value:=DateDelta+Round(D);
    PXPutDate(RecordHandle^, FieldNo, Value);
  end;
end;

var
  OldExit:TDBExitProc;
  OldBegin:TDBExitProc;
  OldDBOpen:TDBOpenProc;


procedure InitDataSets;far;
var c:Longint;
begin
   C:=PXInit;
   if C<>0 then DBError(edbPXInitError, C);
   OldBegin;
end;

procedure DoneDataSets;far;
var c:integer;
begin
   C:=PXExit;
   OldExit;
end;

function DBOpenProc(AFileName:FNameStr):PDataSet;far;
var P:PDataSet;
    FName:FNameStr;
begin
  P:=nil;
  FName:=ReplaceExt(FullName(AFileName), '.db', false);
  if (ExtractExt(FName)='.DB') and FileExists(FName) then P:=New(PPXTable, Init(AFileName, stOpen));
  if (P=nil) and Assigned(OldDBOpen) then P:=OldDBOpen(AFileName);
  DBOpenProc:=P;
end;

begin
  OldExit:=DBExitProc;
  DBExitProc:=DoneDataSets;
  OldBegin:=DBBeginProc;
  DBBeginProc:=InitDataSets;
  OldDBOpen:=DBOpen;
  DBOpen:=DBOpenProc;
end.