{*******************************************************}
{                                                       }
{       Turbo Pascal 7                                  }
{       Free Pascal                                     }
{       Delphi 2 - 5                                    }
{       Virtual Pascal 2.1                              }
{                                                       }
{       FPT object for DBF files                        }
{                                                       }
{       Copyright (c) 1995-2001 by Alexs                }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit _FPTUnit;
interface
uses Objects, DAE, DAE_Type, Ext_Type;
const
  DefExt          = '.fpt';
  fptBlockSize    = $40;

type
  PFPTFileHeader = ^TFPTFileHeader;
  TFPTFileHeader = packed record
      NextFreeBlock:Longint;
      Reserved:TrueWord;
      BlockSize:TrueWord;
  end;

type
  PFPTMemo = ^TFPTMemo;
  TFPTMemo = object(TMemoFile)
     MemoStream:PDBStream;
     FileHeader:PFPTFileHeader;
     constructor Init(AOwner:PDataSet);
     destructor Done;virtual;
     function GetSize(Offset:LongInt):Word;virtual;
     procedure Get(Offset:LongInt;var D:TMemoRec);virtual;
     procedure AtFree(Offset:LongInt);virtual;
     function Put(var D:TMemoRec):LongInt;virtual;
     procedure Close;virtual;
     procedure Open;virtual;
     procedure RenameMemo;virtual;
  private
     procedure ReadHeader;
     procedure WriteHeader;
  end;

type
  PVFPTMemo = ^TVFPTMemo;
  TVFPTMemo = object(TFPTMemo)
    function CalcOffset(P:Pointer):Longint;virtual;
    procedure StoreOffset(P:Pointer; Value:Longint);virtual;
  end;

function CreateFPTMemoFile(ADataSet:PDataSet):PMemoFile;far;
function CreateVFPTMemoFile(ADataSet:PDataSet):PMemoFile;
implementation
uses Ext_Dos, Ext_Math, Ext_Str;

function CreateFPTMemoFile(ADataSet:PDataSet):PMemoFile;
begin
   CreateFPTMemoFile:=New(PFPTMemo, Init(ADataSet));
end;

function CreateVFPTMemoFile(ADataSet:PDataSet):PMemoFile;
begin
   CreateVFPTMemoFile:=New(PVFPTMemo, Init(ADataSet));
end;

  { TFPTMemo }
constructor TFPTMemo.Init(AOwner:PDataSet);
begin
  inherited Init(AOwner);
  AssignStr(FileName,ReplaceExt(AOwner^.TableName^, DefExt, true));
  Open;
end;

procedure TFPTMemo.Close;
begin
  WriteHeader;
  Session^.FreeDBStream(MemoStream);
  MemoStream:=nil;
  FileHeader:=nil;
end;

procedure TFPTMemo.Open;
procedure DoCreate;
var Rec:PFPTFileHeader;
    P:Pointer absolute Rec;
begin
  MemoStream:=New(PDBStream, Init(FileName^, stCreate));
  GetMem(P, 512);
  FillChar(P^, 512, 0);
  Rec^.Reserved:=fptBlockSize;
  Rec^.BlockSize:=Swap(Rec^.Reserved);
  Rec^.NextFreeBlock:=SwapLong((512 div Rec^.Reserved) + 1);
  MemoStream^.Write(Rec^, 512);
  FreeMem(P, 512);
end;
begin
  MemoStream:=Session^.GetDBStream(FileName^);
  if MemoStream=nil then
  begin
     if FileExists(FileName^) then
        MemoStream:=New(PDBStream, Init(FileName^, stOpen))
     else DoCreate;
     FileHeader:=MemoStream^.CreateInfoBlock(SizeOf(TFPTFileHeader));
     ReadHeader;
  end
  else FileHeader:=MemoStream^.Info;
end;

procedure TFPTMemo.ReadHeader;
begin
  with MemoStream^ do
  begin
    Seek(0);
    Read(FileHeader^, SizeOf(TFPTFileHeader));
    FileHeader^.NextFreeBlock:=SwapLong(FileHeader^.NextFreeBlock);
    FileHeader^.BlockSize:=Swap(FileHeader^.BlockSize);
  end;
end;

procedure TFPTMemo.RenameMemo;
var NewName:string;
begin
  NewName:=ReplaceExt(Owner^.TableName^, defExt, true);
  if FileExists(NewName) then DeleteFile(NewName);
  RenameFile(FullName(FileName^), FullName(NewName));
  AssignStr(FileName, NewName);
end;

procedure TFPTMemo.WriteHeader;
var Rec:TFPTFileHeader;
begin
  with MemoStream^ do
  begin
    Rec.NextFreeBlock:=SwapLong(FileHeader^.NextFreeBlock);
    Rec.Reserved:=FileHeader^.BlockSize;
    Rec.BlockSize:=Swap(FileHeader^.BlockSize);
    Seek(0);
    Write(Rec, SizeOf(TFPTFileHeader));
  end;
end;

destructor TFPTMemo.Done;
begin
  Close;
  inherited Done;
end;

function TFPTMemo.GetSize(Offset:LongInt):Word;
var MemoSize : longint;
begin
  if Offset=0 then GetSize:=0
  else
  begin
    MemoStream^.Seek((Offset*FileHeader^.BlockSize)+4);
    MemoStream^.Read(MemoSize,4);
    GetSize := SwapLong(MemoSize);
  end;
end;

procedure TFPTMemo.Get(Offset:LongInt;var D:TMemoRec);
begin
  D.Length := GetSize(offset);
  MemoStream^.Seek((Offset*FileHeader^.BlockSize)+8);
  MemoStream^.Read(D.Text, D.Length);
end;

procedure TFPTMemo.AtFree(Offset:LongInt);
var L:longint;
begin
{  exit; !!!!!}
  MemoStream^.Seek((offset*FileHeader^.BlockSize));
  L:=SwapLong(FileHeader^.NextFreeBlock);
  MemoStream^.Write(L,4);
  FileHeader^.NextFreeBlock := Offset;
  WriteHeader;
end;

function TFPTMemo.Put(var D:TMemoRec):LongInt;
var NFreeBl,PrevFreeblock,FoundBlock,
        countBlocks,Size, tempS,
        Offset : longint;
        Gap:Integer;
        P:Pointer;
function FreeSpace(Offset:LongInt):Word;
begin
  if Offset>=CountBlocks then FreeSpace:=$FFFF
  else FreeSpace:=(GetSize(Offset)+FileHeader^.BlockSize+7) div FileHeader^.BlockSize * FileHeader^.BlockSize;
end;

begin
  NFreeBl := FileHeader^.NextFreeBlock;
  CountBlocks := (MemoStream^.GetSize + FileHeader^.BlockSize - 1) div FileHeader^.BlockSize;
  PrevFreeBlock:=0;
  while (NFreeBl < CountBlocks) and  (FreeSpace(NFreeBl) < D.Length+8) do
  begin     { 室 訩, 㤮⢮騩 ࠧࠬ  }
    MemoStream^.Seek(NFreeBl*FileHeader^.BlockSize);
    PrevFreeBlock:=NFreeBl;
    MemoStream^.Read(NFreeBl,4);
    NFreeBl:=SwapLong(NFreeBl);
  end;
  FoundBlock:=NFreeBl;
  Put:=NFreeBl;
  Offset:=NFreeBl*FileHeader^.BlockSize;
  MemoStream^.Seek(NFreeBl*FileHeader^.BlockSize);
  if NFreebl=CountBlocks then
  begin {Append file}
    MemoStream^.Seek(MemoStream^.GetSize);
    gap:= Offset-MemoStream^.GetSize;
    if Gap>0 then
    begin
      GetMem(P,Gap);
      FillChar(P^,Gap,0);
      MemoStream^.Write(P^,Gap);
      FreeMem(P,Gap);
      NFreeBl:=CountBlocks+(D.Length+7+FileHeader^.BlockSize) div FileHeader^.blockSize;
    end
  end
  else {Read Pointer to next free block}
  begin
    MemoStream^.Seek(Offset);
    MemoStream^.Read(NFreeBl,4);
    MemoStream^.Read(Size,4);
    Size:=SwapLong(Size);
    Size:=(Size+FileHeader^.BlockSize-1) div FileHeader^.BlockSize - (D.Length+7+FileHeader^.BlockSize)
                                         div FileHeader^.BlockSize;
    if Size>0 then
    begin {Create FreeBlock from rest of found block}
      MemoStream^.Seek(Offset+Size*FileHeader^.BlockSize);
      tempS:=SwapLong(NFreeBl);
      MemoStream^.Write(NFreeBl,4);
      tempS:=SwapLong(Size);
      MemoStream^.Write(tempS,4);
      NFreeBl:=FoundBlock+Size;
    end;
    MemoStream^.Seek(Offset);
  end;
  Size:=SwapLong(D.Length);
  CountBlocks:=1;
  CountBlocks:=SwapLong(CountBlocks);
  MemoStream^.Write(CountBlocks,4);
  MemoStream^.Write(Size,4);
  MemoStream^.Write(D.Text,D.Length);
  MemoStream^.Seek(PrevFreeBlock*FileHeader^.BlockSize);
  if PrevFreeBlock=0 then FileHeader^.NextFreeBlock:=NFreeBl;
  NFreeBl:=SwapLong(NFreeBl);
  MemoStream^.Write(NFreeBl,4);
end;

  { TVFPTMemo }
function TVFPTMemo.CalcOffset(P:Pointer):Longint;
var L:^Longint absolute P;
begin
  CalcOffset:=L^;
end;

procedure TVFPTMemo.StoreOffset(P:Pointer; Value:Longint);
var L:^Longint absolute P;
begin
  L^:=Value;
end;


end.
