unit EditWind;
interface
uses Objects, DAE, Views, Dialogs, OutLine, Drivers;
const
  CSystemViewer = #6#17#16#8;
const
  cmFocusNode   = 199;
  cmNewAlias    = 198;
  cmEditAlias   = 197;
  cmDeleteAlias = 196;

type
  PSystemNode = ^TSystemNode;
  TSystemNode = object(TObject)
    Expanded:Boolean;
    Child:PSystemNode;
    Next:PSystemNode;
    constructor Init(AChild, ANext:PSystemNode);
    destructor Done;virtual;
    function GetNumChildren: Integer;
    function GetChild(I: Integer): Pointer;
    function GetText: String; virtual;
    function HasChildren: Boolean;
    function GetEditList:PCollection;virtual;
  end;

type
  PRootNode = ^TRootNode;
  TRootNode = object(TSystemNode)
    Text:PString;
    constructor Init(AText:string; Expand:boolean; AChild, ANext:PSystemNode);
    destructor Done;virtual;
    function GetText: String; virtual;
  end;

type
  TDriversFieldType = (fDateSseparator,fTimeSeparator,fExpandDate,
                   fDecimalSeparator,fCurrencyString,fThousandSeparator,
                   fCurrencyDecimals,fTextForTrue,fTextForFalse,
                   fShortTimeFormat,fLongTimeFormat,fShortDateFormat,
                   fLongDateFormat, fSystemPath, fDefaultDriver);
  TDriversFieldSet = set of TDriversFieldType;

type
  PFormatNode = ^TFormatNode;
  TFormatNode = object(TRootNode)
    DriversFieldSet:TDriversFieldSet;
    constructor Init(AText:string; aDriversFieldSet:TDriversFieldSet; ANext:PSystemNode);
    function GetEditList:PCollection;virtual;
  end;

type
  PAliasNode = ^TAliasNode;
  TAliasNode = object(TSystemNode)
    Alias:PAliasRecord;
    constructor Init(AAlias:PAliasRecord; ANext:PSystemNode);
    function GetText: String; virtual;
    function GetEditList:PCollection;virtual;
  end;

type
  PAbstractField = ^TAbstractField;
  TAbstractField = object(TObject)
    Name:PString;
    destructor Done;virtual;
    function GetText:string;virtual;
    procedure EditField(Sender:PListBox);virtual;
  end;

type
  TAliasFieldType = ( aWorkDir,aAliasName,aWaitTime, aisNet);

type
  PAliasField = ^TAliasField;
  TAliasField = object(TAbstractField)
    aFieldType:TAliasFieldType;
    Alias:PAliasRecord;
    constructor Init(AAlias:PAliasRecord; aaFieldType:TAliasFieldType);
    destructor Done;virtual;
    function GetText:string;virtual;
    procedure EditField(Sender:PListBox);virtual;
  end;

type
  PDriversField = ^TDriversField;
  TDriversField = object(TAbstractField)
    DriversFieldType:TDriversFieldType;
    constructor Init(ADriversFieldType:TDriversFieldType);
    destructor Done;virtual;
    function GetText:string;virtual;
    procedure EditField(Sender:PListBox);virtual;
  end;


type
  PCodePageDrvField = ^TCodePageDrvField;
  TCodePageDrvField = object(TAbstractField)
    CPD:PCodePageDrv;
    Alias:PAliasRecord;
    //DriversFieldType:TDriversFieldType;
    constructor Init(ACPD:PCodePageDrv; AAlias:PAliasRecord);
    function GetText:string;virtual;
    procedure EditField(Sender:PListBox);virtual;
  end;

type
  PAliasCodePageDrvField = ^TAliasCodePageDrvField;
  TAliasCodePageDrvField = object(TCodePageDrvField)
  end;

type
  PSystemViewer = ^TSystemViewer;
  TSystemViewer = object(TOutlineViewer)
    Root: PSystemNode;
    constructor Init(var Bounds: TRect; AHScrollBar,
      AVScrollBar: PScrollBar; ARoot: PSystemNode);
    procedure Adjust(Node: Pointer; Expand: Boolean); virtual;
    function GetRoot: Pointer; virtual;
    function GetNumChildren(Node: Pointer): Integer; virtual;
    function GetChild(Node: Pointer; I: Integer): Pointer; virtual;
    function GetText(Node: Pointer): String; virtual;
    function IsExpanded(Node: Pointer): Boolean; virtual;
    function HasChildren(Node: Pointer): Boolean; virtual;
    function GetPalette: PPalette; virtual;
    procedure Focused(I: Integer); virtual;
  end;

type
  PEditListBox = ^TEditListBox;
  TEditListBox = object(TListBox)
    function GetText(Item: Integer; MaxLen: Integer): String; virtual;
    procedure HandleEvent(var Event:TEvent);virtual;
  end;

type
  PAliasEditWindow = ^TAliasEditWindow;
  TAliasEditWindow = object(TDialog)
    SystemViewer:PSystemViewer;
    EditBox:PListBox;
    constructor Init;
    procedure HandleEvent(var Event:TEvent);virtual;
    procedure DeleteAlias;
{    procedure EditAlias;}
    procedure NewAlias;
  end;

const
  Modified: boolean = false;

function NewSystemRoot(Text:string; Expand:boolean; Child,Next:PSystemNode):PSystemNode;
implementation
uses App, ResTools, Clases, Ext_Math, Ext_Dos, Validate, Register, Ext_Str,
     MsgBox, Ext_Sys, AddCntrl;

function BooleanToStr(F:boolean):string;
begin
  if F then BooleanToStr:='true' else BooleanToStr:='false';
end;
  {TSystemNode}
constructor TSystemNode.Init(AChild, ANext:PSystemNode);
begin
  inherited Init;
  Child:=AChild;
  Next:=ANext;
end;

destructor TSystemNode.Done;
begin
  if Child<>nil then Dispose(Child, Done);
  if Next<>nil then Dispose(Next, Done);
  inherited Done;
end;

function TSystemNode.GetNumChildren: Integer;
var i:integer;
    AChild:PSystemNode;
begin
  AChild:=Child;
  i:=0;
  while AChild<>nil do
  begin
    AChild:=AChild^.Next;
    inc(i);
  end;
  GetNumChildren:=i;
end;

function TSystemNode.GetChild(I: Integer): Pointer;
var AChild:PSystemNode;
begin
  AChild:=Child;
  while (i<>0) and (AChild<>nil) do
  begin
    AChild:=AChild^.Next;
    Dec(i);
  end;
  GetChild:=AChild;
end;

function TSystemNode.GetText: String;
begin
end;

function TSystemNode.HasChildren: Boolean;
begin
  HasChildren:=Child<>nil;
end;

function TSystemNode.GetEditList:PCollection;
begin
  GetEditList:=nil;
end;

  { TRootNode }
constructor TRootNode.Init(AText:string; Expand:boolean; AChild, ANext:PSystemNode);
begin
  inherited Init(AChild, ANext);
  Text:=NewStr(AText);
  Expanded:=Expand;
end;

destructor TRootNode.Done;
begin
  DisposeStr(Text);
  inherited Done;
end;


function TRootNode.GetText: String;
begin
  GetText:=Text^;
end;

function NewSystemRoot(Text:string; Expand:boolean; Child,Next:PSystemNode):PSystemNode;
begin
  NewSystemRoot:=New(PRootNode, Init(Text, Expand, Child, Next));
end;

function AliasNodes:PAliasNode;
function DoAliasNode(i:integer):PAliasNode;
begin
  with Session^.AliasCollection^ do
  if I<Count then DoAliasNode:=New(PAliasNode, Init(At(i),DoAliasNode(i+1)))
  else DoAliasNode:=nil;
end;
begin
  AliasNodes:=DoAliasNode(0);
end;

function NewFormatNode(AName:string; Types:TDriversFieldSet; Next:PSystemNode):PFormatNode;
begin
  NewFormatNode:=New(PFormatNode, Init(AName, Types, Next));
end;

  { TFormatNode }
constructor TFormatNode.Init(AText:string; aDriversFieldSet:TDriversFieldSet; ANext:PSystemNode);
begin
  inherited Init(Atext, false, nil, ANext);
  DriversFieldSet:=aDriversFieldSet;
end;

function TFormatNode.GetEditList:PCollection;
var i:TDriversFieldType;
    P:PCollection;
begin
  P:=New(PCollection, Init(10,10));
  for i:=Low(TDriversFieldType) to High(TDriversFieldType) do
    if i in DriversFieldSet then
      if i=fDefaultDriver then
        P^.Insert(New(PCodePageDrvField, Init(Session^.DefaultDriver, nil)))
      else P^.Insert(New(PDriversField, Init(i)));
  GetEditList:=P;
end;

  { TAliasNode }
constructor TAliasNode.Init(AAlias:PAliasRecord; ANext:PSystemNode);
begin
  inherited Init(nil,ANext);
  Alias:=AAlias;
end;

function TAliasNode.GetText: String;
begin
  if Alias<>nil then GetText:=DinamicStr(Alias^.AliasName)
  else GetText:='';
end;

function TAliasNode.GetEditList:PCollection;
var P:PCollection;
begin
  P:=New(PCollection, Init(10,5));
  P^.Insert(New(PAliasField, Init(Alias,aWorkDir)));
  P^.Insert(New(PAliasField, Init(Alias,aAliasName)));
  P^.Insert(New(PAliasField, Init(Alias,aWaitTime)));
  P^.Insert(New(PAliasField, Init(Alias,aisNet)));
  P^.Insert(New(PCodePageDrvField, Init(Alias^.DefaultDriver, Alias)));
  GetEditList:=P;
end;

  {TAbstractField}
destructor TAbstractField.Done;
begin
  DisposeStr(Name);
  inherited Done;
end;

function TAbstractField.GetText:string;
begin
end;

procedure TAbstractField.EditField(Sender:PListBox);
begin
end;


function EditeDate(var S:string; ListBox:PListBox; MaxLen:byte; V:PValidator):boolean;
var R:TRect;
    P:PInputLine;
    Prop:PAbstractField;
    ModalResult:word;
begin
  with ListBox^ do
  begin
    Prop:=List^.At(Focused);
    S:=Prop^.GetText;
    R.Assign(Origin.X+Length(Prop^.Name^)+3, Origin.Y+Focused-TopItem,
             Origin.X+Size.X, Origin.Y+Focused-TopItem+1);
    P:=New(PInputLine, Init(R, MaxLen));
    P^.SetValidator(V);
    P^.SetData(S);
    ModalResult:=Owner^.ExecView(P);
  end;
  if ModalResult=cmOk then P^.GetData(S);
  EditeDate:=ModalResult=cmOk;
  P^.Free;
end;

  {TAliasField}
constructor TAliasField.Init(AAlias:PAliasRecord; aaFieldType:TAliasFieldType);
begin
  inherited Init;
  Alias:=AAlias;
  aFieldType:=aaFieldType;
  case aFieldType of
    aWorkDir  :Name:=NewStr('WorkDir');
    aAliasName:Name:=NewStr('AliasName');
    aWaitTime :Name:=NewStr('WaitTime');
    aisNet    :Name:=NewStr('isNet');
  end;
end;

destructor TAliasField.Done;
begin
  inherited Done;
end;

function TAliasField.GetText:string;
begin
  if Alias=nil then GetText:=''
  else
  case aFieldType of
    aWorkDir  :GetText:=DinamicStr(Alias^.WorkDir);
    aAliasName:GetText:=DinamicStr(Alias^.AliasName);
    aWaitTime :GetText:=IntToStr(Alias^.WaitTime);
    aisNet    :GetText:=BooleanToStr(Alias^.isNet);
  end;
end;

procedure TAliasField.EditField(Sender:PListBox);
var S:string;
begin
  if Alias=nil then exit;
  if not EditeDate(S, Sender, 255, nil) then exit;
  Modified:=true;
  case aFieldType of
    aWorkDir  :AssignStr(Alias^.WorkDir, S);
    aAliasName:begin
                 AssignStr(Alias^.AliasName, S);
                 if Sender^.Owner<>nil then
                    Sender^.Owner^.Redraw;
               end;
    aWaitTime :Alias^.WaitTime:=StrToInt(S);
    aisNet    :Alias^.isNet:=chUpCase(S[1]) in ['T','Y',''];
  end;
end;

  { TDriversField }
constructor TDriversField.Init(ADriversFieldType:TDriversFieldType);
begin
  inherited Init;
  DriversFieldType:=aDriversFieldType;
  case DriversFieldType of
    fDateSseparator    : Name:=NewStr('Date separator');
    fTimeSeparator     : Name:=NewStr('Time separator');
    fExpandDate        : Name:=NewStr('Expand date');
    fDecimalSeparator  : Name:=NewStr('Decimal separator');
    fCurrencyString    : Name:=NewStr('Currency string');
    fThousandSeparator : Name:=NewStr('Thousand separator');
    fCurrencyDecimals  : Name:=NewStr('Currency decimals');
    fTextForTrue       : Name:=NewStr('Text for true');
    fTextForFalse      : Name:=NewStr('Text for false');
    fShortTimeFormat   : Name:=NewStr('Short time format');
    fLongTimeFormat    : Name:=NewStr('Long time format');
    fShortDateFormat   : Name:=NewStr('Short date format');
    fLongDateFormat    : Name:=NewStr('Long date formate');
    fSystemPath        : Name:=NewStr('System path');
  end;
end;

destructor TDriversField.Done;
begin
  inherited Done;
end;

function TDriversField.GetText:string;
begin
  case DriversFieldType of
    fDateSseparator    :GetText:=DateSeparator;
    fTimeSeparator     :GetText:=TimeSeparator;
    fExpandDate        :GetText:=BooleanToStr(ExpandDate);
    fDecimalSeparator  :GetText:=DecimalSeparator;
    fCurrencyString    :GetText:=CurrencyString;
    fThousandSeparator :GetText:=ThousandSeparator;
    fCurrencyDecimals  :GetText:=IntToStr(CurrencyDecimals);
    fTextForTrue       :GetText:=BooleanText[true];
    fTextForFalse      :GetText:=BooleanText[false];
    fShortTimeFormat   :GetText:=ShortTimeFormat;
    fLongTimeFormat    :GetText:=LongTimeFormat;
    fShortDateFormat   :GetText:=ShortDateFormat;
    fLongDateFormat    :GetText:=LongDateFormat;
    fSystemPath        :GetText:=Session^.SystemPath;
  else
    GetText:='Other';
  end;
end;

procedure TDriversField.EditField(Sender:PListBox);
var S:string;
begin
  if not EditeDate(S, Sender, 255, nil) then exit;
  Modified:=true;
  case DriversFieldType of
    fDateSseparator    :DateSeparator:=S[1];
    fTimeSeparator     :TimeSeparator:=S[1];
    fExpandDate        :ExpandDate:=chUpCase(S[1]) in ['T','Y',''];
    fDecimalSeparator  :DecimalSeparator:=S[1];
    fCurrencyString    :CurrencyString:=S;
    fThousandSeparator :ThousandSeparator:=S[1];
    fCurrencyDecimals  :CurrencyDecimals:=StrToInt(S);
    fTextForTrue       :BooleanText[true]:=S;
    fTextForFalse      :BooleanText[false]:=S;
    fShortTimeFormat   :ShortTimeFormat:=S;
    fLongTimeFormat    :LongTimeFormat:=S;
    fShortDateFormat   :ShortDateFormat:=S;
    fLongDateFormat    :LongDateFormat:=S;
    fSystemPath        :Session^.SetSystemPath(S);
  end;
end;

  { TCodePageDrvField }
constructor TCodePageDrvField.Init(ACPD:PCodePageDrv; AAlias:PAliasRecord);
begin
  inherited Init;
  CPD:=ACPD;
  Alias:=AAlias;
  Name:=NewStr('Defult code page');
end;

function TCodePageDrvField.GetText:string;
var fCpd:PCodePageDrv;
begin
  if Alias<>nil then fCpd:=Alias^.DefaultDriver
  else fCpd:=Session^.DefaultDriver;
  if fCPD<>nil then GetText:=fCPD^.Name
  else GetText:='';
end;

procedure TCodePageDrvField.EditField(Sender:PListBox);
var R:TRect;
    P:PLockUpInputLine;
    Prop:PAbstractField;
    fCPD:PCodePageDrv;
    Selected,i:Longint;
begin
  with Sender^ do
  begin
    Prop:=List^.At(Focused);
    Selected:=-1;
    R.Assign(Origin.X+Length(Prop^.Name^)+3, Origin.Y+Focused-TopItem,
             Origin.X+Size.X, Origin.Y+Focused-TopItem+1);
    P:=New(PLockUpInputLine, Init(R, nil));
    for i:=0 to Session^.CodePageCollections^.Count-1 do
    begin
      fCPD:=Session^.CodePageCollections^.At(i);
      if fCPD=CPD then Selected:=i;
      P^.List^.AtInsert(i, NewStr(fCPD^.Name));
    end;
    P^.SetData(Selected);
    if Owner^.ExecView(P)=cmOk then
    begin
      P^.GetData(Selected);
      if Alias<>nil then
      begin
        if Selected=-1 then Alias^.DefaultDriver:=nil
        else Alias^.DefaultDriver:=Session^.CodePageCollections^.At(Selected);
      end
      else
      begin
        if Selected=-1 then Session^.DefaultDriver:=nil
        else Session^.DefaultDriver:=Session^.CodePageCollections^.At(Selected);
      end;
    end;
  end;
  P^.Free;
end;

  { TSystemViewer }
constructor TSystemViewer.Init(var Bounds: TRect; AHScrollBar,
      AVScrollBar: PScrollBar; ARoot: PSystemNode);
begin
  inherited Init(Bounds,AHScrollBar,AVScrollBar);
  Root:=ARoot;
  Update;
end;

procedure TSystemViewer.Adjust(Node: Pointer; Expand: Boolean);
var ANode:PSystemNode absolute Node;
begin
  ANode^.Expanded:=Expand;
end;

function TSystemViewer.GetRoot: Pointer;
begin
  GetRoot:=Root;
end;

function TSystemViewer.GetNumChildren(Node: Pointer): Integer;
var ANode:PSystemNode absolute Node;
begin
  GetNumChildren:=ANode^.GetNumChildren;
end;

function TSystemViewer.GetChild(Node: Pointer; I: Integer): Pointer;
var ANode:PSystemNode absolute Node;
begin
  GetChild:=ANode^.GetChild(i);
end;

function TSystemViewer.GetText(Node: Pointer): String;
var ANode:PSystemNode absolute Node;
begin
  GetText:=ANode^.GetText;
end;

function TSystemViewer.IsExpanded(Node: Pointer): Boolean;
var ANode:PSystemNode absolute Node;
begin
  IsExpanded:=ANode^.Expanded;
end;

function TSystemViewer.HasChildren(Node: Pointer): Boolean;
var ANode:PSystemNode absolute Node;
begin
  HasChildren:=ANode^.HasChildren;
end;

function TSystemViewer.GetPalette: PPalette;
const
  P: String[Length(CSystemViewer)] = CSystemViewer;
begin
  GetPalette := @P;
end;

procedure TSystemViewer.Focused(I: Integer);
begin
  inherited Focused(I);
  Message(Owner, evBroadcast, cmFocusNode, GetNode(I));
end;

  { TEditListBox }
function TEditListBox.GetText(Item: Integer; MaxLen: Integer): String;
var P:PAbstractField;
begin
  P:=List^.At(Item);
  GetText:=P^.Name^+' : '+P^.GetText;
end;

procedure TEditListBox.HandleEvent(var Event:TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What=evKeyDown) and (Event.KeyCode=kbF4) and
     (List<>nil) and (List^.Count>0) then SelectItem(Focused)
end;

  {TAliasEditWindow}
constructor TAliasEditWindow.Init;
var R:TRect;
    SH,SV:PScrollBar;
begin
  Desktop^.GetExtent(R);
  inherited Init(R,'Configure Data Access Engine utilite');
  Palette:=dpBlueDialog;
  GetExtent(R);
  Inc(R.A.Y);
  Dec(R.B.Y);
  R.A.X:=21;
  R.B.X:=R.A.X+1;
  SV:=Insert(New(PScrollBar, Init(R)));
  GetExtent(R);
  Inc(R.A.X);
  R.A.Y:=R.B.Y-1;
  R.B.X:=R.A.X+20;
  SH:=Insert(New(PScrollBar, Init(R)));
  GetExtent(R);
  R.Grow(-1,-1);
  R.B.X:=R.A.X+20;
  SystemViewer:=Insert(New(PSystemViewer, Init(R,SH,SV,
    NewSystemRoot('Root', true,
      NewSystemRoot('Alias', false,
        AliasNodes,
      NewSystemRoot('Drivers', false,
        NewFormatNode('Formats', [fDateSseparator, fTimeSeparator,
                                  fExpandDate, fDecimalSeparator,
                                  fThousandSeparator, fCurrencyString,
                                  fCurrencyDecimals],
        NewFormatNode('System',[fTextForTrue,fTextForFalse,fShortTimeFormat,
                                fLongTimeFormat,fShortDateFormat,
                                fLongDateFormat, fSystemPath, fDefaultDriver],
        nil)),
      nil)),
    nil)
    )));
  GetExtent(R);
  Inc(R.A.Y);
  Dec(R.B.Y);
  R.A.X:=R.B.X-1;
  SV:=Insert(New(PScrollBar, Init(R)));
  GetExtent(R);
  R.Grow(-1,-1);
  inc(R.A.X,21);
  EditBox:=Insert(New(PEditListBox, Init(R,1,SV)));
  EditBox^.Options:=EditBox^.Options or ofFramed;
  SystemViewer^.Select;
  SystemViewer^.Focused(0);
end;

procedure TAliasEditWindow.HandleEvent(var Event:TEvent);
procedure DoAddEditor;
var P:PCollection;
    Node:PSystemNode;
begin
  Node:=Event.InfoPtr;
  if Node<>nil then P:=Node^.GetEditList
  else P:=nil;
  EditBox^.NewList(P);
  SetCmdState([cmEditAlias,cmDeleteAlias], TypeOf(Node^)=TypeOf(TAliasNode));
end;

procedure EditField;
var Field:PAbstractField;
begin
  Field:=EditBox^.List^.At(EditBox^.Focused);
  Field^.EditField(EditBox);
  EditBox^.DrawView;
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evBroadcast:
    case Event.Command of
      cmFocusNode:DoAddEditor;
      cmListItemSelected:EditField;
    else
      exit;
    end;
    evCommand:
    case Event.Command of
      cmDeleteAlias:DeleteAlias;
      cmNewAlias:NewAlias;
    else
      exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TAliasEditWindow.DeleteAlias;
var
  Node,N1:PAliasNode;
begin
  Node:=SystemViewer^.GetNode(SystemViewer^.Foc);
  if (TypeOf(Node^)=TypeOf(TAliasNode)) and
     (MessageBox('Delete alias '''+Node^.GetText+''' ?',
                  nil, mfConfirmation+mfOkCancel)<>cmCancel) then
  begin
    if PAliasNode(SystemViewer^.Root^.Child^.Child)=Node then
       SystemViewer^.Root^.Child^.Child:=Node^.Next
    else
    begin
      N1:=PAliasNode(SystemViewer^.Root^.Child^.Child);
      while (PAliasNode(N1^.Next)<>Node) and (N1^.Next<>nil) do
        N1:=PAliasNode(N1^.Next);
      if N1^.Next<>nil then N1^.Next:=Node^.Next;
    end;
    Node^.Next:=nil;
    Session^.AliasCollection^.Delete(Node^.Alias);
    Session^.AliasCollection^.FreeItem(Node^.Alias);
    Dispose(Node, Done);
    SystemViewer^.Update;
    SystemViewer^.Focused(2);
    SystemViewer^.DrawView;
    Modified:=true;
  end;
end;

procedure TAliasEditWindow.NewAlias;
var S:string;
    A:PAliasRecord;
    P:PAliasNode;
begin
  S:='';
  if InputBox('Enter alias name.','~N~ame',S, 60)<>cmOk then exit;
  Modified:=true;
  A:=Session^.AddAlias('',S);
  P:=New(PAliasNode, Init(A, SystemViewer^.Root^.Child^.Child));
  SystemViewer^.Root^.Child^.Child:=P;
  SystemViewer^.Adjust(SystemViewer^.Root^.Child, true);
  SystemViewer^.ExpandAll(SystemViewer^.Root^.Child^.Child);
  SystemViewer^.Update;
  SystemViewer^.Focused(2);
  SystemViewer^.DrawView;
end;

end.

